// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef CYCLEDETECT_H
#define CYCLEDETECT_H

#include <seqan/graph_algorithms.h>
#include <seqan/map.h>
#include <Types.h>
#include <SparsePairGraph.h>



/**
.Function.dijkstra:
..cat:Graph
..summary:Computes shortest paths from a single source in a graph.
..remarks:Edge weights have to be nonnegative.
..signature:dijkstra(g, source, weight, predecessor, distance)
..param.g:In-parameter:A graph.
...type:Spec.Directed Graph
..param.source:In-parameter:A source vertex.
...type:Metafunction.VertexDescriptor
..param.weight:In-parameter:A weight map.
...remarks:A property map with edge weights. Edge weights have to be nonnegative.
..param.predecessor:Out-parameter:A property map.
...remarks:A property map that represents predecessor relationships among vertices. It determines a shortest-paths tree.
..param.distance:Out-parameter:A property map.
...remarks:Indicates for each vertex the distance from the source.
..returns:void
..see:Function.dagShortestPath
..see:Function.bellmanFordAlgorithm
..include:seqan/graph_algorithms.h
*/
template<typename TSpec, typename TVertexDescriptor, typename TWeightMap, typename TPredecessorMap, typename TDistanceMap, typename TDistType>
void
dijkstra(seqan::Graph<TSpec> const& g,
         TVertexDescriptor const source,
         TWeightMap const& weight,
         TPredecessorMap& predecessor,
         TDistanceMap& distance,
         TDistType bound)
{
    using namespace seqan;

    SEQAN_CHECKPOINT
    typedef Graph<TSpec> TGraph;
    typedef typename Size<TGraph>::Type TSize;
    typedef typename Value<TDistanceMap>::Type TDistVal;
    typedef typename Iterator<TGraph, VertexIterator>::Type TVertexIterator;
    typedef typename Iterator<TGraph, OutEdgeIterator>::Type TOutEdgeIterator;

    // Initialization
    resizeVertexMap(g,predecessor);
    resizeVertexMap(g,distance);

    // S is initially empty
    String<bool> setS;
    resize(setS, getIdUpperBound(_getVertexIdManager(g)), false);

    // Set-up the priority queue
    typedef Pair<TVertexDescriptor, TDistVal> TKeyValue;
    typedef HeapTree<TKeyValue, std::less<TDistVal>, KeyedHeap<> > TKeyedHeap;
    TKeyedHeap priorityQueue;
    TDistVal infDist = _getInfinityDistance(weight);
    TVertexDescriptor nilVertex = getNil<typename VertexDescriptor<TGraph>::Type>();
    TVertexIterator it(g);
    for(;!atEnd(it);goNext(it)) {
        assignProperty(predecessor, value(it), nilVertex);
        assignProperty(distance, value(it), infDist);
        heapInsert(priorityQueue, TKeyValue(value(it), infDist));
    }
    assignProperty(distance, source, 0);
    heapChangeValue(priorityQueue, source, 0);

    // Run Dijkstra
    while (!empty(priorityQueue)) {
        // Extract min
        TVertexDescriptor u = heapExtractRoot(priorityQueue).i1;
        if (getProperty(distance, u) > bound)
            break;

        assignProperty(setS, u, true);
        TOutEdgeIterator itout(g, u);
        for(;!atEnd(itout);++itout) {
            TVertexDescriptor v = targetVertex(itout);
            if (property(setS, v) == true) continue;
            if (getProperty(distance, v) > getProperty(distance,u) + getProperty(weight,value(itout)) &&
                    (getProperty(distance,u) + getProperty(weight,value(itout)) < bound))
            {
                assignProperty(distance, v, getProperty(distance,u) + getProperty(weight,value(itout)));
                assignProperty(predecessor, v, u);
                heapChangeValue(priorityQueue, v, getProperty(distance,u) + getProperty(weight,value(itout)));
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// detect cycles -- small modification of seqan DFS code
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

template<typename TVertexDescriptor, typename TPredecessorMap, typename TDiscoveryTimeMap, typename TFinishingTimeMap, typename TVal, typename TCycleString, typename TScoreString, typename TWeightMap>
void
cycleVisit(const TDirGraph & g,
           TVertexDescriptor const u,
           TPredecessorMap& predecessor,
           TDiscoveryTimeMap& disc,
           TFinishingTimeMap& finish,
           TVal& time,
           TWeightMap vertexWeights,
           TCycleString & cycles,
           TScoreString & scores,
           double & threshold
           )
{
    using namespace seqan;

    SEQAN_CHECKPOINT

    typedef typename Iterator<TDirGraph, AdjacencyIterator>::Type TAdjacencyIterator;

    //assignProperty(tokenMap, u, true);
    assignProperty(disc, u, time);
    //std::cerr << "activate vertex: " << u << std::endl;
    ++time;
    TAdjacencyIterator itad(g,u);
    for(;!atEnd(itad);goNext(itad))
    {
        TVertexDescriptor v = getValue(itad);
        //std::cerr << "peek vertex: " << v << "  disc:  " << getProperty(disc, v) << std::endl;
        if (getProperty(disc, v) == -1)
        {
            assignProperty(predecessor, v, u);
            cycleVisit(g, v, predecessor, disc, finish, time, vertexWeights, cycles, scores, threshold);
        }
        else if(getProperty(finish, v) == -1) //found a cycle
        {
            //std::cerr << "Found a cycle! " << std::endl;
            //traceback the cycle and add to the report list

            seqan::String<TVertexDescriptor> tmp;
            //resize(cycles, length(cycles) + 1);
            TVertexDescriptor trace = u;
            appendValue(tmp, v);
            appendValue(tmp, u);
            double score = getProperty(vertexWeights, u) + getProperty(vertexWeights, v);
            while ( (trace = getProperty(predecessor, trace)) != v && score <= threshold)
            {
                //std::cerr << "TB:" << trace << std::endl;
                score += getProperty(vertexWeights, trace);
                appendValue(tmp, trace);                
            }
            /*if(score < topScore && len < minLen) //this is the new best violated cycle
            {
                seqan::swap(cycles, tmp);
                topScore = score;
            }*/
            if (score <= threshold)
            {
                seqan::appendValue(cycles, tmp);
                seqan::appendValue(scores, score);
            }
        }
    }    
    assignProperty(finish, u, time);
    ++time;
}


//////////////////////////////////////////////////////////////////////////////

/**
.Function.depthFirstSearch:
..cat:Graph
..summary:Implements a depth-first search on a graph.
..remarks:In contrast to a breadth-first search the depth-first search is repeated from multiple sources if the graph is not connected.
Hence, depth-first search produces a depth-first forest. To ensure each vertex ends up in exactly one tree we need not just a distance but a
discovery and finishing time.
..signature:depthFirstSearch(g, predecessor, discovery, finish)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.predecessor:Out-parameter:A property map.
...remarks:Predecessor subgraph produced by the depth-first search.
..param.discovery:Out-parameter:A property map.
...remarks:The discovery time of a vertex v.
..param.finish:Out-parameter:A property map.
...remarks:The time when v's adjacency list has been fully explored.
..returns:void.
..see:Function.breadthFirstSearch
..include:seqan/graph_algorithms.h
*/
template<typename TPredecessorMap, typename TDiscoveryTimeMap, typename TFinishingTimeMap, typename TCycleString, typename TScoreString, typename TWeightMap>
void
detectCycles(TDirGraph const& g,
                   TPredecessorMap& predecessor,
                   TDiscoveryTimeMap& disc,
                   TFinishingTimeMap& finish,
                   TWeightMap & vertexWeights,
                   TCycleString & cycle,
                   TScoreString & scores,
                   double threshold)
{
    using namespace seqan;
    SEQAN_CHECKPOINT
    typedef typename Size<TDirGraph>::Type TSize;
    typedef typename Iterator<TDirGraph, EdgeIterator>::Type TEdgeIterator;
    typedef typename Iterator<TDirGraph, VertexIterator>::Type TVertexIterator;
    typedef typename VertexDescriptor<TDirGraph>::Type TVertexDescriptor;
    typedef typename Value<TPredecessorMap>::Type TPredVal;

    // Initialization
    resizeVertexMap(g,predecessor);
    resizeVertexMap(g,disc, -1);
    resizeVertexMap(g,finish, -1);
    TPredVal nilPred = getNil<TVertexDescriptor>();

    //String<bool> tokenMap;
    //seqan::resizeVertexMap(g, tokenMap);
    TVertexIterator it(g);
    for(;!atEnd(it);goNext(it))
    {
        assignProperty(predecessor, getValue(it), nilPred);
    }

    TSize time = 0;    

    goBegin(it);
    for(;!atEnd(it);goNext(it)) {
        TVertexDescriptor u = getValue(it);
        if (getProperty(disc, u) == -1) {
            cycleVisit(g, u, predecessor, disc, finish, time, vertexWeights, cycle, scores, threshold);
        }
    }

    //std::cerr << "best cycle has score: " << topScore << std::endl;
    return;// topScore;
}



//////////////////////////////////////////////////////////////////////////////

/**
.Function.depthFirstSearch:
..cat:Graph
..summary:Implements a depth-first search on a graph.
..remarks:In contrast to a breadth-first search the depth-first search is repeated from multiple sources if the graph is not connected.
Hence, depth-first search produces a depth-first forest. To ensure each vertex ends up in exactly one tree we need not just a distance but a
discovery and finishing time.
..signature:depthFirstSearch(g, predecessor, discovery, finish)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.predecessor:Out-parameter:A property map.
...remarks:Predecessor subgraph produced by the depth-first search.
..param.discovery:Out-parameter:A property map.
...remarks:The discovery time of a vertex v.
..param.finish:Out-parameter:A property map.
...remarks:The time when v's adjacency list has been fully explored.
..returns:void.
..see:Function.breadthFirstSearch
..include:seqan/graph_algorithms.h
*/
template<typename TPredecessorMap, typename TDiscoveryTimeMap, typename TFinishingTimeMap, typename TCycleString, typename TScoreString, typename TWeightMap>
void
detectCycles2(TDirGraph const& g,
                   TPredecessorMap& predecessor,
                   TDiscoveryTimeMap& disc,
                   TFinishingTimeMap& finish,
                   TWeightMap & vertexWeights,
                   TCycleString & cycle,
                   TScoreString & scores,
                   double threshold)
{
    using namespace seqan;
    SEQAN_CHECKPOINT
    typedef typename Size<TDirGraph>::Type TSize;
    typedef typename Iterator<TDirGraph, EdgeIterator>::Type TEdgeIterator;
    typedef typename Iterator<TDirGraph, VertexIterator>::Type TVertexIterator;
    typedef typename VertexDescriptor<TDirGraph>::Type TVertexDescriptor;
    typedef typename Value<TPredecessorMap>::Type TPredVal;

    // Initialization
    resizeVertexMap(g,predecessor);
    resizeVertexMap(g,disc, -1);
    resizeVertexMap(g,finish, -1);
    TPredVal nilPred = getNil<TVertexDescriptor>();

    //String<bool> tokenMap;
    //seqan::resizeVertexMap(g, tokenMap);
    TVertexIterator it(g);
    for(;!atEnd(it);goNext(it))
    {
        assignProperty(predecessor, getValue(it), nilPred);
    }



    goBegin(it);
    for(;!atEnd(it);goNext(it)) {
        TVertexIterator it2(g);
        for(;!atEnd(it2);goNext(it2))
        {
            assignProperty(predecessor, getValue(it2), nilPred);
            assignProperty(finish, getValue(it2), -1);
            assignProperty(disc, getValue(it2), -1);
        }

        //std::cerr << "from vertext: " << getValue(it) << std::endl;
        TSize time = 0;
        TVertexDescriptor u = getValue(it);
        //if (getProperty(disc, u) == -1) {
        cycleVisit(g, u, predecessor, disc, finish, time, vertexWeights, cycle, scores, threshold);
        //}
    }

    //filter duplicate cycles
    std::set<String<TVertexDescriptor> > uniques;
    TCycleString cycletmp;
    TScoreString scorestmp;

    //std::cerr << "temporarily have cycles: " << length(cycle) << std::endl;
    for (size_t i = 0; i < length(cycle); ++i)
    {
        std::sort(begin(cycle[i]), end(cycle[i]));

        if(uniques.insert(cycle[i]).second)
        {
            appendValue(cycletmp, cycle[i]);
            appendValue(scorestmp, scores[i]);
        }
    }

    cycle = cycletmp;
    scores = scorestmp;

    //std::cerr << "best cycle has score: " << topScore << std::endl;
    return;// topScore;
}



/**
.Function._unblock:
..cat:Graph
..summary:Helper function for enumerateAllCylces
..remarks:A version of an ordinary DFS based on the algorithm by Donald B. Johnson (1975)
..signature:_unblock(g, v, ccl, blocked, pred)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.blocked:In-parameter:A property map.
...remarks:A boolean property map - for every vertex this map stores a flag whether it is blocked (true) or not (false).
..param.v:In-parameter:A vertex descriptor.
...remarks:The vertex being explored.
..param.pred:In-parameter:A property map.
...remarks:Internal datastructure for the enumerateAllCycles function.
..returns:bool.
..include:seqan/graph_algorithms.h
*/
template<typename TGraph, typename TVertexDescriptor, typename TBoolListPM, typename TPredList>
void
_unblock(TGraph const& g,
         TVertexDescriptor const& v,
         TBoolListPM & blocked,
         TPredList & pred)
{
    using namespace seqan;

    typedef typename Value<TPredList>::Type TPredListElemType;
    typedef typename Iterator<TPredListElemType>::Type TPredIterator;


    assignProperty(blocked, v, false);

    TPredIterator it = begin(property(pred, v));
    TPredIterator itEnd = end(property(pred, v));

    while(it != itEnd)
    {
        erase(property(pred, v), it);

        if (getProperty(blocked,v))
        {
            _unblock(g, *it, blocked, pred);
        }
        ++it;
    }
}



/**
.Function._circuit:
..cat:Graph
..summary:Helper function for enumerateAllCylces
..remarks:A version of an ordinary DFS based on the algorithm by Donald B. Johnson (1975)
..signature:_circuit(g, ccl, blocked, v, s, stack, cycles)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.ccl:In-parameter:A property map.
...remarks:ccl A boolean property map - for every vertex this map stores whether it belongs to the strongly connected component (true) or not (false).
..param.blocked:In-parameter:A property map.
...remarks:A boolean property map - for every vertex this map stores a flag whether it is blocked (true) or not (false).
..param.v:In-parameter:A vertex descriptor.
...remarks:The vertex being explored.
..param.s:In-parameter:A vertex descriptor.
...remarks:The actual root vertex.
..param.stack:Out-parameter:A block string allowing for the stack interface.
..param.cycles:Out-parameter:A string set of vertex descriptors containing detected cycles.
..returns:bool.
..include:seqan/graph_algorithms.h
*/
template<typename TGraph, typename TVertexDescriptor, typename TBoolListPM, typename TStack, typename TCycleString, typename TVertexListPM, typename TScoreString, typename TWeightMap>
bool
_circuit(TGraph const& g,
        TWeightMap const& edgeWeights,
        TBoolListPM const& ccl,
        TBoolListPM & blocked,
        TVertexDescriptor const& v,
        TVertexDescriptor const& s,
        TStack & stack,
        TCycleString & cycles,
        TScoreString & cScores,
        TVertexListPM & pred,
        double threshold,
        double dist)
{
    using namespace seqan;
    typedef typename Iterator<TGraph, AdjacencyIterator>::Type TAdjacencyIterator;
    typedef typename Iterator<TGraph, OutEdgeIterator>::Type TOutEdgeIterator;
    std::cerr << "#cycles:" << length(cycles) << std::endl;



    bool found = false;
    appendValue(stack, v);
    assignProperty(blocked, v, true);

    TOutEdgeIterator itad(g,v);
    for(;!atEnd(itad);goNext(itad))
    {
        double dist2 = dist + getProperty(edgeWeights, *itad);
        TVertexDescriptor targ = targetVertex(itad);
        if (!getProperty(ccl, targ))
            continue; //only consider vertices in strongly connected component

        if (dist2 > threshold)
        {
            found = true;
        }
        else if (targ == s)
        {
            appendValue(cycles, stack);
            appendValue(cScores, dist2);
            found = true;
        }
        else if (getProperty(blocked, targ) == false)
        {
            found = _circuit(g, edgeWeights, ccl, blocked, targ, s, stack, cycles, cScores, pred, threshold, dist2);
        }
    }

    if (found)
    {
        _unblock(g, v, blocked, pred);
    }
    else
    {
        TAdjacencyIterator it(g,v);
        for (;!atEnd(it);goNext(it))
        {
            if (!getProperty(ccl, *it))
                continue; //only consider vertices in strongly connected component

            insert(property(pred, *it),v);
        }
    }

    resize(stack, length(stack)-1); //remove v from stack
    return found;
}



/**
.Function.depthFirstSearch:
..cat:Graph
..summary:Implements a depth-first search on a graph.
..remarks:In contrast to a breadth-first search the depth-first search is repeated from multiple sources if the graph is not connected.
Hence, depth-first search produces a depth-first forest. To ensure each vertex ends up in exactly one tree we need not just a distance but a
discovery and finishing time.
..signature:depthFirstSearch(g, predecessor, discovery, finish)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.predecessor:Out-parameter:A property map.
...remarks:Predecessor subgraph produced by the depth-first search.
..param.discovery:Out-parameter:A property map.
...remarks:The discovery time of a vertex v.
..param.finish:Out-parameter:A property map.
...remarks:The time when v's adjacency list has been fully explored.
..returns:void.
..see:Function.breadthFirstSearch
..include:seqan/graph_algorithms.h
*/
template<typename TGraph, typename TCycleString, typename TScoreString, typename TWeightMap>
void
enumerateAllCycles(TGraph const& g,
                   TCycleString & cycles,
                   TScoreString & cScores,
                   TWeightMap const& edgeWeights,
                   double threshold)
{
    using namespace seqan;
    SEQAN_CHECKPOINT

    typedef typename VertexDescriptor<TGraph>::Type TVertexDescriptor;
    typedef typename Iterator<TGraph, VertexIterator>::Type TVertexIterator;
    typedef String<TVertexDescriptor> TVertexString;
    typedef String<TVertexString> TVertexStringSet;

    String<bool> ccl, blocked;
    resizeVertexMap(g, ccl);
    resizeVertexMap(g, blocked);

    typedef Map<TVertexDescriptor> TPredMapType;
    String<TPredMapType> pred;
    resizeVertexMap(g, pred);

    String<TVertexDescriptor> stack;

    // step 1: compute strongly connected components of g
    String<unsigned int> components;
    stronglyConnectedComponents(g, components);

    unsigned int maxId = 0;
    TVertexIterator it(g);
    while (!atEnd(it))
    {
        maxId = _max(maxId, getProperty(components, *it));
        ++it;
    }

    TVertexString cc; //vertices of the actual conn. comp.

    //for each strongly connected component do
    for (unsigned int i = 0; i <= maxId; ++i)
    {        
        clear(cc);
        TVertexIterator itC(g);
        while (!atEnd(itC))
        {
            if (getProperty(components, *itC) == i)
            {
                //std::cerr << "vertex " << *itC << std::endl;
                assignProperty(ccl, *itC, true);
                appendValue(cc, *itC);
            }
            else
                assignProperty(ccl, *itC, false);

            ++itC;
        }

        // start the circuit function from every vertex in cc
        //std::cerr << "length  " << length(cc) << std::endl;
        typename Iterator<TVertexString>::Type itV = begin(cc);
        typename Iterator<TVertexString>::Type itVEnd = end(cc);

        while (itV != itVEnd)
        {            
            //std::cerr << "vertex2 " << *itV << std::endl;
            typename Iterator<TVertexString>::Type itV2 = begin(cc);
            typename Iterator<TVertexString>::Type itV2End = end(cc);
            while (itV2 != itV2End)
            {
                assignProperty(blocked, getValue(itV2), false);
                assignProperty(pred, getValue(itV2), TPredMapType());
                ++itV2;
            }

            _circuit(g, edgeWeights, ccl, blocked, *itV, *itV, stack, cycles, cScores, pred, threshold, 0.);
            assignProperty(ccl, *itV, false); //remove root from the component
            ++itV;
        }
    }
}




/**
.Function.shortestCycleV:
..cat:Graph
..summary:Implements a depth-first search on a graph.
..remarks:In contrast to a breadth-first search the depth-first search is repeated from multiple sources if the graph is not connected.
Hence, depth-first search produces a depth-first forest. To ensure each vertex ends up in exactly one tree we need not just a distance but a
discovery and finishing time.
..signature:depthFirstSearch(g, predecessor, discovery, finish)
..param.g:In-parameter:A graph.
...type:Spec.Undirected Graph
...type:Spec.Directed Graph
..param.predecessor:Out-parameter:A property map.
...remarks:Predecessor subgraph produced by the depth-first search.
..param.discovery:Out-parameter:A property map.
...remarks:The discovery time of a vertex v.
..param.finish:Out-parameter:A property map.
...remarks:The time when v's adjacency list has been fully explored.
..returns:void.
..see:Function.breadthFirstSearch
..include:seqan/graph_algorithms.h
*/
template<typename TGraph, typename TWeightMap, typename TCycleString>
double
shortestCycleV(TGraph const& g,
              TWeightMap const& vertexWeights,
              TCycleString & cycle)
{
    using namespace seqan;

    typedef typename VertexDescriptor<TGraph>::Type TVertexDescriptor;
    typedef typename Iterator<TGraph, EdgeIterator>::Type TEdgeIterator;

    TWeightMap edgeWeights;
    resizeEdgeMap(g, edgeWeights);

    TEdgeIterator it(g);
    while (!atEnd(it))
    {
        assignProperty(edgeWeights, *it, getProperty(vertexWeights, sourceVertex(g, *it)));
        ++it;
    }

    return shortestCycle(g, edgeWeights, cycle);
}

template<typename TGraph, typename TWeightMap, typename TCycleString, typename TBoundType>
double
shortestCycle(TGraph const& g,
              TWeightMap const& edgeWeights,
              TCycleString & cycle,
              TBoundType bound,
              float eps = 0.0001)
{
    using namespace seqan;

    typedef typename VertexDescriptor<TGraph>::Type TVertexDescriptor;
    typedef typename Iterator<TGraph, VertexIterator>::Type TVertexIterator;
    typedef typename Iterator<TGraph, EdgeIterator>::Type TEdgeIterator;
    typedef typename EdgeDescriptor<TGraph>::Type TEdgeDescriptor;
    typedef String<String<TEdgeDescriptor> > TEdgeList;
    typedef typename Iterator<String<TEdgeDescriptor> >::Type TInEdgeIterator;
    typedef typename Value<TWeightMap>::Type TDistVal;

    TVertexDescriptor nilpred = getNil<TVertexDescriptor>();
    TDistVal minLen = _getInfinityDistance(edgeWeights);

    //create list of incoming edges for each vertex
    TEdgeIterator itE(g);
    TEdgeList IncomingEdges;
    resizeVertexMap(g, IncomingEdges);

    while(!atEnd(itE))
    {
        appendValue(property(IncomingEdges, targetVertex(g, *itE)), *itE);
        ++itE;
    }

    //run dijkstra for each node v and get distance (v -> v)
    TVertexIterator itV(g);
    while (!atEnd(itV))
    {
        //std::cerr << "dijkstra from: " << *itV << std::endl;
        String<TVertexDescriptor> pred;
        TWeightMap dist;
        dijkstra(g, *itV, edgeWeights, pred, dist, bound);

        TVertexDescriptor minV = nilpred;
        TInEdgeIterator itIn = begin(property(IncomingEdges, *itV)), itInEnd = end(property(IncomingEdges, *itV));
        while (itIn != itInEnd)
        {
            TDistVal tmpDist = getProperty(dist, sourceVertex(g, *itIn)) + getProperty(edgeWeights, *itIn);
            if (tmpDist < minLen && tmpDist <= bound)
            {
                minLen = tmpDist;
                minV = sourceVertex(g, *itIn);
            }
            ++itIn;
        }

        if (minV != nilpred) //new shortest cycle found
        {
            clear(cycle);
            TVertexDescriptor itBt = minV;

            while (itBt != *itV)
            {
                std::cerr << itBt << "  ";
                appendValue(cycle, itBt);
                itBt = getProperty(pred, itBt);
            }
            appendValue(cycle, *itV);
            std::cerr << *itV << std::endl;
            
            if (minLen < eps) //can't get much better. Stop here
                return minLen;
        }
        ++itV;
    }
    return minLen;
}







#endif // CYCLEDETECT_H
