// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef DUPLICATIONEVENTHOLDER_H
#define DUPLICATIONEVENTHOLDER_H

#include <DuplicationEventHolderBase.h>

struct SparsePairGraph;

struct DuplicationEventHolder : public DuplicationEventHolderBase
{
    friend struct SparsePairGraph;
    friend struct PairedDuplicationEventHolder3;
    friend void testMixedCliques(); //get rid of this by accessor methods
    friend void testDuplications();
    
    using DuplicationEventHolderBase::DuplicationEvent;

    struct ViolatedDuplicationCycle
    {
        TSizeArray dupIds;
        size_t cycleLen;
        double score;

        bool operator< (const ViolatedDuplicationCycle & rhs) const
        {
            if (cycleLen - score != rhs.cycleLen - rhs.score)
                return (cycleLen - score > rhs.cycleLen - rhs.score);
            else
                return cycleLen < rhs.cycleLen;
                //return dupIds.size() > rhs.dupIds.size();
        }
    };
                

    typedef seqan::Graph<seqan::Directed<> > TGraph;
    typedef seqan::VertexDescriptor<TGraph>::Type TVertexDescriptor;
    typedef std::vector<TVertexDescriptor> TVertexList;
                
    typedef std::vector<std::vector<size_t> > TDuplicationMap;
    typedef std::vector<std::vector<size_t> > TIndexMatrix;

    typedef std::vector<std::vector<std::vector<size_t> > > TDuplicationIntervalMap;
    typedef std::vector<std::vector<double> > TScoreMatrix;

    typedef std::vector<DuplicationEvent> TDuplicationEventList;
    typedef std::vector<DuplicationEvent>::const_iterator TDuplicationEventConstIter;

    typedef std::vector<ViolatedDuplicationCycle> TviolatedDupList;

    //MEMBERS

    ///find all possible duplications using lcp tab scan
    void createDuplicationEventsLCP(const GeneString & str, bool inversions = false);

    ///obtain violated duplication cycles for a given set of activated duplications
    void getDuplicationCycle(const IloNumArray & scores, TviolatedDupList &cycles, double threshold = INFINITY, bool all = true)
    {
        return _getDupCycle(scores, cycles, threshold, all, false);
    }

    ///obtain lifted violated duplication cycles for a given set of activated duplications
    void getLiftedDuplicationCycle(const IloNumArray & scores, TviolatedDupList & cycles, double threshold = INFINITY, bool all = true)
    {
        return _getDupCycle(scores, cycles, threshold, all, true);
    }    
    

protected:
    
    TIndexMatrix _duplicationCliques;                                                                                                                                                                                                                                     
    TIndexMatrix _dupToCliques;                                                                                                                                                                                                                                           
    TIndexMatrix _cliquesAdjList;                                                                                                                                                                                                                                         
    //EventMode _mode;

    //void _reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos, bool)
    //{
    //    return DuplicationEventHolderBase::_reportEvent(startPos, endPos, len, sa, catPos, true, true);
    //};
    


    ///helper creates the duplication graph. Only duplications with non-zero vScoreIlo entry are added to the graph
    void _createDuplicationGraph(TGraph & graph,
                                 TVertexList & vertices,
                                 TPosPropertyMap & verticesInv,
                                 TGraphWeightPropertyMap & vScores,
                                 const IloNumArray & vScoresIlo);    


    void _createLiftedDuplicationGraph(TGraph & graph,
                             TVertexList & vertices,
                             TPosPropertyMap & verticesInv,
                             TGraphWeightPropertyMap & vScores,
                             const IloNumArray & vScoresIlo);

    void _createDuplicationCliques();

    ///compute duplication cycles for lifted or non lifted
    void _getDupCycle(const IloNumArray & scores, TviolatedDupList &cycles, double threshold = INFINITY, bool all = true, bool lifted = false);
};




/*
//Implementation of function templates
template<typename SAString>
inline void DuplicationEventHolder::_reportDuplicates(size_t startPos, size_t endPos, size_t len, const SAString &sa)
{
    //std::cerr << " report duplicates " << startPos << " -- " << endPos << "  " << len << std::endl;
    for (size_t i = startPos; i < endPos; ++i)
    {
        for (size_t j = i + 1; j <= endPos; ++j)
        {
            size_t l = seqan::_min(sa[i],sa[j]);
            size_t r = seqan::_max(sa[i],sa[j]);
            //test if they are non-overlapping
            if(r - l >= len)
            {
                size_t len2 = len - 1;
                duplications.push_back(DuplicationEvent(l, l + len2, r, r + len2));
                duplications.push_back(DuplicationEvent(r, r + len2, l, l + len2));

                for(size_t k = 0; k <= len2; ++k)
                {
                    //std::cout << " try to access: " << sa[j] + k << std::endl;
                    dupMap[r + k].push_back(seqan::length(duplications)-2);
                    dupMap[l + k].push_back(seqan::length(duplications)-1);
                }

                dupInterMap[r][r + len2].push_back(seqan::length(duplications)-2);
                dupInterMap[l][l + len2].push_back(seqan::length(duplications)-1);
            }
        }
    }
}
*/



#endif // DUPLICATIONEVENTHOLDER_H
