// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef DUPLICATIONEVENTHOLDERBASE_H
#define DUPLICATIONEVENTHOLDERBASE_H


//#include <CycleDetect.h>
#include <EventHolderBase.h>


struct SparsePairGraph;

struct DuplicationEventHolderBase : public EventHolderBase
{    
    //enum EventMode{DUPNONINV, DUPALL};
    
    ///Nested type: A single Duplication Event
    struct DuplicationEvent
    {
        size_t sourceBegin, sourceEnd, targetBegin, targetEnd;
        
        DuplicationEvent(){};
        DuplicationEvent(size_t sb, size_t se, size_t tb, size_t te):
        sourceBegin(sb), sourceEnd(se), targetBegin(tb), targetEnd(te){}
    };
    
    typedef std::vector<std::vector<size_t> > TDuplicationMap;
    typedef std::vector<std::vector<size_t> > TIndexMatrix;
    
    typedef std::vector<std::vector<std::vector<size_t> > > TDuplicationIntervalMap;
    typedef std::vector<std::vector<double> > TScoreMatrix;
    
    typedef std::vector<DuplicationEvent> TDuplicationEventList;
    typedef std::vector<DuplicationEvent>::const_iterator TDuplicationEventConstIter;
    
    
    //MEMBERS
    
    ///find all possible duplications using lcp tab scan
    virtual void createDuplicationEventsLCP(const GeneString & str, bool inversions = false) = 0;    
    
    bool hasDup(size_t i, size_t j){return !dupInterMap[i][j].empty();}
    
    void DPCompute(const IloNumArray & scores);
    
    ///List of all duplication events for the given string
    TDuplicationEventList duplications;    
    
    ///map from string positions to covering duplication events
    TDuplicationMap dupMap;
    
    //map containing for each pair of positions i,j the duplications with target site [i,j]
    TDuplicationIntervalMap dupInterMap;    
    //piV(i,j) is the summed score of all duplications whose target site start at position i and do not end before j
    TScoreMatrix piV;
    //pirV(i,j) is the summed score of all duplications whose target site end at position j and do not start after i
    TScoreMatrix pirV;
    //sigV(i,j) is the summed score of all duplications whose target sites include the interval [i,j]
    TScoreMatrix sigV;
    ///The sigma(i,j) map containing the duplication indices contributing for each sigma(i,j)
    TDuplicationIntervalMap sigI;
    ///The pi(i,j) map containing the duplication indices contributing for each pi(i,j)
    TDuplicationIntervalMap piI;    
    
protected:
    
    //EventMode _mode;
    
    //template<typename SAString>
    //virtual void _reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos, bool inverted);    
    virtual void _reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos);    
    
    //initialize sigI and piI arrays after duplication events are created
    void _init();
    
    bool _inversions;
};


//new implementation which allows to identify possible inversions or inverted duplications
//to find possible inversion concatenate the genome with the inversion of the second genome
//to find possible inverted duplications concat the genome with its own inversion
inline void DuplicationEventHolderBase::_reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos)
{
    size_t len2 = len - 1;
    size_t lenG2 = seqan::length(sa) - catPos - 1;
    
    for (size_t i = startPos; i < endPos; ++i)
    {
        for (size_t j = i + 1; j <= endPos; ++j)
        {
            size_t l = seqan::_min(sa[i],sa[j]);
            size_t r = seqan::_max(sa[i],sa[j]);
            
            if (l > catPos)
                continue; //skip if both in inverted concat string
            
            if (!_inversions && (r - l < len || r > catPos))
                continue; //skip overlapping or right end in concat part of string
            
            if (_inversions && r - l < len && r < catPos)
                continue; //skip overlapping (when both on same side of concat string)
            
            if (r > catPos)
            {
                if (_inversions)
                    r = lenG2 - (r - catPos) - len2; // for inverted duplication perform index correction
                else
                    r = r - catPos - 1;
            }
            
            duplications.push_back(DuplicationEvent(l, l + len2, r, r + len2));
            duplications.push_back(DuplicationEvent(r, r + len2, l, l + len2));
            
            for(size_t k = 0; k <= len2; ++k)
            {
                dupMap[r + k].push_back(seqan::length(duplications)-2);
                dupMap[l + k].push_back(seqan::length(duplications)-1);
            }
            dupInterMap[r][r + len2].push_back(seqan::length(duplications)-2);
            dupInterMap[l][l + len2].push_back(seqan::length(duplications)-1);
        }
    }
}



#endif 


