// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef FUNCTIONS_H
#define FUNCTIONS_H

#include <Types.h>
#include <fstream>
#include <string>


struct TGenome
{
    std::string id;
    GeneString geneStr;
};

typedef std::map<std::string, int> TNameStore;
typedef std::vector<TGenome> TGenomeList;

inline void parseGenomes(const std::string & GenFile, TNameStore & namesStore, TGenomeList & genomes, bool inversions = false)
{
    int newId = 1;
    std::ifstream infile(GenFile.c_str());
    if (!infile.is_open())
    {
        std::cerr << "Cannot open Genome File: " << GenFile << "  --terminating" << std::endl;
    }
    
    //set the first 1000 genes in case of numeric ids no transformation is required
    for (size_t i = 1; i < 1000; ++i)
    {
        std::stringstream strstr;
        strstr << i;
        namesStore[strstr.str()] = i;
        ++newId;
    }

    std::string line;
    while (getline(infile, line))
    {
        if (line[0] == '#')
        {
            genomes.push_back(TGenome());
            genomes.back().id = line.substr(2);
            //remove trailing whitespaces or other line endings
            genomes.back().id.erase(genomes.back().id.find_last_not_of(" \n\r\t")+1);
        }
        else
        {
            size_t pos1 = 0, pos2 = 0;

            while(pos2 != std::string::npos)
            {
                pos2 = line.find_first_of(",;",pos1);

                if (pos2 == std::string::npos && pos1 == 0)
                    break;

                std::string gene = line.substr(pos1, pos2 - pos1);
                //std::cerr << seqan::length(genomes.back().geneStr) <<": " << gene << std::endl;
                pos1 = pos2 + 1;

                TNameStore::iterator finder;
                int sign = 1;
                if (inversions && gene[0] == '-')
                {
                    gene = gene.substr(1);
                    sign = -1;
                }
                if ((finder = namesStore.find(gene)) != namesStore.end())
                {
                    //std::cerr << "appending " << sign * finder->second << std::endl;
                    seqan::appendValue(genomes.back().geneStr, sign * finder->second);
                }
                else
                {                    
                    namesStore[gene] = sign * newId; //get strand info
                    seqan::appendValue(genomes.back().geneStr, namesStore[gene]);
                    ++newId;
                    //std::cerr << "appending new gene" << namesStore[gene] << std::endl;
                }
            }
        }
    }
}

void modifyGenome(GeneString & gen, unsigned numMoves, seqan::Rng<seqan::MersenneTwister> &rng)
{
    
    //sample duplication length
    //static seqan::Rng<seqan::MersenneTwister> rng(0);
    seqan::Pdf<seqan::Normal> normal(5, 2);
    seqan::Pdf<seqan::Uniform<int> > uniformInt(0,1);
    
    for (size_t i = 0; i < numMoves; ++i)
    {
        //cerr << "gen: " << gen << endl;
        //sample operation
        int op = pickRandomNumber(rng, uniformInt);
        //int op = rand() % 2;
        
        if (op == 1) //loss
        {
            //sample position for loss
            seqan::Pdf<seqan::Uniform<int> > uniformInt1(0, seqan::length(gen)-1);
            unsigned pos = pickRandomNumber(rng, uniformInt1);
            //unsigned pos = rand() % seqan::length(gen);
            seqan::erase(gen, pos);
        }
        
        if (op == 0) //duplication
        {
            int dlen;
            do
            {
                dlen = pickRandomNumber(rng, normal);
                std::cerr << "dlen: " << dlen << std::endl;
            }
            while (dlen <= 0 || dlen >= seqan::length(gen));
            
            //sample start
            unsigned posS;
            seqan::Pdf<seqan::Uniform<int> > uniformInt2(0, seqan::length(gen) - dlen);
            posS = pickRandomNumber(rng, uniformInt2);
            //posS = rand() % (seqan::length(gen) - dlen + 1);
            unsigned posT;
            do
            {
                seqan::Pdf<seqan::Uniform<int> > uniformInt3(0, seqan::length(gen) - 1);
                posT = pickRandomNumber(rng, uniformInt3);
                //posT = rand() % seqan::length(gen);
            }
            while (posT <= posS + dlen - 1 && posT >= posS);

            GeneString substr(seqan::infix(gen, posS, posS + dlen));
//            seqan::insert(gen, posT, seqan::infix(gen, posS, posS + dlen));
            seqan::insert(gen, posT, substr);
//            seqan::resize(substr, dlen);
//            for (size_t j = 0; j < dlen; ++j)
//            {
//                substr[j] = gen[posS+j];
//            }
//            for (size_t j = 0; j < dlen; ++j)
//            {
//                seqan::insertValue(gen, posT + j, substr[j]);
//            }
        }
    }
}


void generateInstance(unsigned len,
                      unsigned moves,
                      unsigned alphaSize,
                      GeneString & ancestor,
                      GeneString & CA,
                      GeneString & CB,
                      seqan::Rng<seqan::MersenneTwister> &rng)
{
    GeneString root;
    seqan::resize(root, len);
    //sample the root genome R
    seqan::Pdf<seqan::Uniform<int> > uniformInt(1, alphaSize);
    
    for (size_t i = 0; i < len; ++i)
    {
        root[i] = pickRandomNumber(rng, uniformInt);
        //root[i] = rand() % alphaSize;
    }
    
    //create ancestor by applying "moves" moves to the root
    ancestor = root;
    modifyGenome(ancestor, moves, rng);
    
    //generate the two extant genomes
    CA = ancestor;
    CB = ancestor;
    
    modifyGenome(CA, moves, rng);
    modifyGenome(CB, moves, rng);
    
    std::cerr << "anc: ";
    for (size_t i = 0; i < seqan::length(ancestor); ++i)
        std::cerr << ancestor[i]<<" ";
    std::cerr << std::endl;
    
    std::cerr << "CA: ";
    for (size_t i = 0; i < seqan::length(CA); ++i)
        std::cerr << CA[i]<<" ";
    std::cerr << std::endl;
    
    std::cerr << "CB: ";
    for (size_t i = 0; i < seqan::length(CB); ++i)
        std::cerr << CB[i]<<" ";
    std::cerr << std::endl;    
}

void printGenome(const GeneString & str , std::ostream & stream = std::cerr)
{
    for (size_t i = 0; i + 1 < seqan::length(str); ++i)
    {
        stream << str[i] <<',';
    }
    stream << seqan::back(str) << std::endl;
}





#endif // FUNCTIONS_H
