// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef GENETREE_H
#define GENETREE_H

#include <seqan/graph_algorithms.h>
#include <Types.h>
#include <Functions.h>
#include <IlpMedian.h>

struct GeneTree
{
    typedef seqan::Graph<seqan::Tree<> > Ttree;
    typedef seqan::VertexDescriptor<Ttree>::Type TVertexDescriptor;
    typedef seqan::Iterator<Ttree, seqan::DfsPreorder>::Type TPreorderIt;
    typedef seqan::Iterator<Ttree, seqan::OutEdgeIterator>::Type TEdgeIt;
    
protected:
    Ttree tree;
    seqan::String<std::string> _nodeNames;
    seqan::String<GeneString> _groundTruth;
    seqan::String<std::vector<GeneString> > _genomesPredicted;
    seqan::String<size_t> _nodeDepth;
    seqan::String<double> _edgeCost;
    
    size_t _iteration;
    
    TNameStore _geneNameStore;
    
    std::string _recursiveWrite(TVertexDescriptor vertex);
    
public:

    void parseTree(const std::string & text, const std::string & fileGenomes);
    void writeTree(const std::string & text, const std::string & fileGenomes);
    
    void evaluatePredictions(const IlpOptions & opts);
    int reoptimize(bool init, const IlpOptions & opts);
    
    void simulate(size_t generations, 
                  size_t length, 
                  size_t moves, 
                  size_t alpha,
                  seqan::Rng<seqan::MersenneTwister> &rng);
    
    void print();
};


void inline GeneTree::simulate(size_t generations, 
                               size_t length,
                               size_t moves, 
                               size_t alpha, 
                               seqan::Rng<seqan::MersenneTwister> &rng)
{
    std::vector<TVertexDescriptor> currentGen, nextGen;
    
    
    //generate the root genome
    GeneString grand, P1, P2;    
    generateInstance(length, moves, alpha, grand, P1, P2, rng);
    
    currentGen.push_back(seqan::addVertex(tree)); //root
    for (size_t i = 0; i < generations; ++i)
    {
        nextGen.clear();
        for (size_t j = 0; j < currentGen.size(); ++j)
        {
            nextGen.push_back(seqan::addChild(tree, currentGen[j]));
            nextGen.push_back(seqan::addChild(tree, currentGen[j]));
        }
        currentGen.swap(nextGen);
    }
    
    //simulate the genomes for all vertices.
    seqan::resizeVertexMap(tree, _groundTruth);
    seqan::resizeVertexMap(tree, _genomesPredicted);
    seqan::resizeVertexMap(tree, _nodeDepth);
    seqan::resizeVertexMap(tree, _nodeNames);
    seqan::resizeEdgeMap(tree, _edgeCost);
    
    TVertexDescriptor root = seqan::getRoot(tree);
    seqan::assignProperty(_groundTruth, root, grand);
    std::stringstream strroot;
    strroot << root;
    seqan::assignProperty(_nodeNames, root, strroot.str());
    
    //traverse tree and construct the genomes
    TPreorderIt preIt(tree, root);
    seqan::assignProperty(_nodeDepth, *preIt, 0);
    goNext(preIt); //skip root
    for(;!atEnd(preIt);goNext(preIt)) 
    {
        TVertexDescriptor parent = seqan::parentVertex(tree, *preIt);
        seqan::assignProperty(_groundTruth, *preIt, seqan::getProperty(_groundTruth, parent));
        seqan::assignProperty(_nodeDepth, *preIt, seqan::getProperty(_nodeDepth, parent) + 1);
        
        modifyGenome(seqan::property(_groundTruth, *preIt), moves, rng);
        std::cerr << "LEN: " << seqan::length(seqan::property(_groundTruth, *preIt)) << " " << seqan::length(seqan::getProperty(_groundTruth, parent)) << std::endl;
        
        std::stringstream strstr;
        strstr << *preIt;
        seqan::assignProperty(_nodeNames, *preIt, strstr.str());
    }
    
    for(seqan::goBegin(preIt); !atEnd(preIt); goNext(preIt))
    {
        std::cerr << "Vertex " << *preIt << ": ";
        printGenome(seqan::getProperty(_groundTruth, *preIt));
    }
    std::cerr << tree << std::endl;
}


int inline GeneTree::reoptimize(bool init, const IlpOptions & opts)
{
    const bool acceptWorse = true;
    typedef seqan::Iterator<Ttree, seqan::DfsPreorder>::Type TPreorderIt;
    typedef seqan::Iterator<Ttree, seqan::OutEdgeIterator>::Type TOutEdgeIter;
    typedef seqan::Iterator<Ttree, seqan::EdgeIterator>::Type TEdgeIter;
    typedef seqan::EdgeDescriptor<Ttree>::Type TtreeEdgeDescriptor;

    seqan::String<seqan::String<TVertexDescriptor> > nodesByDepth;
    
    if (init)
        _iteration = 1;
    else
        ++_iteration;
    
    
    TPreorderIt preIt(tree, seqan::getRoot(tree));
    
    
    seqan::resize(nodesByDepth, 1);
    seqan::appendValue(nodesByDepth[0], *preIt);
    seqan::goNext(preIt);
    
    for(;!atEnd(preIt);goNext(preIt)) 
    {
        if (seqan::isLeaf(tree, *preIt))
            continue;

        size_t d = seqan::getProperty(_nodeDepth, *preIt);
        
        if (d >= seqan::length(nodesByDepth))
            seqan::resize(nodesByDepth, d+1);
        
        seqan::appendValue(nodesByDepth[d], *preIt);
    }
    
    double totalCost = 0;
    for (int d = seqan::length(nodesByDepth)-1; d >= 0; --d)
    {
        //work through all nodes at the same depth of the tree in parallel
        #pragma omp parallel for schedule(dynamic)
        for (size_t i = 0 ; i < seqan::length(nodesByDepth[d]); ++i)
        {   
            TVertexDescriptor & v = nodesByDepth[d][i];
            TOutEdgeIter outIt(tree, v);
            
            std::vector<GeneString> genomesTmp(2);
            std::vector<TtreeEdgeDescriptor> edges;
            for(size_t i = 0;!atEnd(outIt);goNext(outIt), ++i) 
            {
                edges.push_back(*outIt);
                if (i > 1)
                {
                    std::cerr << "ERROR ONLY BINARY TREES SUPPORTED - EXIT" << std::endl;
                    exit(-1);
                }        
                TVertexDescriptor child = seqan::targetVertex(tree, *outIt);
                if (seqan::isLeaf(tree, child))             
                    genomesTmp[i] = seqan::getProperty(_groundTruth, child);
                else
                    genomesTmp[i] = seqan::getProperty(_genomesPredicted, child).back();
            }
            
            TAlignment alignTmp;
            std::string vName = seqan::getProperty(_nodeNames, v);
            if (init || seqan::isRoot(tree, v)) 
            {

#ifdef DEBUGDLC
                #pragma omp critical
                {
                    std::cerr << "initialize Node " <<  vName << std::endl;
                    std::cerr << "Genome A: ";
                    printGenome(genomesTmp[0]);
                    std::cerr << "Genome B: ";
                    printGenome(genomesTmp[1]);
                }
#endif
                
                double cost = MedianILPThreeLeaf(genomesTmp, alignTmp, opts);
                totalCost += cost;
                seqan::property(_genomesPredicted, v).resize(_iteration);
                getMedianGenome(alignTmp, seqan::property(_genomesPredicted, v).back());
                std::vector<double> dists;
                getDistancesToMedian(alignTmp, dists, false);
                seqan::assignProperty(_edgeCost, edges[0], dists[0]);
                seqan::assignProperty(_edgeCost, edges[1], dists[1]);

#ifdef DEBUGDLC
                #pragma omp critical
                {
                    std::cerr << "single costs for " << vName << "(" << seqan::targetVertex(tree, edges[0]) << ", " << seqan::targetVertex(tree, edges[1]) << "): ";
                    std::cerr << "single costs for " << vName << ": " <<dists[0] << " - " << dists[1] << std::endl;
                    std::cerr << "MEDIAN: ";
                    printGenome(seqan::getProperty(_genomesPredicted, v).back());

                    if (std::abs(dists[0] + dists[1] - cost) >= 0.01)
                    {
                        std::cerr << "ATTENTION DIFFERING COSTS -- SOME ERROR -- EXIT" << std::endl;
                        std::cerr << (int)(dists[0] + dists[1]) << " != " << (int)cost << std::endl;
                        exit(-1);
                    }
                }
#endif
            }
            else
            {
                TVertexDescriptor parent = seqan::parentVertex(tree, v);
                genomesTmp.insert(genomesTmp.begin(), seqan::getProperty(_genomesPredicted, parent).back());

#ifdef DEBUGDLC
                #pragma omp critical
                {
                    std::cerr << "reoptimize Node " << vName << std::endl;
                    std::cerr << "Genome X: ";
                    printGenome(genomesTmp[0]);
                    std::cerr << "Genome A: ";
                    printGenome(genomesTmp[1]);
                    std::cerr << "Genome B: ";
                    printGenome(genomesTmp[2]);
                }
#endif
                
                double cost = MedianILPThreeLeaf(genomesTmp, alignTmp, opts);
                totalCost += cost;
                
                //compare new cost to old cost
                edges.insert(edges.begin(), seqan::findEdge(tree, parent, v));

                std::vector<double> dists(3);
                dists[0] = seqan::getProperty(_edgeCost, edges[0]);
                dists[1] = seqan::getProperty(_edgeCost, edges[1]);
                dists[2] = seqan::getProperty(_edgeCost, edges[2]);
                
                seqan::property(_genomesPredicted, v).resize(_iteration);
                std::cerr << "Cost compare: " << cost << " " << dists[0] << " + " << dists[1] << " + " << dists[2] << std::endl;
                if (cost  <= dists[0] + dists[1] + dists[2] + 0.01 || acceptWorse)
                {
                    getMedianGenome(alignTmp, seqan::property(_genomesPredicted, v).back());
                    dists.clear();
                    getDistancesToMedian(alignTmp, dists, true);
                    
                    seqan::assignProperty(_edgeCost, edges[0], dists[0]);
                    seqan::assignProperty(_edgeCost, edges[1], dists[1]);            
                    seqan::assignProperty(_edgeCost, edges[2], dists[2]);
                }
                else
                {
#ifdef DEBUGDLC
                    std::cerr << "No Improvement for node " << vName << " -- keep old median" << std::endl;
#endif
                    //copy the previous median
                    seqan::property(_genomesPredicted, v).back() = seqan::property(_genomesPredicted, v)[_iteration - 2];
                }
                
#ifdef DEBUGDLC
                #pragma omp critical
                {
                    std::cerr << "single costs for " << vName << "(" << seqan::sourceVertex(tree, edges[0]) << ", " << seqan::targetVertex(tree, edges[1]) <<
                    ", " << seqan::targetVertex(tree, edges[2]) << "): ";
                    
                    std::cerr << dists[0] << " - " << dists[1] << " - " << dists[2] << std::endl;
                    std::cerr << "MEDIAN: ";
                    printGenome(seqan::getProperty(_genomesPredicted, v).back());
                    
                    if (std::abs(dists[0] + dists[1] + dists[2] - cost) >= 0.01)
                    {
                        std::cerr << "ATTENTION DIFFERING COSTS -- SOME ERROR -- EXIT" << std::endl;
                        std::cerr << (int)(dists[0] + dists[1] + dists[2]) << " != " << (int)cost << std::endl;
                        exit(1);
                    }
                }
#endif
            }
        }
    }
    
    //Sum up the costs along all edges
    TEdgeIter edgeIt(tree);
    double total_edgeCost = 0;
    for (; !atEnd(edgeIt); goNext(edgeIt))
    {
        total_edgeCost += seqan::property(_edgeCost, *edgeIt);
#ifdef DEBUGDLC
        std::cerr << seqan::sourceVertex(tree, *edgeIt) << " --> " << seqan::targetVertex(tree, *edgeIt) << " :: " << seqan::property(_edgeCost, *edgeIt) << std::endl;
#endif
    }
    
    return total_edgeCost;
}


void inline GeneTree::evaluatePredictions(const IlpOptions & opts)
{
    typedef seqan::Iterator<Ttree, seqan::DfsPreorder>::Type TPreorderIt;
    
    TPreorderIt preIt(tree, seqan::getRoot(tree));
    double totalDist = 0;
    size_t totalNodes = 0;
    std::map<size_t, size_t> distInLevel;
    std::map<size_t, size_t> nodesInLevel;
    
    for(;!atEnd(preIt);goNext(preIt)) 
    {
        if (seqan::isLeaf(tree, *preIt))
            continue;
        
        std::vector<GeneString> genomesTmp(2);
        genomesTmp[0] = seqan::getProperty(_groundTruth, *preIt);
        genomesTmp[1] = seqan::getProperty(_genomesPredicted, *preIt).back();
        
        TAlignment align;
        double distance = MedianILPThreeLeaf(genomesTmp, align, opts);
        totalDist += distance;
        ++totalNodes;
        size_t d = seqan::getProperty(_nodeDepth, *preIt);
        distInLevel[d] += distance;
        ++nodesInLevel[d];
        
        std::cerr << "Evaluated Distance from Ground truth for Node " << *preIt << " = " << distance << std::endl;
        
        std::cerr << "GROUND TRUTH " << *preIt << ": ";
        for (size_t i = 0; i < seqan::length(genomesTmp[0]); ++i)
        {
            std::cerr << genomesTmp[0][i] << " ";
        }
        std::cerr << std::endl;
        std::cerr << "MEDIAN: ";
        for (size_t i = 0; i < seqan::length(genomesTmp[1]); ++i)
        {
            std::cerr << genomesTmp[1][i] << " ";
        }
        std::cerr << std::endl;
    }
    
    std::cerr << "Total Distance from Ground truth: " << totalDist << std::endl;
    std::cerr << "Average Distance from Ground truth: " << totalDist/totalNodes << std::endl;
    
    for(std::map<size_t, size_t>::const_iterator it = nodesInLevel.begin(); it != nodesInLevel.end(); ++it)
    {
        std::cerr << "Average Distance from Ground in Level " << it->first << " : " << (double)distInLevel[it->first]/it->second << std::endl;
    }    
}


void inline GeneTree::parseTree(const std::string & fileTree, const std::string & fileGenomes)
{
    std::ifstream infile(fileTree.c_str());
    std::string text;
    
    //get the genome sequences
    TGenomeList genomesTmp;
    parseGenomes(fileGenomes, _geneNameStore, genomesTmp);
    
    std::map<std::string, GeneString> genomes;
    for (size_t i = 0; i < genomesTmp.size(); ++i)
    {
        genomes[genomesTmp[i].id] = genomesTmp[i].geneStr;
    }
    
    
    //get the tree
    if(infile.is_open())
    {
        std::getline(infile, text);
        infile.close();
    }
    else
    {
        std::cerr << "Could not open File " << fileTree << std::endl;
        std::cerr << "Aborting" << std::endl;
        exit(1);
    }
    
    typedef seqan::Iterator<Ttree, seqan::VertexIterator>::Type TVIter;
    seqan::clear(tree);
    TVertexDescriptor root = seqan::addVertex(tree);
    std::map<TVertexDescriptor, std::string> keys;
    
    //perform some validity checks
    if (text.find(':') != std::string::npos)
    {
        std::cerr << "detected \":\" symbol -- no branch lengths are supported. Please remove and try again!" << std::endl;
        exit(1);
    }
    if (text[text.length()-1] != ';')
    {
        std::cerr << "Tree not properly terminated by \";\". Please check and try again!" << std::endl;
        exit(1);
    }
    
    //pop terminating ';'
    text.erase(text.length()-1);
    
    std::string token("(),");
    
    TVertexDescriptor v = root;
    
    //check for root node identifier
    size_t lp = text.find_last_of(')');
    if(lp != text.length()-1)
    {
        std::string key = text.substr(lp+1);
        keys[root] = key;
    }
    else
    {
        keys[root] = "root";
    }
    
    size_t lastPos = 0, pos = text.find_first_of(token, 0);
    
    while (pos != std::string::npos)
    {
        if (text[pos] == '(')
        {
            //create new child vertex and go there
            v = seqan::addChild(tree, v);
        }
        if (text[pos] == ',')
        {
            //parse entry until ',' and create sibling vertex
            if (lastPos != pos)
            {
                std::string key = text.substr(lastPos,  pos - lastPos);
                keys[v] = key;
                //std::cerr << key << std::endl;
            }
            //create sibling and go there
            v = seqan::addChild(tree, seqan::parentVertex(tree, v));
        }
        if (text[pos] == ')')
        {
            //add entry to current node and go up
            if (lastPos != pos)
            {
                std::string key = text.substr(lastPos,  pos - lastPos);
                keys[v] = key;
            }
            v = seqan::parentVertex(tree, v);
            
            //check for inner vertex identifier
            size_t posTmp = text.find_first_of(token, pos+1);
            if (posTmp > pos+1 && posTmp != std::string::npos)
            {
                std::string key = text.substr(pos+1,  posTmp - pos - 1);
                keys[v] = key;
                pos = posTmp-1;
            }
        }
        lastPos = pos + 1;
        pos = text.find_first_of(token, lastPos);
    }
    
    seqan::resizeVertexMap(tree, _nodeNames);
    seqan::resizeVertexMap(tree, _groundTruth);
    seqan::resizeVertexMap(tree, _genomesPredicted);
    seqan::resizeVertexMap(tree, _nodeDepth);
    
    seqan::resizeEdgeMap(tree, _edgeCost);

    TPreorderIt it(tree, seqan::getRoot(tree));
    for (; !atEnd(it); goNext(it))
    {
        if (seqan::isRoot(tree, *it))
            seqan::assignProperty(_nodeDepth, *it, 0);
        else
        {
            seqan::assignProperty(_nodeDepth, *it, seqan::getProperty(_nodeDepth, seqan::parentVertex(tree, *it)) + 1);
        }
        
        if (seqan::isLeaf(tree, *it))
        {
            if (!keys.count(*it))
            {
                std::cerr << "Missing node id for leaf -- Terminating" << std::endl;
                exit(1);
            }
            else if (keys.count(*it) && !genomes.count(keys[*it]))
            {
                std::cerr << "Missing genome for leaf " << keys[*it] << " -- Terminating" << std::endl;
                exit(1);
            }
            else
            {
                seqan::assignProperty(_nodeNames, *it, keys[*it]);
                seqan::assignProperty(_groundTruth, *it, genomes[keys[*it]]);
            }
        }
        else if (keys.count(*it)) //if no leaf but identifier given
        {
            seqan::assignProperty(_nodeNames, *it, keys[*it]);
            
            if (!genomes.count(keys[*it])) // genome given also for internal node (for testing reasons only)
                seqan::assignProperty(_groundTruth, *it, genomes[keys[*it]]);
        }
    }
    //do a second run to assign internal node names if not in input file
    seqan::goBegin(it);
    for (; !atEnd(it); goNext(it))
    {
        if (seqan::outDegree(tree, *it) > 2)
        {
            std::cerr << "ERROR ONLY BINARY TREES SUPPORTED - EXIT" << std::endl;
            exit(1);
        }
        
        if (seqan::empty(seqan::getProperty(_nodeNames, *it)))
        {
            std::stringstream tmps;
            TEdgeIt edgeIt(tree, *it);
            for (size_t c = 0; !seqan::atEnd(edgeIt); seqan::goNext(edgeIt), ++c)
            {
                TVertexDescriptor child = seqan::targetVertex(tree, *edgeIt);
                if(c)
                    tmps << "_";
                tmps << seqan::getProperty(_nodeNames, child);
            }
            
            seqan::assignProperty(_nodeNames, *it, tmps.str());
        }
    }    
    //print();
}

void inline GeneTree::print()
{
    TPreorderIt preIt(tree, seqan::getRoot(tree));

    for(;!atEnd(preIt); goNext(preIt))
    {
        std::cerr << seqan::getProperty(_nodeNames, *preIt) << " -> ";
        TEdgeIt eIt(tree, *preIt);
        for(; !atEnd(eIt); goNext(eIt))
        {
            std::cerr << seqan::getProperty(_nodeNames, *eIt) << ',';
        }
        std::cerr << std::endl;
    }
}


void GeneTree::writeTree(const std::string & fileTree, const std::string & fileGenomes)
{
    std::string treeStr = _recursiveWrite(seqan::getRoot(tree)) + ";";
    
    std::ofstream treeFStr(fileTree.c_str());
    
    if (!treeFStr.is_open())
    {
        std::cerr << "Could not open output File: " << fileTree << std::endl;
        exit(1);
    }
    
    treeFStr << treeStr << std::endl;
    treeFStr.close();
    
    std::ofstream genFStr(fileGenomes.c_str());
    
    if (!genFStr.is_open())
    {
        std::cerr << "Could not open output File: " << fileGenomes << std::endl;
        exit(1);
    }

    //genereate inverse gene namestore
    std::map<int, std::string> invGeneNameStore;
    TNameStore::const_iterator nIt;
    for (nIt = _geneNameStore.begin(); nIt != _geneNameStore.end(); ++ nIt)
    {
        invGeneNameStore[nIt->second] = nIt->first;
    }
    
    TPreorderIt preIt(tree, seqan::getRoot(tree));
    
    for(;!atEnd(preIt); goNext(preIt))
    {
//        if (seqan::isLeaf(tree, *preIt))
//        {
//            genFStr << "# " << seqan::getProperty(_nodeNames, *preIt) << std::endl;
//            printGenome(seqan::getProperty(_groundTruth, *preIt), genFStr);
//        }
        if (!seqan::isLeaf(tree, *preIt) && !seqan::getProperty(_genomesPredicted, *preIt).empty())
        {
            genFStr << "# " << seqan::getProperty(_nodeNames, *preIt) << std::endl;
            
            //transform back to old gene names
            GeneString gen = seqan::getProperty(_genomesPredicted, *preIt).back();
            for (size_t i = 0; i < seqan::length(gen); ++i)
            {
                if(i)
                    genFStr << ',';
                if(gen[i] < 0)
                    genFStr << '-';
                
                genFStr << invGeneNameStore[gen[i]];
            }
            genFStr << std::endl;
        }
    }
    genFStr.close();
}


std::string GeneTree::_recursiveWrite(TVertexDescriptor vertex)
{
    if (seqan::isLeaf(tree, vertex))
    {
        std::string str = seqan::getProperty(_nodeNames, vertex);
        return str;
    }
    
    else
    {
        std::string str;
        TEdgeIt eIt(tree, vertex);
        for(; !atEnd(eIt); goNext(eIt))
        {
            //get branch length
            std::stringstream strstr;
            strstr << ":" << seqan::getProperty(_edgeCost, *eIt);

            TVertexDescriptor v = seqan::targetVertex(tree, *eIt);
            str = _recursiveWrite(v) + strstr.str() + "," + str;
        }
        str.resize(str.length()-1);
        str = "(" + str + ")" + seqan::getProperty(_nodeNames, vertex);
        return str;
    }
}

#endif
