// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef INVERSIONEVENTHOLDER_H
#define INVERSIONEVENTHOLDER_H

#include <EventHolderBase.h>


struct SparsePairGraph;

struct InversionEventHolder : EventHolderBase
{
  friend struct SparsePairGraph;

  struct InversionEvent
  {
      size_t G1Begin, G1End, G2Begin, G2End;

      InversionEvent(){};
      InversionEvent(size_t sb, size_t se, size_t tb, size_t te):
          G1Begin(sb), G1End(se), G2Begin(tb), G2End(te){}
  };


  void createInversionEvents(const GeneString & strA, const GeneString & strB);

  bool hasInversion(size_t posA, size_t posB)
  {
    return (posA < EdgeMap.size() && posB < EdgeMap[posA].size() && !EdgeMap[posA][posB].empty());
  }

  size_t numInversions() const {return inversions.size();}

  //DATA MEMBERS

  typedef std::vector<InversionEvent> TInversionEventList;
  typedef std::vector<std::vector<size_t> > TInversionMap;
  typedef std::vector<std::vector<std::vector<size_t> > > TEdgeMap;
  typedef std::vector<std::vector<std::pair<size_t, size_t> > > TInvEdgeMap;

  ///List of all inversion events for the given string
  TInversionEventList inversions;

  ///map from string positions (GenomeA and GenomeB) to covering inversion events
  TInversionMap invMapA, invMapB;

  ///map from alignment edges i,j to inversions that are represented by that edge
  TEdgeMap EdgeMap;

  ///map from inversion to alignement edges representing this inversions
  TInvEdgeMap invEdgeMap;


protected:
  void _reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos);
};


//to find possible inversion concatenate the genome with the inversion of the second genome
inline void InversionEventHolder::_reportEvent(size_t startPos, size_t endPos, size_t len, const TSuffArray &sa, size_t catPos)
{
    size_t len2 = len - 1;
    size_t lenG2 = seqan::length(sa) - catPos - 1;

    for (size_t i = startPos; i < endPos; ++i)
    {
        for (size_t j = i + 1; j <= endPos; ++j)
        {
            size_t l = seqan::_min(sa[i],sa[j]);
            size_t r = seqan::_max(sa[i],sa[j]);

            if (l > catPos || r < catPos)
                continue; //skip if both in same string

            r = lenG2 - (r - catPos) - len2; // for inverted duplication perform index correction

            inversions.push_back(InversionEvent(l, l + len2, r, r + len2));
            invEdgeMap.resize(inversions.size());

            for(size_t k = 0; k <= len2; ++k)
            {
                invMapA[l + k].push_back(inversions.size()-1);
                invMapB[r + k].push_back(inversions.size()-1);
                EdgeMap[l + k][r + k].push_back(inversions.size()-1);
                invEdgeMap[inversions.size()-1].push_back(std::make_pair(l+k, r+k));
            }
        }
    }
}


#endif // INVERSIONEVENTHOLDER_H
