// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef SPARSEPAIRGRAPH_H
#define SPARSEPAIRGRAPH_H

#include <seqan/graph_algorithms.h>
#include <Types.h>
#include <DuplicationEventHolder.h>
#include <PairedDuplicationEventHolder3.h>
#include <InversionEventHolder.h>

struct MixedCycleFinder; //forward declaration

struct SparsePairGraph{
    
    friend struct MixedCycleFinder;
    
    typedef TDirGraph TGraph;
    typedef seqan::VertexDescriptor<TGraph>::Type TVertexDescriptor;
    typedef seqan::EdgeDescriptor<TGraph>::Type TEdgeDescriptor;
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;
    
    typedef seqan::String<TVertexDescriptor> TPredMap;
    typedef std::vector<TVertexDescriptor> TVertexList;
    typedef seqan::String<TVertexDescriptor> TVertexString;
    
    typedef seqan::String<std::pair<size_t, size_t> >TPos2DPropertyMap;
    typedef seqan::String<seqan::String<size_t> > TInversionPropertyMap;
    typedef std::vector<TVertexList> TVertexMatrix;

    
    struct MixedClique
    {
        char seq;
        size_t left;
        size_t right;
        double score;
        TVertexList vIds;
        TSizeArray alnEdgeIds;
        TSizeArray dupIds;
        TSizeArray inversionIds;
        TSizeArray dupIdsP;        

        void clear()
        {
            vIds.clear();
            alnEdgeIds.clear();
            dupIds.clear();
            inversionIds.clear();
        }
    };
    
    typedef std::vector<MixedClique> TMixedCliqueList;

    
    TGraph graph;
    
    ///find a violated clique that can be used as cutting plane    
    void findViolatedCliques(const IloNumArray &inversionScores,
                             MixedClique &clique,
                             IloNum eps = 0.
                             );

    void findViolatedCliques(MixedClique &clique,
                             IloNum eps = 0.
                             );
    
    void assignWeights(const IloNumArray & alnScores);

    ///create sparse pairgraph
    ///Idea: Not by row but by column. Top down with lookback in the corresponding column
    void createPG(const GeneString & strA,
                  const GeneString & strB,
                  bool inversions = false);

    ///Idea: Like createPG but an augmented PG is constructed to allow for interval queries
    void createAugmentedPG(const GeneString & strA,
                           const GeneString & strB,
                           bool inversions = false);

    void getMixedCliques(DuplicationEventHolder &holderA,
                         DuplicationEventHolder &holderB,
                         const IloNumArray &inversionScores,
                         const IloNumArray &dupAScores,
                         const IloNumArray &dupBScores,
                         TMixedCliqueList &cliques,
                         IloNum eps = 0.);

    void getMixedCliques(DuplicationEventHolder &holderA,
                         DuplicationEventHolder &holderB,
                         const IloNumArray &dupAScores,
                         const IloNumArray &dupBScores,
                         TMixedCliqueList &cliques,
                         IloNum eps = 0.);

    void getMixedCliques(DuplicationEventHolder &holder,
                         const IloNumArray &dupScores,
                         TMixedCliqueList &cliques,
                         IloNum eps = 0.,
                         size_t id = 0);
    
    void getMixedCliques(DuplicationEventHolder &holder,
                         const IloNumArray &inversionScores,
                         const IloNumArray &dupScores,
                         TMixedCliqueList &cliques,
                         IloNum eps = 0.,
                         size_t id = 0);
    
    void getMixedCliquesPaired(PairedDuplicationEventHolder3 &holderP,
                               DuplicationEventHolder & holderC,
                               const IloNumArray &dupPScores,
                               const IloNumArray &dupCScores,
                               TMixedCliqueList &cliques,
                               IloNum eps = 0.);
    
    //compute the maximal Clique within the interval [startPos, endPos] in genome id
    //store the corresponding IloIds in the clique vector and return the score;
    //requires prepareIntervalQueries to be called first!!
    double getMaxIntervalClique(size_t startPos,
                                size_t endPos,
                                size_t id,
                                std::vector<size_t> & clique);
    
    //Function prepares the datastructures to use getMaxIntervalClique
    void prepareIntervalQueries();


    void createShortestPathTree(TDistanceMap & distMap,
                                TPredMap & predMap,
                                TVertexDescriptor start);

    size_t strLenA() const {return _strLenA;}
    size_t strLenB() const {return _strLenB;}

    size_t numAlignEdges() const {return _numEdges;}
    size_t numInversions() const {return _inversionHolder.numInversions();}


    bool hasAlignEdge(size_t i, size_t j) const
    {
        return (_vertices[i][j] != seqan::getNil<TVertexDescriptor>() &&  vertexToIloId(i,j) != -1);
    }

    size_t vertexToIloId(TVertexDescriptor v) const 
    {
        return seqan::property(_vertexToIloId, v);
    }

    size_t vertexToIloId(size_t i, size_t j) const
    {
        return vertexToIloId(_vertices[i][j]);
    }
    
    TVertexDescriptor getVertexDesc(size_t i, size_t j) const
    {
        return _vertices[i][j];
    }
    
    void getVertexInv(TVertexDescriptor v, size_t &p1, size_t &p2) const
    {
        const std::pair<size_t, size_t> & tmp = _verticesInv[v];
        p1 = tmp.first;
        p2 = tmp.second;
    }
    
    void iloIdToVertex(size_t id, size_t &p1, size_t &p2) const
    {   const std::pair<size_t, size_t> & tmp = _verticesInv[_iloIdToVertex[id]];
        p1 = tmp.first;
        p2 = tmp.second;        
    }
    
    const InversionEventHolder & getInversionHolder() const
    {
        return _inversionHolder;
    }


protected:

    ///access vertex from matrix cell indices (eg. alignment edge ij)
    TVertexMatrix _vertices;
    ///access matrix cell indices (eg. alignment edge ij) from vertex
    TPos2DPropertyMap _verticesInv;
    ///get vertex corresponding array index of cplex variable
    TVertexList _iloIdToVertex;
    ///get array index of cplex variable for corresponding vertex
    TPosPropertyMap _vertexToIloId;
    ///weights for the edges in the pairgraph
    TGraphWeightPropertyMap _edgeWeights;
    ///maps to every node (alignment edge) the inversion events it is representing
    TInversionPropertyMap _inversionMap;
    ///Inversion holder in case that inversion are considered
    InversionEventHolder _inversionHolder;
    
    size_t _strLenA;
    size_t _strLenB;
    size_t _numEdges;
    
    TVertexDescriptor _source;
    TVertexDescriptor _sink;
    
    seqan::String<seqan::String<TVertexDescriptor> > _predMapsA, _predMapsB;
    seqan::String<seqan::String<double> > _distMapsA, _distMapsB;
    
    bool _inversions;
    
    void _inversionsToEdges(const IloNumArray &inversionScores, IloNum eps);
};


inline void SparsePairGraph::_inversionsToEdges( const IloNumArray &inversionScores,
                                                 IloNum eps)
{
    //put inversion scores on edges scores
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;
    for (unsigned i = 0; i < inversionScores.getSize(); ++i)
    {
        if (inversionScores[i] < eps)
            continue;

        double score = (-1) * inversionScores[i];
        for (size_t j = 0; j < _inversionHolder.invEdgeMap[i].size(); ++j)
        {
            //std::cerr << "adjust score of edge " << inversionHolder.invEdgeMap[i][j].first << " -- " <<inversionHolder.invEdgeMap[i][j].second <<std::endl;
            TVertexDescriptor v = _vertices[_inversionHolder.invEdgeMap[i][j].first][_inversionHolder.invEdgeMap[i][j].second];
            TOutEdgeIterator outIter(graph, v);

            while(!seqan::atEnd(outIter))
            {
                //std::cerr << "set edge " << seqan::getValue(outIter) << " to: " << score << std::endl;
                //seqan::assignProperty(edgeWeights, seqan::getValue(outIter), score);
                seqan::property(_edgeWeights, seqan::getValue(outIter)) += score;
                ++outIter;
            }
        }
    }
}

inline void SparsePairGraph::findViolatedCliques(const IloNumArray &inversionScores,
                                                 MixedClique &clique,
                                                 IloNum eps)
{
    if (_inversions)
        _inversionsToEdges(inversionScores, eps);

//    else
//        std::cerr << "No Inversions Defined - Ignoring scores" << std::endl;
    
    return findViolatedCliques(clique, eps);
}



inline void SparsePairGraph::createShortestPathTree(TDistanceMap & distMap,
                                                    TPredMap & predMap,
                                                    TVertexDescriptor start)
{
    //compute shortest path in pairgraph
    seqan::dagShortestPath(graph, start, _edgeWeights, predMap, distMap);
}


inline void SparsePairGraph::assignWeights(const IloNumArray & alnScores)
{
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;
    
    //reset edge weights
    std::fill(seqan::begin(_edgeWeights), seqan::end(_edgeWeights), 0.0);
    
    //put vertex scores on edges
    for (unsigned i = 0; i < alnScores.getSize(); ++i)
    {
        TVertexDescriptor v = _iloIdToVertex[i];
        
        double score = (-1) * alnScores[i];
        TOutEdgeIterator outIter(graph, v);
        
        while(!seqan::atEnd(outIter))
        {
            seqan::property(_edgeWeights, seqan::getValue(outIter)) += score;
            ++outIter;
        }
    }
    return;
}

//implementation of template member functions
inline void SparsePairGraph::findViolatedCliques(MixedClique &clique,
                                                 IloNum eps)

{
    clique.clear();
    TPredMap predMap;
    TDistanceMap distMap;

    //compute shortest path in pairgraph
    createShortestPathTree(distMap, predMap, _source);

    //if found a violated clique report it    
    if(seqan::getProperty(distMap, _sink) < -(1 + eps))
    {
#ifdef DEBUGILP
        std::cerr << "violated clique score: " << seqan::getProperty(distMap, _sink) << std::endl;
#endif
        //traceback to get the violated clique
        TVertexDescriptor v = _sink;
        while((v = seqan::getProperty(predMap, v)) != _source)
        {
            clique.vIds.push_back(v);
            if (seqan::getProperty(_vertexToIloId, v) != -1)
                clique.alnEdgeIds.push_back(seqan::getProperty(_vertexToIloId, v));

            if (_inversions)
                clique.inversionIds.insert(clique.inversionIds.end(), seqan::begin(seqan::getProperty(_inversionMap, v)), seqan::end(seqan::getProperty(_inversionMap, v)));
        }
    }    
}

#endif // SPARSEPAIRGRAPH_H
