// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#include <iostream>
#include <GeneTree.h>
#include <seqan/arg_parse.h>

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;

struct DuPloCutOptions
{
    std::string treeFilename;
    std::string genFilename;
    std::string outFilePrefix;
    unsigned threads;
    IlpOptions ilpOpts;
    unsigned reoptRounds;
    
    DuPloCutOptions() : treeFilename(""),
                        genFilename(""),
                        outFilePrefix(""),
                        threads(1),
                        ilpOpts(IlpOptions()),
                        reoptRounds(5)
    {}
};

seqan::ArgumentParser::ParseResult
parseCommandLine(DuPloCutOptions & options, int argc, char const ** argv)
{
    // Setup ArgumentParser.
    seqan::ArgumentParser parser("DuPloCut");
    setShortDescription(parser, "DuPloCut Duplication Loss Evolution using Branch and Cut");
    setVersion(parser, "0.1");
    setDate(parser, "March 2013");
    
    // Define usage line and long description.
    addUsageLine(parser,
                 "[\\fIOPTIONS\\fP] \"\\fITEXT\\fP\"");
    addDescription(parser,
                   "This program allows to compute the median genomes under the duplication-loss model"
                   " for all internal nodes in the phylogenetic tree given the genomes of the leaves.");
    // Define Options
    addOption(parser, seqan::ArgParseOption("i", "tree-file",
                                            "A file containing the tree structure in Newick format",
                                            seqan::ArgParseArgument::INPUTFILE));
    setRequired(parser, "tree-file");
    
    addOption(parser, seqan::ArgParseOption("g", "genome-file",
                                            "A file containing the genomes (ids must match those in tree file)",
                                            seqan::ArgParseArgument::INPUTFILE));
    setRequired(parser, "genome-file");
    
    addOption(parser, seqan::ArgParseOption("o", "output-prefix",
                                            "Two files will be generated. prefix.tree and prefix.medians",
                                            seqan::ArgParseArgument::OUTPUTFILE));
    setRequired(parser, "output-prefix");

    
    
    addOption(parser, seqan::ArgParseOption("j", "threads", "Number of threads to use.",
                                            seqan::ArgParseArgument::INTEGER, "INT"));
    setMinValue(parser, "threads", "1");
    setDefaultValue(parser, "threads", "1");
    
    addOption(parser, seqan::ArgParseOption("r", "rounds", "Number of re-optimization rounds for the tree.",
                                            seqan::ArgParseArgument::INTEGER, "INT"));
    setMinValue(parser, "rounds", "0");
    setDefaultValue(parser, "rounds", "3");
    
    addSection(parser, "Model");
    
    addOption(parser, seqan::ArgParseOption("s", "visible-source",
                                            "Use the strict model of visible history for source of paired duplications."));

    addOption(parser, seqan::ArgParseOption("t", "visible-target",
                                            "Use the strict model of visible history for target of paired duplications."));
    
       // Parse command line.
    seqan::ArgumentParser::ParseResult res = seqan::parse(parser, argc, argv);
    // Only extract  options if the program will continue after parseCommandLine()
    if (res != seqan::ArgumentParser::PARSE_OK)
        return res;
    
    // Extract option values.
    getOptionValue(options.threads, parser, "threads");
    options.ilpOpts.sourceDisruptConstraints = isSet(parser, "visible-source");
    options.ilpOpts.targetDisruptConstraints = isSet(parser, "visible-target");
    getOptionValue(options.reoptRounds, parser, "rounds");
    getOptionValue(options.treeFilename, parser, "tree-file");
    getOptionValue(options.genFilename, parser, "genome-file");
    getOptionValue(options.outFilePrefix, parser, "output-prefix");
    
    return seqan::ArgumentParser::PARSE_OK;
}

void optimizeTree(const DuPloCutOptions &opts)
{
    GeneTree tree;
    tree.parseTree(opts.treeFilename, opts.genFilename);
    //return;
    
    double cost;
    cost = tree.reoptimize(true, opts.ilpOpts);
    //tree.evaluatePredictions(opts.ilpOpts);
    //cerr << "TotalCost of init step: " << cost << endl;
    cerr << "Tree Initialization Done" << endl;
    
    for (size_t r = 0; r < opts.reoptRounds; ++r)
    {
        cerr << "Reoptimization round " << r + 1 << " of " << opts.reoptRounds << endl;
        cost = tree.reoptimize(false, opts.ilpOpts);
        //cerr << "TotalCost of reopt Phase " << r+1 << ": " << cost << endl;
        //tree.evaluatePredictions(opts.ilpOpts);
    }
    cerr << "Complete Tree has Dup-Loss cost: " <<  cost << endl;
    tree.writeTree(opts.outFilePrefix + ".tree", opts.outFilePrefix + ".medians");
    return;
}


int main (int argc, const char ** argv)
{
    // Parse the command line.
    DuPloCutOptions options;
    seqan::ArgumentParser::ParseResult res = parseCommandLine(options, argc, argv);
    
    // If parsing was not successful then exit with code 1 if there were errors.
    // Otherwise, exit with code 0 (e.g. help was printed).
    if (res != seqan::ArgumentParser::PARSE_OK)
        return res == seqan::ArgumentParser::PARSE_ERROR;
    
    double startTime = sysTime();
    
#ifdef _OPENMP
    omp_set_num_threads(options.threads);
#endif
    
    optimizeTree(options);
    
    std::cerr << "Total Runtime " << sysTime() - startTime << " s." << std::endl;
}
