// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <DuplicationEventHolder.h>
#include <CycleDetect.h>
#include <seqan/index.h>


using namespace std;

void DuplicationEventHolder::createDuplicationEventsLCP(const GeneString & str, bool inversions)
{   
    _inversions = inversions;
    
    dupMap.clear();
    dupMap.resize(seqan::length(str));    
    dupInterMap.assign(seqan::length(str), vector<vector<size_t> >(seqan::length(str)));
    duplications.clear();

    piV.assign(seqan::length(str),  TFloatArray(seqan::length(str), 0.));
    pirV.assign(seqan::length(str),  TFloatArray(seqan::length(str), 0.));
    sigV.assign(seqan::length(str), TFloatArray(seqan::length(str), 0.));
    piI.assign(seqan::length(str), std::vector<TSizeArray>(seqan::length(str)));
    sigI.assign(seqan::length(str), std::vector<TSizeArray>(seqan::length(str)));    

    size_t catPos = seqan::length(str);
    if (_inversions)
    {
        GeneString str2 = str, inv = str;
        seqan::appendValue(str2,0);
        seqan::reverse(inv);
        for (int i = 0; i < seqan::length(str); ++i)
            inv[i] *= -1;

        seqan::append(str2, inv);
        //std::cerr << seqan::length(str2) << "  "  << catPos << std::endl;
        _createEventsLCP(str2, catPos);
    }
    else
    {
        _createEventsLCP(str, catPos);
    }

    _init();
    _createDuplicationCliques();
}



void DuplicationEventHolder::_createDuplicationGraph(TGraph & graph,
                                                     TVertexList & vertices,
                                                     TPosPropertyMap & verticesInv,
                                                     TGraphWeightPropertyMap & vScores,
                                                     const IloNumArray & vScoresIlo)
{
    float eps = 0.000001;
    vertices.reserve(duplications.size());
    for(size_t i = 0; i < duplications.size(); ++i)
    {
        if(vScoresIlo[i] >= eps)
        {
            vertices.push_back(seqan::addVertex(graph));
        }
    }

    seqan::resizeVertexMap(graph, vScores);
    seqan::resizeVertexMap(graph, verticesInv);
    for(size_t i = 0, id = 0; i < duplications.size(); ++i)
    {
        if(vScoresIlo[i] >= eps)
        {
            seqan::assignProperty(vScores, vertices[id], 1 - vScoresIlo[i]);
            seqan::assignProperty(verticesInv, vertices[id], i);
            ++id;
        }
    }

    for (size_t i = 0; i < vertices.size(); ++i)
    {
        const DuplicationEvent & dupI = duplications[seqan::getProperty(verticesInv, vertices[i])];

        for (size_t j = i+1; j < vertices.size(); ++j)
        {

            const DuplicationEvent & dupJ = duplications[seqan::getProperty(verticesInv, vertices[j])];

        //    std::cerr << i << '\t' << j << '\t' << dupI.targetBegin <<'\t' << dupJ.sourceBegin << '\t' << dupI.targetEnd << '\t' << std::endl;

            if(!(dupI.targetEnd <  dupJ.sourceBegin || dupI.targetBegin > dupJ.sourceEnd))
            {
                seqan::addEdge(graph, vertices[i], vertices[j]);
//                std::cerr << i << " --> " << j << "   " << "  "  << vScores[i] << "::" << vScores[j] << std::endl;
//                std::cerr << '[' << dupI.sourceBegin << " " << dupI.sourceEnd << "][" << dupI.targetBegin << " " << dupI.targetEnd << " --> " << '[' << dupJ.sourceBegin << " " << dupJ.sourceEnd << "][" << dupJ.targetBegin << " " << dupJ.targetEnd << ']' <<std::endl;
            }
            if(!(dupJ.targetEnd < dupI.sourceBegin ||  dupJ.targetBegin > dupI.sourceEnd))
            {
                seqan::addEdge(graph, vertices[j], vertices[i]);
//                std::cerr << j << " --> " << i << "  " << vScores[j] << "::" << vScores[i] << std::endl;
//                std::cerr << '[' << dupJ.sourceBegin << " " << dupJ.sourceEnd << "][" << dupJ.targetBegin << " " << dupJ.targetEnd << " --> " << '[' << dupI.sourceBegin << " " << dupI.sourceEnd << "][" << dupI.targetBegin << " " << dupI.targetEnd << ']' <<std::endl;
            }
        }
    }
}



void DuplicationEventHolder::_createLiftedDuplicationGraph(TGraph & graph,
                                                     TVertexList & vertices,
                                                     TPosPropertyMap & verticesInv,
                                                     TGraphWeightPropertyMap & vScores,
                                                     const IloNumArray & vScoresIlo)
{
    float eps = 0.000001;
    TVertexDescriptor NIL = seqan::getNil<TVertexDescriptor>();

    vector<double> cliqueScores(_duplicationCliques.size(), 0);
    vector<TVertexDescriptor> vertexMap(_duplicationCliques.size(), NIL);

    for(size_t i = 0; i < duplications.size(); ++i)
    {
        for(size_t j = 0; j < _dupToCliques[i].size(); ++j)
        {
            cliqueScores[_dupToCliques[i][j]] += vScoresIlo[i];
        }
    }

    vertices.reserve(_duplicationCliques.size());
    for(size_t i = 0; i < _duplicationCliques.size(); ++i)
    {
        if(cliqueScores[i] >= eps)
        {
            vertices.push_back(seqan::addVertex(graph));
            vertexMap[i] = vertices.back();
        }
    }


    seqan::resizeVertexMap(graph, vScores);
    seqan::resizeVertexMap(graph, verticesInv);
    for(size_t i = 0; i < _duplicationCliques.size(); ++i)
    {
        if(vertexMap[i] != NIL)
        {
            seqan::assignProperty(vScores, vertexMap[i], 1 - cliqueScores[i]);
            seqan::assignProperty(verticesInv, vertexMap[i], i);
        }
    }

    //create edges
    for(size_t i = 0; i < _duplicationCliques.size(); ++i)
    {
        if (vertexMap[i] != NIL)
        {
            for(size_t j = 0; j < _cliquesAdjList[i].size(); ++j)
            {
                if (vertexMap[_cliquesAdjList[i][j]] != NIL)
                {
                    seqan::addEdge(graph, vertexMap[i], vertexMap[_cliquesAdjList[i][j]]);
                }
            }
        }
    }
}


void DuplicationEventHolder::_createDuplicationCliques()
{

    TDuplicationMap mapS(dupMap.size()), mapT(dupMap.size());

    typedef std::map<std::vector<size_t>, size_t> TCliqueMap;

    TCliqueMap detectedCliques;

    //for each position pos store all duplication that cover pos with source or target
    for (size_t i = 0; i < dupMap.size(); ++i)
    {
        for (size_t j = 0; j < dupMap[i].size(); ++j)
        {
            if (dupMap[i][j] % 2 == 1)
            {
                mapS[i].push_back(dupMap[i][j] - 1);
                mapT[i].push_back(dupMap[i][j]);
            }
            else
            {
                mapT[i].push_back(dupMap[i][j]);
                mapS[i].push_back(dupMap[i][j] + 1);
            }
        }
        std::sort(mapS[i].begin(), mapS[i].end());
        std::sort(mapT[i].begin(), mapT[i].end());
    }

    //for each pair of positions i,j store all duplications with source covering i and target covering j
    //these are the "cliques"
    TDuplicationMap mapCS(dupMap.size()), mapCT(dupMap.size());
    size_t id = 0;
    for (size_t i = 0; i < dupMap.size(); ++i)
    {
        for (size_t j = 0; j < dupMap.size(); ++j)
        {
            if (i == j)
                continue;

            std::vector<size_t> intersec(std::min(mapS[i].size(), mapT[j].size()));
            std::vector<size_t>::iterator it = std::set_intersection(mapS[i].begin(), mapS[i].end(), mapT[j].begin(), mapT[j].end(), intersec.begin());
            size_t offset = it - intersec.begin();

            if (offset == 0)
                continue;

            intersec.resize(offset);
            std::sort(intersec.begin(), intersec.end());

            TCliqueMap::iterator itC = detectedCliques.find(intersec);
            if (itC == detectedCliques.end())
            {
                detectedCliques[intersec] = id;
                ++id;
            }

            size_t index = detectedCliques[intersec];
            //add clique as source to pos i
            mapCS[i].push_back(index);
            //add clique as target to pos j
            mapCT[j].push_back(index);
        }
    }

    //fill the accessor maps that:
    //    - allow to access all cliques containing a certain duplication
    //    - allow to access all duplications contained in a certain clique
    size_t numCliques = detectedCliques.size();
    _duplicationCliques.resize(numCliques);
    _dupToCliques.resize(duplications.size());

    TCliqueMap::iterator itC = detectedCliques.begin();
    while (itC != detectedCliques.end())
    {
        _duplicationCliques[itC->second] = itC->first;

        for  (size_t i = 0; i < itC->first.size(); ++i)
        {
            _dupToCliques[itC->first[i]].push_back(itC->second);
        }
        ++itC;
    }

    //create adjacency lists (temporary in set to remove duplicates)
    vector<set<size_t> > tmpAdjList(numCliques);
    for (size_t i = 0; i < dupMap.size(); ++i)
    {
        for (size_t t = 0; t < mapCT[i].size(); ++t)
        {
            for (size_t s = 0; s < mapCS[i].size(); ++s)
            {
                tmpAdjList[mapCT[i][t]].insert(mapCS[i][s]);
            }
        }
    }

    //from sets to vectors
    _cliquesAdjList.resize(numCliques);
    for (size_t i = 0; i < numCliques; ++i)
    {
        _cliquesAdjList[i].insert(_cliquesAdjList[i].end(), tmpAdjList[i].begin(), tmpAdjList[i].end());
    }
}

void DuplicationEventHolder::_getDupCycle(const IloNumArray & scores, TviolatedDupList & cycles, double threshold, bool all, bool lifted)
{
    typedef seqan::Iterator<TGraph, seqan::EdgeIterator>::Type TEdgeIterator;

    //construct the duplication graph
    TGraph graph;
    TVertexList vertices;
    TPosPropertyMap verticesInv;
    TGraphWeightPropertyMap vertexScores;

    if (lifted)
        _createLiftedDuplicationGraph(graph, vertices, verticesInv, vertexScores, scores);

    else
        _createDuplicationGraph(graph, vertices, verticesInv, vertexScores, scores);

    if(seqan::empty(graph))
        return;
    
    //put vertex weights onto outgoing edges
    TGraphWeightPropertyMap edgeWeights;
    resizeEdgeMap(graph, edgeWeights);
    TEdgeIterator it(graph);
    while (!atEnd(it))
    {
        seqan::assignProperty(edgeWeights, *it, seqan::getProperty(vertexScores, seqan::sourceVertex(graph, *it)));
        ++it;
    }

#ifdef DEBUG
    //std::cerr << "graph has " << seqan::numVertices(graph) << "  Vertices and " << seqan::numEdges(graph) << " Edges" <<std::endl;
#endif

    seqan::String<seqan::String<TVertexDescriptor> >cyclesV; //cycle as strings of vertices
    vector<double> cScores;

    if(!all) //compute only the shortest cycle
    {        
        seqan::String<TVertexDescriptor> minCycle;
        double score = shortestCycle(graph, edgeWeights, minCycle, threshold);

        if(score < threshold)
        {
            cScores.push_back(score);
            seqan::appendValue(cyclesV, minCycle);
        }
    }
    else //compute all cycles
    {
        enumerateAllCycles(graph, cyclesV, cScores, edgeWeights, threshold);
    }

    cycles.resize(seqan::length(cyclesV));
    for (size_t i = 0; i < seqan::length(cyclesV); ++i)
    {
        if (!lifted)
            seqan::resize(cycles[i].dupIds, seqan::length(cyclesV[i]));

        for (int j = 0; j < seqan::length(cyclesV[i]); ++j)
        {
            if (lifted)
            {
                size_t cliqueId = seqan::getProperty(verticesInv, cyclesV[i][j]);
                cycles[i].dupIds.insert(cycles[i].dupIds.end(), _duplicationCliques[cliqueId].begin(), _duplicationCliques[cliqueId].end());
            }
            else
            {
                cycles[i].dupIds[j] = seqan::getProperty(verticesInv, cyclesV[i][j]);
            }
        }
        cycles[i].cycleLen = seqan::length(cyclesV[i]);
        cycles[i].score = cycles[i].cycleLen - cScores[i];
    }
}


//void DuplicationEventHolder::getLiftedDuplicationCycle(const IloNumArray & scores, TviolatedDupList & cycles, double threshold, bool all)
//{
//    typedef seqan::Iterator<TGraph, seqan::EdgeIterator>::Type TEdgeIterator;

//    //construct the duplication even graph
//    TGraph graph;
//    TVertexList vertices;
//    TPosPropertyMap verticesInv;
//    TGraphWeightPropertyMap vertexScores;

//    _createLiftedDuplicationGraph(graph,
//                                  vertices,
//                                  verticesInv,
//                                  vertexScores,
//                                  scores);

//    if(seqan::empty(graph))
//        return;

//    //put vertex weights onto outgoing edges
//    TGraphWeightPropertyMap edgeWeights;
//    resizeEdgeMap(graph, edgeWeights);
//    TEdgeIterator it(graph);
//    while (!atEnd(it))
//    {
//        seqan::assignProperty(edgeWeights, *it, seqan::getProperty(vertexScores, seqan::sourceVertex(graph, *it)));
//        ++it;
//    }


//#ifdef DEBUG
//    //std::cerr << "graph has " << seqan::numVertices(graph) << "  Vertices and " << seqan::numEdges(graph) << " Edges" <<std::endl;
//#endif


//    seqan::String<seqan::String<TVertexDescriptor> >cyclesV; //cycle as strings of vertices
//    vector<double> cScores;

//    if(!all) //compute only the shortest cycle
//    {
//        seqan::String<TVertexDescriptor> minCycle;
//        double score = shortestCycle(graph, edgeWeights, minCycle, threshold);

//        if(score < threshold)
//        {
//            cScores.push_back(score);
//            seqan::appendValue(cyclesV, minCycle);
//        }
//    }
//    else //compute all cycles
//    {
//        enumerateAllCycles(graph, cyclesV, cScores, edgeWeights, threshold);
//    }

///*  //old version
//    TVertexString pred;
//    seqan::String<int> disc, fin;
//    detectCycles2(graph, pred, disc, fin, vertexScores, cycleV, cScores, 1. - eps);
//*/
///* //even older version
//    //just for testing now:
//    TVertexString pred2;
//    seqan::String<int> disc2, fin2;
//    seqan::String<seqan::String<TVertexDescriptor> >cycleV2; //cycle as strings of vertices
//    vector<double> cScores2;
//    detectCycles2(graph, pred2, disc2, fin2, vertexScores, cycleV2, cScores2, 1. - eps);
//    if(length(cycleV2) != length(cycleV))
//        std::cerr << "DIFFER: " << length(cycleV2) << " vs. " << length(cycleV) << std::endl;
//*/

//    cycles.resize(seqan::length(cyclesV));

//    for (size_t i = 0; i < seqan::length(cyclesV); ++i)
//    {
//        //seqan::resize(cycles[i], seqan::length(cycleV[i]));
//        //transform vertices into cplex variable indices
//        for (int j = 0; j < seqan::length(cyclesV[i]); ++j)
//        {
//            size_t cliqueId = seqan::getProperty(verticesInv, cyclesV[i][j]);
//            cycles[i].dupIds.insert(cycles[i].dupIds.end(), _duplicationCliques[cliqueId].begin(), _duplicationCliques[cliqueId].end());
//        }
//        cycles[i].cycleLen = seqan::length(cyclesV[i]);
//        cycles[i].score = cycles[i].cycleLen - cScores[i];
//        //cycleLens[i] = seqan::length(cycleV[i]);
//    }

////    std::sort(cycles.begin(), cycles.end());
//}







