// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#include <IlpMedian.h>
#include <MixedCycleFinder.h>

#include <lemon/list_graph.h>
#include <lemon/concepts/maps.h>
#include <Subtour.h>



//#define PAIRWISECONFLICT
//#define DEBUGILP
//#define DEBUGDLC

using namespace std;

const size_t MINUSONE = -1;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//-----------------------ALIGNMENT PRINTING FUNCTIONS---------------------------
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void _printColumn(const column & col, const std::vector<GeneString> & genomes)
{
    for (size_t i = 0; i < col.alPos.size(); ++i)
    {
        if (col.alPos[i] == MINUSONE)
            cerr <<"-\t";
        else
            //cerr << genomes[i][col.alPos[i]] <<'\t';
            cerr << col.alPos[i] <<'\t';
    }
    if (col.type == column::LOSS)
        cerr << "L";
    else if (col.type == column::LOSSCONS)
        cerr << "LC";
    else if (col.type == column::DUP)
        cerr << "D_" << col.id.second;
    else if (col.type == column::PAIRED)
        cerr << "P_" << col.id.second;
    else if (col.type == column::DUPLOSS)
        cerr << "PL_" << col.id.second;
    
    cerr<<endl;
    
    if (col.type == column::UNDEF)
    {
        cerr << " UNDEF Column EXIT" << std::endl;
        exit(-1);
    }
}

//@DEPRECATED is replaced by _getAlignment2
void _getAlignment(const ValHolderMedian2 &vals,
                   const CallBackDataMedian2 & cbData,
                   const std::vector<GeneString> & genomes,
                   TAlignment & alignment,
                   double eps,
                   bool asymmetric,
                   IloNumArray penaltyVals)
{
    size_t numGenomes = genomes.size();
    std::vector<std::vector<column> > columns;
    columns.resize(numGenomes);
    
    std::vector<std::vector<bool> >  covered(genomes.size());
    
    for (size_t i = 0; i < genomes.size(); ++i)
        covered[i].assign(seqan::length(genomes[i]), false);
    
    for (size_t i = 0; i < numGenomes; ++i)
    {
        for (size_t pos = 0; pos < seqan::length(genomes[i]); ++pos)
        {
            if (covered[i][pos])
                continue;
            
            bool done = false;
            
            column col;
            col.type = column::UNDEF;
            col.alPos.assign(numGenomes, -1);
            col.alPos[i] = pos;
            col.consensus = genomes[i][pos];
            
            //check whether duplication
            if (!cbData.dupEvents[i].duplications.empty())
            {
                for (size_t dup = 0; dup < cbData.dupEvents[i].dupMap[pos].size(); ++dup)
                {
                    if (vals.duplicationVals[i][cbData.dupEvents[i].dupMap[pos][dup]] > eps)
                    {
                        col.type = column::DUP;
                        col.id = std::make_pair(i, cbData.dupEvents[i].dupMap[pos][dup]);
                        done = true;
                        break;
                    }
                }
            }
            
            //check whether paired or duploss scenario for 1
            if (i == 1 && !cbData.pairedDupEvents[0].duplications.empty())
            {
                for (size_t dup = 0; dup < cbData.pairedDupEvents[0].dupMap[pos].size(); ++dup)
                {
                    size_t pId = cbData.pairedDupEvents[0].dupMap[pos][dup];
                    
                    if (vals.pairedDuplicationVals[0][pId] > eps)
                    {
                        cerr << "paired dup for 1: " << pos <<  "  " << pId << endl;
                        for (size_t pos2 = 0; pos2 < seqan::length(genomes[2]); ++pos2)
                        {
                            if (cbData.pairGraphs[1][2].hasAlignEdge(pos, pos2))
                            {
                                size_t aId = cbData.pairGraphs[1][2].vertexToIloId(pos, pos2);
                                if (vals.wVals[1][aId] > eps)
                                {
                                    col.type = column::PAIRED;
                                    col.id = std::make_pair(1, pId);
                                    col.alPos[2] = pos2;
                                    covered[2][pos2] = true;
                                    cerr << "cover 2: " << pos2 << endl;
                                    done = true;
                                    break;
                                }
                            }
                        }
                        if (!done)
                        {
                            col.type = column::DUPLOSS;
                            col.id = std::make_pair(1, pId);
                            done = true;
                            break;
                        }
                    }
                }
            }
            //check whether paired or duploss scenario for 2
            if (i == 2 && !cbData.pairedDupEvents[1].duplications.empty())
            {
                for (size_t dup = 0; dup < cbData.pairedDupEvents[1].dupMap[pos].size(); ++dup)
                {
                    size_t pId = cbData.pairedDupEvents[1].dupMap[pos][dup];
                    
                    if (vals.pairedDuplicationVals[1][pId] > eps)
                    {
                        cerr << "paired dup for 2: " << pos << "  " << pId << endl;
                        for (size_t pos1 = 0; pos1 < seqan::length(genomes[1]); ++pos1)
                        {
                            if (cbData.pairGraphs[1][2].hasAlignEdge(pos1, pos))
                            {
                                size_t aId = cbData.pairGraphs[1][2].vertexToIloId(pos1, pos);
                                if (vals.wVals[0][aId] > eps)
                                {
                                    col.type = column::PAIRED;
                                    col.alPos[1] = pos1;
                                    col.id = std::make_pair(2, pId);
                                    covered[1][pos1] = true;
                                    cerr << "cover 1: " << pos1 << endl;
                                    done = true;
                                    break;
                                }
                            }
                        }
                        if (!done)
                        {
                            col.type = column::DUPLOSS;
                            col.id = std::make_pair(2, pId);
                            done = true;
                            break;
                        }
                    }
                }
            }
            
            //check whether align column
            if (!done && ((i == 0) || !asymmetric))
            {
                bool align = (i == 0); //only align column if also genome 0 is aligned.
                for (size_t j = i + 1; j < numGenomes; ++j)
                {
                    bool alignJ = false;
                    for (size_t posJ = 0; posJ < seqan::length(genomes[j]); ++posJ)
                    {
                        if (cbData.pairGraphs[i][j].hasAlignEdge(pos, posJ))
                        {
                            size_t aId = cbData.pairGraphs[i][j].vertexToIloId(pos, posJ);
                            if (vals.alignEdgesVals[i][j][aId] > eps)
                            {
                                col.alPos[j] = posJ;
                                covered[j][posJ] = true;
                                cerr << "cover (AL)" << j << "pos: " << posJ << endl;
                                alignJ = true;
                                break;
                            }
                        }
                    }
                    align = align && alignJ;
                }
                if (align)
                {
                    col.type = column::AL;
                    done = true;
                }
            }
            if(!done && (i == 0 || !asymmetric) ) //has to be a loss column
            {
                if (!asymmetric)
                    col.type = column::LOSSCONS;
                
                else if ((size_t)penaltyVals.getSize() != seqan::length(genomes[0]))
                {
                    std::cerr << "must be called with valid penalty vals! - Exit! " << std::endl;
                    exit(-1);
                }
                else if (col.alPos[1] == MINUSONE && col.alPos[2] == MINUSONE && penaltyVals[pos] <= eps)
                    col.type = column::LOSS; //does not appear in median
                
                else
                    col.type = column::LOSSCONS; // does appear in median
                
                done = true;
            }
            
            if (done)
            {
                covered[i][pos] = true;
                columns[i].push_back(col);
            }
        }
    }
    
    seqan::clear(alignment);
    alignment.insert(alignment.end(), columns[0].begin(), columns[0].end());
    
    //now merge the remaining columns that had no position in
    for (size_t i = 1; i < numGenomes; ++i)
    {
        std::vector<column>::iterator itAl = alignment.begin();
        for (size_t posCol = 0; posCol < columns[i].size(); ++posCol)
        {
            while (itAl != alignment.end() && !(columns[i][posCol] < *itAl))
            {
                ++itAl;
            }
            itAl = alignment.insert(itAl, columns[i][posCol]);
        }
    }
}

void _getAlignment2(const ValHolderMedian2 &vals,
                    const CallBackDataMedian2 & cbData,
                    const std::vector<GeneString> & genomes,
                    TAlignment & alignment,
                    double eps,
                    bool asymmetric,
                    IloNumArray penaltyVals)
{
    size_t numGenomes = genomes.size();
    std::vector<std::vector<bool> >  covered(genomes.size());
    
    for (size_t i = 0; i < genomes.size(); ++i)
        covered[i].assign(seqan::length(genomes[i]), false);
    
    std::vector<std::vector<column> > columns;
    columns.resize(numGenomes);
    
    for (size_t i = 0; i < numGenomes; ++i)
    {
        for (size_t pos = 0; pos < seqan::length(genomes[i]); ++pos)
        {
            if (covered[i][pos])
                continue;
            
            columns[i].push_back(column());
            column & col = columns[i].back();
            
            col.type = column::UNDEF;
            col.alPos.assign(numGenomes, -1);
            col.alPos[i] = pos;
            col.consensus = genomes[i][pos];
            
            //Find all aligned genes in subsequent genomes
            for (size_t j = i + 1; j < numGenomes; ++j)
            {
                for (size_t posJ = 0; posJ < seqan::length(genomes[j]); ++posJ)
                {
                    if (cbData.pairGraphs[i][j].hasAlignEdge(pos, posJ))
                    {
                        size_t aId = cbData.pairGraphs[i][j].vertexToIloId(pos, posJ);
                        if (vals.alignEdgesVals[i][j][aId] > eps)
                        {
                            col.alPos[j] = posJ;
                            covered[j][posJ] = true;
                            break;
                        }
                    }
                }
            }
            
            //get the column type
            
            //alignment column? (only if aligned Gene in all Genomes)
            size_t c = 0;
            for (; col.alPos[c] != MINUSONE && c < numGenomes; ++c);
            
            if (c == numGenomes)
                col.type = column::AL;
            
            if (col.type != column::UNDEF)
                continue;
            
            //loss column
            size_t numL = 0;
            for (size_t j = 0; j < numGenomes; ++j)
            {
                if (i == j)
                    continue;
                
                if (asymmetric && j == 0)
                    continue;
                
                if (vals.lossVals[i][j][pos] > eps)
                    ++numL;
            }
            
            if (numL > 0)
            {
                if (!asymmetric)
                    col.type = column::LOSSCONS; //appears in median
                
                else if (i == 0 && ( (penaltyVals[pos] > eps) || numL < numGenomes - 1)) //appears in median
                    col.type = column::LOSSCONS;
                
                else if (i == 0 && !(penaltyVals[pos] > eps) && numL == numGenomes - 1) //does not appear in median
                    col.type = column::LOSS;
            }
            
            if (col.type != column::UNDEF)
                continue;
            
            //duplication column
            if (!cbData.dupEvents[i].duplications.empty())
            {
                for (size_t dup = 0; dup < cbData.dupEvents[i].dupMap[pos].size(); ++dup)
                {
                    if (vals.duplicationVals[i][cbData.dupEvents[i].dupMap[pos][dup]] > eps)
                    {
                        col.type = column::DUP;
                        col.id = std::make_pair(i, cbData.dupEvents[i].dupMap[pos][dup]);
                        break;
                    }
                }
            }
            
            if (col.type != column::UNDEF)
                continue;
            
            //shared duplication
            if (asymmetric && i == 1)
            {
                const IloNumArray & pDupValsA = vals.pairedDuplicationVals[0];
                
                //covered by pairedduplication directly
                for (size_t d = 0; d < cbData.pairedDupEvents[0].dupMap[pos].size(); ++d)
                {
                    size_t dupId = cbData.pairedDupEvents[0].dupMap[pos][d];
                    if (pDupValsA[dupId] > eps)
                    {
                        col.id = std::make_pair(1, dupId);;
                        
                        if (numL == 0)
                            col.type = column::PAIRED;
                        else
                            col.type = column::DUPLOSS;
                        
                        break;
                    }
                }
                
                if (col.type != column::UNDEF)
                    continue;
                
                //covered via aligned paired dupl Gene
                if (col.alPos[2] != MINUSONE)
                {
                    size_t pos2 = col.alPos[2];
                    const IloNumArray & pDupValsB = vals.pairedDuplicationVals[1];
                    for (size_t d = 0; d < cbData.pairedDupEvents[1].dupMap[pos2].size(); ++d)
                    {
                        size_t dupId = cbData.pairedDupEvents[1].dupMap[pos2][d];
                        if (pDupValsB[dupId] > eps)
                        {
                            col.id = std::make_pair(2, dupId);
                            col.type = column::PAIRED;
                            break;
                        }
                    }
                }
            }
            else if (asymmetric && i == 2)
            {
                const IloNumArray & pDupValsB = vals.pairedDuplicationVals[1];
                for (size_t d = 0; d < cbData.pairedDupEvents[1].dupMap[pos].size(); ++d)
                {
                    size_t dupId = cbData.pairedDupEvents[1].dupMap[pos][d];
                    if (pDupValsB[dupId] > eps)
                    {
                        col.id = std::make_pair(2, dupId);
                        col.type = column::DUPLOSS;
                        break;
                    }
                }
            }
        }
    }
    
    seqan::clear(alignment);
    alignment.insert(alignment.end(), columns[0].begin(), columns[0].end());
    
    //now merge the remaining columns that had no position in
    for (size_t i = 1; i < numGenomes; ++i)
    {
        std::vector<column>::iterator itAl = alignment.begin();
        for (size_t posCol = 0; posCol < columns[i].size(); ++posCol)
        {
            while (itAl != alignment.end() && !(columns[i][posCol] < *itAl))
            {
                ++itAl;
            }
            itAl = alignment.insert(itAl, columns[i][posCol]);
        }
    }
}




void _printAlignment(const TAlignment & align, const std::vector<GeneString> & genomes)
{
    for (size_t i = 0; i < align.size(); ++i)
        _printColumn(align[i], genomes);
}

void getMedianGenome(const TAlignment & align, GeneString & median)
{
    seqan::clear(median);
    for (size_t i = 0; i < align.size(); ++i)
    {
        if (align[i].type != column::DUP && align[i].type != column::LOSS)
        {
            seqan::appendValue(median, align[i].consensus);
        }
    }
}

void getDistancesToMedian(const TAlignment & align,
                           std::vector<double> & dist,
                           bool asymmetric)
{
    size_t numGenomes = align[0].alPos.size();
    dist.assign(numGenomes, 0);
    
    vector< pair<size_t, size_t> > lastDup(numGenomes, make_pair(-1,-1));
    pair<size_t, size_t> lastPairedDup(-1,-1);
    
    
    TAlignment::const_iterator col;
    //size_t id = 0;
    for (col = align.begin(); col != align.end(); ++col)
    {
        if (col->type == column::DUP)
        {
            //find genome with duplication
            size_t g = 0;
            for (; col->alPos[g] == MINUSONE; ++g);
            
            if (col->id != lastDup[g])
            {
                dist[g] += 1;
                lastDup[g] = col->id;
            }
        }
        
        if (col->type == column::PAIRED && col->id != lastPairedDup)
        {
            dist[0] += 1;
            lastPairedDup = col->id;
        }
        
        if (col->type == column::DUPLOSS)
        {
            if (col->id != lastPairedDup)
                dist[0] += 1;
            if (col->alPos[1] == MINUSONE)
                dist[1] += 1;
            if (col->alPos[2] == MINUSONE)
                dist[2] += 1;
            
            lastPairedDup = col->id;
        }
        
        if (col->type == column::LOSS)
        {
            if (!asymmetric)
            {
                for (size_t g = 0; g < col->alPos.size(); ++g)
                {
                    if (col->alPos[g] == MINUSONE)
                        dist[g] += 1;
                }
            }
            else
            {
                dist[0] += 1;
            }
        }
        
        if (col->type == column::LOSSCONS)
        {
            for (size_t g = 0; g < col->alPos.size(); ++g)
            {
                if (col->alPos[g] == MINUSONE)
                    dist[g] += 1;
            }
        }
        //
        //                if (col->alPos[0] == -1) //this is for the lossHack
        //                {
        //                    dist[0] += 1; //one for the duplication
        //                    for (size_t g = 1; g < col->alPos.size(); ++g)
        //                    {
        //                        if (col->alPos[g] == -1)
        //                            dist[g] += 1; //one for the subsequent loss
        //                    }
        //                }
        //                else
        //                {
        //                    bool LossInMedian = true;
        //                    for (size_t g = 1; g < col->alPos.size(); ++g)
        //                    {
        //                        LossInMedian = LossInMedian && (col->alPos[g] == -1);
        //                    }
        //
        //                    if (LossInMedian)
        //                        dist[0] += 1;
        //
        //                    else
        //                    {
        //                        for (size_t g = 1; g < col->alPos.size(); ++g)
        //                        {
        //                            if (col->alPos[g] == -1)
        //                                dist[g] += 1;
        //                        }
        //                    }
        //                }
        //            }
        //        }
        //cerr << dist[1] << "  " << id++ << endl;
    }
}

void _getProjectedInterval(size_t begin,
                           size_t end,
                           size_t & proj_begin,
                           size_t & proj_end,
                           const CallBackDataMedian2 & cbData,
                           const ValHolderMedian2 & vals,
                           size_t genS,
                           size_t genT,
                           double eps)
{
    //find the left_most and right most alignment edge
    bool rev = genS > genT;
    if (rev)
    {
        size_t tmp = genS;
        genS = genT;
        genT = tmp;
    }
    const SparsePairGraph & pg = cbData.pairGraphs[genS][genT];
    
    size_t left = -1, right = -1;
    size_t s1 = 0;
    for (size_t p = begin; p <= end; ++p)
    {
        size_t len = rev ? pg.strLenA() : pg.strLenB();
        for (size_t l1 = s1; l1 < len; ++l1)
        {
            size_t alId;
            if (!rev)
            {
                if (!pg.hasAlignEdge(p, l1))
                    continue;
                alId = pg.vertexToIloId(p, l1);
            }
            else
            {
                if (!pg.hasAlignEdge(l1, p))
                    continue;
                alId = pg.vertexToIloId(l1, p);
            }
            
            
            if (vals.alignEdgesVals[genS][genT][alId] >= 1 - eps)
            {
                s1 = l1 + 1;
                if (left == MINUSONE)
                    left = alId;
                else
                    right = alId;
                
                break;
            }
        }
    }
    proj_begin = left;
    proj_end = right;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//-------------------------CUT GENERATION FUNCTIONS-----------------------------
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//PAIRED DUPLICATION CUTS (ONLY FOR ASYMMETRIC MEDIAN)
// @DEPRECATED use _generatePairedDupCutsLifted
int _generatePairedDupCuts(const ValHolderMedian2 & vals,
                           const VarHolderMedian2 & vars,
                           CallBackDataMedian2 &cbData,
                           IloNum eps,
                           IloRangeArray &cuts)
{
    typedef std::vector<size_t>::const_iterator TdupMapIter;
    
    size_t cutsAdded = 0;
    
    const vector<IloNumArray> & pDupValsArr = vals.pairedDuplicationVals;
    const vector<IloNumVarArray> & pDupVarsArr = vars.pairedDuplicationVars;
    vector <PairedDuplicationEventHolder3> & pairedDups = cbData.pairedDupEvents;
    
    
    for (size_t s = 0; s < 2; ++s)
    {
        for (size_t pDupId = 0; pDupId < (size_t)pDupValsArr[s].getSize(); ++pDupId)
        {
            if (pDupValsArr[s][pDupId] < eps)
                continue;
            
            //find the left_most and right most alignment edge of the source region
            vector<size_t> leftS(2, -1), rightS(2, -1); //replace by numbers
            
            _getProjectedInterval(pairedDups[s].duplications[pDupId].sourceBegin,
                                  pairedDups[s].duplications[pDupId].sourceEnd,
                                  leftS[0], rightS[0], cbData, vals, 0, s+1, eps);
            
            //find the left_most and right most alignment edge of the target region
            size_t leftT = -1, rightT = -1;
            
            _getProjectedInterval(pairedDups[s].duplications[pDupId].targetBegin,
                                  pairedDups[s].duplications[pDupId].targetEnd,
                                  leftT, rightT, cbData, vals, s+1, 2 - s, eps);
            
            //check for both that these regions are not disrupted by any other duplication coming in from parent
            //TOO STRONG ASSUMPTION OF VISIBLE HISTORY
            //for source
            if (cbData.opt.sourceDisruptConstraints)
            {
                if (leftS[0] != MINUSONE && rightS[0] != MINUSONE)
                {
                    size_t lpos, rpos, tmp1, tmp2;
                    SparsePairGraph & pg = cbData.pairGraphs[0][s+1];
                    pg.iloIdToVertex(leftS[0], tmp1, lpos);
                    pg.iloIdToVertex(rightS[0], tmp2, rpos);
                    
                    float checkSum1 = vals.alignEdgesVals[0][s+1][leftS[0]]
                    + vals.alignEdgesVals[0][s+1][rightS[0]]
                    + pDupValsArr[s][pDupId];
                    
                    for (size_t pos = lpos + 1; pos < rpos; ++pos)
                    {
                        TdupMapIter dupItA, dupItEnd;
                        dupItA = pairedDups[s].dupMap[pos].begin();
                        dupItEnd = pairedDups[s].dupMap[pos].end();
                        
                        for (; dupItA != dupItEnd; ++dupItA)
                        {
                            if (checkSum1 + pDupValsArr[s][*dupItA] > 3 + eps)
                            {
                                //found violation --  add cut
                                IloEnv env = cuts.getEnv();
                                IloExpr cut_lhs(env);
                                cut_lhs = vars.alignEdgesVars[0][s+1][leftS[0]] + vars.alignEdgesVars[0][s+1][rightS[0]] +
                                pDupVarsArr[s][pDupId] + pDupVarsArr[s][*dupItA];
                                cerr << "source disrupt " << s+1 << " " << checkSum1 + pDupValsArr[s][*dupItA] << std::endl;
                                //                          cerr << tmp1 << " " << tmp2 << " ==> " << lpos << " " << rpos << " " << pos << endl;
                                cuts.add(cut_lhs <= 3);
                                cut_lhs.end();
                                ++cutsAdded;
                            }
                        }
                    }
                }
            }
            
            //for target
            if (leftT != MINUSONE && rightT != MINUSONE)
            {
                const SparsePairGraph & pg12 = cbData.pairGraphs[1][2];
                
                size_t lpos, rpos, tmp1, tmp2;
                if (s == 0)
                {
                    pg12.iloIdToVertex(leftT, tmp1, lpos);
                    pg12.iloIdToVertex(rightT, tmp2, rpos);
                    assert (rpos < pg12.strLenB());
                }
                else
                {
                    pg12.iloIdToVertex(leftT, lpos, tmp1);
                    pg12.iloIdToVertex(rightT, rpos, tmp2);
                    assert (rpos < pg12.strLenA());
                }
                
                float checkSum = vals.alignEdgesVals[1][2][leftT]
                + vals.alignEdgesVals[1][2][rightT]
                + pDupValsArr[s][pDupId];
                
                for (size_t pos = lpos + 1; pos < rpos; ++pos)
                {
                    //Cut interrupting duplications from Parent
                    if (cbData.opt.targetDisruptConstraints)
                    {
                        TdupMapIter dupIt, dupEnd;
                        dupIt = pairedDups[1-s].dupMap[pos].begin();
                        dupEnd = pairedDups[1-s].dupMap[pos].end();
                        
                        for (; dupIt != dupEnd; ++dupIt)
                        {
                            if (checkSum + pDupValsArr[1-s][*dupIt] > 3 + eps)
                            {
                                //found violation --  add cut
                                IloEnv env = cuts.getEnv();
                                IloExpr cut_lhs(env);
                                cut_lhs = vars.alignEdgesVars[1][2][leftT] + vars.alignEdgesVars[1][2][rightT] +
                                pDupVarsArr[s][pDupId] + pDupVarsArr[1-s][*dupIt];
                                cerr << "target disrupt Dup: " << checkSum + pDupValsArr[1-s][*dupIt] << std::endl;
                                //cerr << pairedDupsArr[s]->duplications[pDupId].targetBegin << " " << pairedDupsArr[s]->duplications[pDupId].targetEnd << " ==> " << lpos << " " << rpos << " " << pos << endl;
                                cuts.add(cut_lhs <= 3);
                                cut_lhs.end();
                                ++cutsAdded;
                            }
                        }
                    }
                    //Cut interrupting alignments from Parent
                    SparsePairGraph & pgTmp = cbData.pairGraphs[0][2-s];
                    for (size_t posPar = 0; posPar < pgTmp.strLenA(); ++posPar)
                    {
                        if (!pgTmp.hasAlignEdge(posPar, pos))
                            continue;
                        
                        size_t id = pgTmp.vertexToIloId(posPar, pos);
                        if (checkSum + vals.alignEdgesVals[0][2-s][id] > 3 + eps)
                        {
                            //found violation --  add cut
                            IloEnv env = cuts.getEnv();
                            IloExpr cut_lhs(env);
                            cut_lhs = vars.alignEdgesVars[1][2][leftT] + vars.alignEdgesVars[1][2][rightT] +
                            pDupVarsArr[s][pDupId] + vars.alignEdgesVars[0][2-s][id];
                            cuts.add(cut_lhs <= 3);
                            cut_lhs.end();
                            ++cutsAdded;
                            cerr << "target disrupt AL: " << s+1 << " " << checkSum + vals.alignEdgesVals[0][2-s][id] << std::endl;
                            cerr << pairedDups[s].duplications[pDupId].targetBegin << " " << pairedDups[s].duplications[pDupId].targetEnd <<  " " << tmp1 << " " << tmp2 << " ==> " <<lpos << " " << rpos << " :: " << pos << " " << id << endl;
                        }
                    }
                }
            }
        }
    }
    return cutsAdded;
}


//PAIRED DUPLICATION CUTS (ONLY FOR ASYMMETRIC MEDIAN)
int _generatePairedDupCutsLifted(const ValHolderMedian2 & vals,
                                 const VarHolderMedian2 & vars,
                                 CallBackDataMedian2 &cbData,
                                 IloNum eps,
                                 IloRangeArray &cuts)
{
    size_t cutsAdded = 0;
    
    const vector<IloNumArray> & pDupVals = vals.pairedDuplicationVals;
    const vector<IloNumVarArray> & pDupVars = vars.pairedDuplicationVars;
    
    //const IloNumArray * pDupVals[2] = {&vals.pairedDuplicationVals[0], &vals.pairedDuplicationVals[1]};
    //const IloNumVarArray * pDupVars[2] = {&vars.pairedDuplicationVars[0], &vars.pairedDuplicationVars[1]};
    vector<PairedDuplicationEventHolder3> & pDups = cbData.pairedDupEvents;
    pDups[0].DPCompute(pDupVals[0]);
    pDups[1].DPCompute(pDupVals[1]);
    
    cbData.pairGraphsAug[0][1].assignWeights(vals.alignEdgesVals[0][1]);
    cbData.pairGraphsAug[0][2].assignWeights(vals.alignEdgesVals[0][2]);
    cbData.pairGraphsAug[0][1].prepareIntervalQueries();
    cbData.pairGraphsAug[0][2].prepareIntervalQueries();
    
    //TOO STRONG ASSUMPTION OF VISIBLE HISTORY -- DEFAULT TURNED OFF
    // FIND DISRUPTED SOURCE CUTS
    if (cbData.opt.sourceDisruptConstraints)
    {
        for (size_t s = 0; s < 2; ++s)
        {
            //get incidence info for genes in 0 and alignment edges 0-1 and 0-2
            std::vector<TSizeArray> incIdsS;
            incIdsS.resize(cbData.pairGraphsAug[0][s+1].strLenA());
            for (size_t i = 0; i < (size_t)vals.alignEdgesVals[0][s+1].getSize(); ++i)
            {
                if (vals.alignEdgesVals[0][s+1][i] <= eps)
                    continue;
                
                size_t p1, p2;
                cbData.pairGraphsAug[0][s+1].iloIdToVertex(i, p1, p2);
                
                if (vals.alignEdgesVals[0][s+1][i] <= eps)
                    continue;
                
                incIdsS[p1].push_back(i);
            }
            
            
            for (size_t pDup = 0; pDup < (size_t)pDupVals[s].getSize(); ++pDup)
            {
                double coverVal = (pDupVals[s])[pDup];
                if (coverVal <= eps)
                    continue;
                
                size_t sb = pDups[s].duplications[pDup].sourceBegin;
                size_t se = pDups[s].duplications[pDup].sourceEnd;
                
                for (size_t l = sb; l < se; ++l)
                {
                    const TSizeArray & incL = incIdsS[l];
                    for (size_t alL = 0; alL < incL.size(); ++alL)
                    {
                        double alLVal = vals.alignEdgesVals[0][s+1][incL[alL]];
                        if (alLVal + coverVal < 1 + eps)
                            continue;
                        
                        for (size_t r = l+1; r <= se; ++r)
                        {
                            const TSizeArray & incR = incIdsS[r];
                            
                            for (size_t alR = 0; alR < incR.size(); ++alR)
                            {
                                double alRVal = vals.alignEdgesVals[0][s+1][incR[alR]];
                                if (alLVal + alRVal + coverVal <= 2 + eps)
                                    continue;
                                
                                //now check for paired duplication falling within the aligned region
                                size_t lPos, rPos, tmp;
                                cbData.pairGraphs[0][s+1].iloIdToVertex(incL[alL], tmp, lPos);
                                cbData.pairGraphs[0][s+1].iloIdToVertex(incR[alR], tmp, rPos);
                                
                                
                                size_t maxDisrPos = lPos + 1;
                                double maxDisrVal = 0.;
                                
                                for (size_t p = lPos + 1; p < rPos; ++p)
                                {
                                    if (pDups[s].sigV[p][p] > maxDisrVal)
                                    {
                                        maxDisrVal = pDups[s].sigV[p][p];
                                        maxDisrPos = p;
                                    }
                                }
                                
                                //check for violated cut and add it
                                double totScore = alLVal + alRVal + coverVal;
                                if (totScore + maxDisrVal > 3 + eps)
                                {
                                    IloEnv env = cuts.getEnv();
                                    IloExpr cut_lhs(env);
                                    
                                    cut_lhs += vars.alignEdgesVars[0][s+1][incL[alL]];
                                    cut_lhs += vars.alignEdgesVars[0][s+1][incR[alR]];
                                    cut_lhs += pDupVars[s][pDup];
                                    
                                    for (size_t ds = 0; ds < pDups[s].sigI[maxDisrPos][maxDisrPos].size(); ++ds)
                                    {
                                        cut_lhs += pDupVars[s][pDups[s].sigI[maxDisrPos][maxDisrPos][ds]];
#ifdef DEBUGDLC
                                        if (pDupVals[s][pDups[s].sigI[maxDisrPos][maxDisrPos][ds]] > eps)
                                            cerr << pDupVals[s][pDups[s].sigI[maxDisrPos][maxDisrPos][ds]] << endl;
#endif
                                    }
#ifdef DEBUGDLC
                                    cerr << "source disrupt lifted Dup: " << totScore + maxDisrVal << std::endl;
#endif
                                    cuts.add(cut_lhs <= 3);
                                    cut_lhs.end();
                                    ++cutsAdded;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    // FIND DISRUPTED TARGET CUTS
    
    //now get incidence info genes in 1 and 2 (alignment edges 1-2)
    //contains for every position the Ids of all incident edges
    std::vector<std::vector<TSizeArray> > incIds(2);
    incIds[0].resize(cbData.pairGraphsAug[1][2].strLenA());
    incIds[1].resize(cbData.pairGraphsAug[1][2].strLenB());
    
    for (size_t i = 0; i < (size_t)vals.alignEdgesVals[1][2].getSize(); ++i)
    {
        if (vals.alignEdgesVals[1][2][i] <= eps)
            continue;
        
        size_t p1, p2;
        cbData.pairGraphsAug[1][2].iloIdToVertex(i, p1, p2);
        incIds[0][p1].push_back(i);
        incIds[1][p2].push_back(i);
    }
    
    
    for (size_t s = 0; s < 2; ++s)
    {
        size_t len = cbData.pairGraphs[0][s+1].strLenB();
        
        for (size_t i = 0; i < len; ++i)
        {
            const TSizeArray & incIdsI = incIds[s][i];
            for (size_t j = i + 1; j < len; ++j)
            {
                double coverVal = pDups[s].sigV[i][j];
                
                if (coverVal < eps)
                    break;
                
                const TSizeArray & incIdsJ = incIds[s][j];
                
                //check alignment edges
                for (size_t adi = 0; adi < incIdsI.size(); ++adi)
                {
                    double alIVal = vals.alignEdgesVals[1][2][incIdsI[adi]];
                    if (alIVal + coverVal < 1 + eps)
                        break;
                    
                    for (size_t adj = 0; adj < incIdsJ.size(); ++adj)
                    {
                        double alJVal = vals.alignEdgesVals[1][2][incIdsJ[adj]];
                        if (alIVal + alJVal + coverVal < 2 + eps)
                            break;
                        
                        //get the target position of the alignment edges
                        size_t pai, paj, tmp1, tmp2;
                        if (s == 0)
                        {
                            //size_t tmp;
                            cbData.pairGraphsAug[1][2].iloIdToVertex(incIdsI[adi], tmp1, pai);
                            cbData.pairGraphsAug[1][2].iloIdToVertex(incIdsJ[adj], tmp2, paj);
                        }
                        else
                        {
                            //size_t tmp;
                            cbData.pairGraphsAug[1][2].iloIdToVertex(incIdsI[adi], pai, tmp1);
                            cbData.pairGraphsAug[1][2].iloIdToVertex(incIdsJ[adj], paj, tmp2);
                        }
                        
                        //check for interrupting duplications in other genome (2-s)
                        size_t maxDisrPos = pai + 1;
                        double maxDisrVal = 0.;
                        
                        if (cbData.opt.targetDisruptConstraints)
                        {
                            for (size_t p = pai + 1; p < paj; ++p)
                            {
                                if (pDups[1-s].sigV[p][p] > maxDisrVal)
                                {
                                    maxDisrVal = pDups[1-s].sigV[p][p];
                                    maxDisrPos = p;
                                }
                            }
                        }
                        
                        //check for interrupting alignment edge clique from parent to other genome (2-s)
                        double maxDisrAlVal = 0.;
                        std::vector<size_t> tmpClique, maxDisrClique;
                        
                        for (size_t p1 = pai + 1; p1 < paj; ++p1)
                        {
                            for (size_t p2 = p1; p2 < paj; ++p2)
                            {
                                double score = cbData.pairGraphsAug[0][2-s].getMaxIntervalClique(p1, p2, 1, tmpClique);
                                
                                if (score > maxDisrAlVal)
                                {
                                    maxDisrAlVal = score;
                                    maxDisrClique.swap(tmpClique);
                                }
                            }
                        }
                        
                        //check if violated cut and add it
                        double totScore = alIVal + alJVal + coverVal;
                        if (totScore + maxDisrVal > 3 + eps || totScore + maxDisrAlVal > 3 + eps)
                        {
                            IloEnv env = cuts.getEnv();
                            IloExpr cut_lhs(env);
                            cut_lhs += vars.alignEdgesVars[1][2][incIdsI[adi]];
                            cut_lhs += vars.alignEdgesVars[1][2][incIdsJ[adj]];
                            
                            for (size_t dp = 0; dp < pDups[s].sigI[i][j].size(); ++dp)
                                cut_lhs += pDupVars[s][pDups[s].sigI[i][j][dp]];
                            
                            if (cbData.opt.targetDisruptConstraints)
                            {
                                if (totScore + maxDisrVal > 3 + eps)
                                {
                                    IloExpr cutLhsDup(env);
                                    for (size_t ds = 0; ds < pDups[1-s].sigI[maxDisrPos][maxDisrPos].size(); ++ds)
                                        cutLhsDup += pDupVars[1-s][pDups[1-s].sigI[maxDisrPos][maxDisrPos][ds]];
#ifdef DEBUGDLC
                                    cerr << "target disrupt lifted Dup: " << totScore + maxDisrVal << std::endl;
#endif
                                    cuts.add(cut_lhs + cutLhsDup <= 3);
                                    cutLhsDup.end();
                                    ++cutsAdded;
                                }
                            }
                            
                            if (totScore + maxDisrAlVal > 3 + eps)
                            {
                                IloExpr cutLhsAl(env);
                                for (size_t al = 0; al < maxDisrClique.size(); ++al)
                                    cutLhsAl += vars.alignEdgesVars[0][2-s][maxDisrClique[al]];
#ifdef DEBUGDLC
                                cerr << "target disrupt lifted Al: " << totScore + maxDisrAlVal << std::endl;
#endif
                                cuts.add(cut_lhs + cutLhsAl <= 3);
                                cutLhsAl.end();
                                ++cutsAdded;
                            }
                        }
                    }
                }
            }
        }
    }
    return cutsAdded;
}


void _printSolution(const CallBackDataMedian2 &cbData, const ValHolderMedian2 & vals, double eps)
{
    //Print Alignment Edge Values
    cerr << "Alignment edges: " << endl;
    for (size_t i = 0; i < vals.alignEdgesVals.size(); ++i)
    {
        cerr << vals.alignEdgesVals.size() << "  " << vals.alignEdgesVals[i].size() << endl;
        for (size_t j = i+1; j < vals.alignEdgesVals[i].size(); ++j)
        {
            for (size_t e = 0; e < (size_t)vals.alignEdgesVals[i][j].getSize() ; ++e)
            {
                if (vals.alignEdgesVals[i][j][e] > eps)
                {
                    size_t pos1, pos2;
                    cbData.pairGraphs[i][j].iloIdToVertex(e, pos1, pos2);
                    
                    cerr << i << "," << j << " " << pos1 << " -> " << pos2 << " : " << vals.alignEdgesVals[i][j][e] << endl;
                }
            }
            cerr << " ----- " << endl << endl;
        }
    }
    
    cerr << "Duplications: " << endl;
    
    for (size_t i = 0; i < vals.duplicationVals.size(); ++i)
    {
        for (size_t j = 0; j < (size_t)vals.duplicationVals[i].getSize(); ++j)
        {
            if (vals.duplicationVals[i][j] > eps)
            {
                size_t sb = cbData.dupEvents[i].duplications[j].sourceBegin;
                size_t se = cbData.dupEvents[i].duplications[j].sourceEnd;
                size_t tb = cbData.dupEvents[i].duplications[j].targetBegin;
                size_t te = cbData.dupEvents[i].duplications[j].targetEnd;
                
                cerr << i << ": " << "[" << sb << "," << se << "] -> [" << tb << "," << te << "]  " << vals.duplicationVals[i][j] << endl;
            }
        }
        cerr << " ----- " << endl << endl;
    }
    
    cerr << "Shared Duplications: " << endl;
    
    for (size_t i = 0; i < (size_t)vals.pairedDuplicationVals[0].getSize(); ++i)
    {
        if (vals.pairedDuplicationVals[0][i] > eps)
        {
            size_t sb = cbData.pairedDupEvents[0].duplications[i].sourceBegin;
            size_t se = cbData.pairedDupEvents[0].duplications[i].sourceEnd;
            size_t tb = cbData.pairedDupEvents[0].duplications[i].targetBegin;
            size_t te = cbData.pairedDupEvents[0].duplications[i].targetEnd;
            
            cerr << 1 << ": " << "[" << sb << "," << se << "] -> [" << tb << "," << te << "]  " << vals.pairedDuplicationVals[0][i] <<endl;
        }
    }
    cerr << " ----- " << endl << endl;
    
    for (size_t i = 0; i < (size_t)vals.pairedDuplicationVals[1].getSize(); ++i)
    {
        if (vals.pairedDuplicationVals[1][i] > eps)
        {
            size_t sb = cbData.pairedDupEvents[1].duplications[i].sourceBegin;
            size_t se = cbData.pairedDupEvents[1].duplications[i].sourceEnd;
            size_t tb = cbData.pairedDupEvents[1].duplications[i].targetBegin;
            size_t te = cbData.pairedDupEvents[1].duplications[i].targetEnd;
            
            cerr << 2 << ": " << "[" << sb << "," << se << "] -> [" << tb << "," << te << "]  " << vals.pairedDuplicationVals[1][i] << endl;
        }
    }
    cerr << " ----- " << endl << endl;
    
    cerr << "Losses: " << endl;
    for (size_t i = 0; i < vals.lossVals.size(); ++i)
    {
        for (size_t j = 0; j < vals.lossVals[i].size(); ++j)
        {
            if (i == j)
                continue;
            for (size_t l = 0; l < (size_t)vals.lossVals[i][j].getSize(); ++l)
            {
                if (vals.lossVals[i][j][l] > eps)
                {
                    cerr << i << "," << j << " : " << l << "  " << vals.lossVals[i][j][l] << endl;
                }
            }
            cerr << " ----- " << endl << endl;
        }
    }
}

int _generateTransitivityCuts(const ValHolderMedian2 & vals,
                              const VarHolderMedian2 & vars,
                              CallBackDataMedian2 &cbData,
                              IloNum eps,
                              IloRangeArray &cuts)
{
    size_t cutsAdded = 0;
    
    for (size_t i = 0; i < cbData.pairGraphs[0][1].strLenA(); ++i)
    {
        for (size_t j = 0; j < cbData.pairGraphs[0][1].strLenB(); ++j)
        {
            if (!cbData.pairGraphs[0][1].hasAlignEdge(i, j))
                continue;
            
            for (size_t k = 0; k < cbData.pairGraphs[1][2].strLenB(); ++k)
            {
                if (!(cbData.pairGraphs[0][2].hasAlignEdge(i, k) && cbData.pairGraphs[1][2].hasAlignEdge(j, k)))
                    continue;
                
                //get IloIds
                size_t e1 = cbData.pairGraphs[0][1].vertexToIloId(i, j);
                size_t e2 = cbData.pairGraphs[1][2].vertexToIloId(j, k);
                size_t e3 = cbData.pairGraphs[0][2].vertexToIloId(i, k);
                
                if (vals.alignEdgesVals[0][1][e1] + vals.alignEdgesVals[1][2][e2] - vals.alignEdgesVals[0][2][e3] > 1 + eps)
                {
                    cuts.add(vars.alignEdgesVars[0][1][e1] + vars.alignEdgesVars[1][2][e2] - vars.alignEdgesVars[0][2][e3] <= 1);
#ifdef DEBUGDLC
                    cerr << "Not transitive 1: " << i << " " << j << " " << k << std::endl;
#endif
                    ++cutsAdded;
                }
                else if (vals.alignEdgesVals[0][1][e1] - vals.alignEdgesVals[1][2][e2] + vals.alignEdgesVals[0][2][e3] > 1 + eps)
                {
                    cuts.add(vars.alignEdgesVars[0][1][e1] - vars.alignEdgesVars[1][2][e2] + vars.alignEdgesVars[0][2][e3] <= 1);
#ifdef DEBUGDLC
                    cerr << "Not transitive 2: " << i << " " << j << " " << k << std::endl;
#endif
                    ++cutsAdded;
                }
                else if (-vals.alignEdgesVals[0][1][e1] + vals.alignEdgesVals[1][2][e2] + vals.alignEdgesVals[0][2][e3] > 1 + eps)
                {
                    cuts.add(-vars.alignEdgesVars[0][1][e1] + vars.alignEdgesVars[1][2][e2] + vars.alignEdgesVars[0][2][e3] <= 1);
#ifdef DEBUGDLC
                    cerr << "Not transitive 3: " << i << " " << j << " " << k << std::endl;
#endif
                    ++cutsAdded;
                }
            }
        }
    }
    return cutsAdded;
}

//CLIQUE CUTS WITH INVERSIONS(INDEPENDENT OF SYMMETRIC/ASYMMETRIC)
int _generateCliqueCuts(const IloNumArray &alignEdgesVal,
                        const IloNumArray &inversionsVal,
                        const IloNumVarArray &alignEdgesVar,
                        const IloNumVarArray &inversionsVar,
                        SparsePairGraph &pairGraph,
                        IloNum eps,
                        IloRangeArray &cuts)
{
    int cutsAdded = 0;
    //adding cut for violated clique if any
    SparsePairGraph::MixedClique clique;
    pairGraph.assignWeights(alignEdgesVal);
    pairGraph.findViolatedCliques(inversionsVal, clique, eps);
    
    IloExpr cut_lhs(alignEdgesVal.getEnv());
    for (size_t i = 0; i < clique.alnEdgeIds.size(); ++i)
    {
        cut_lhs += alignEdgesVar[clique.alnEdgeIds[i]];
#ifdef DEBUGDLC
        cerr << pairGraph.VertexToIloId[clique.alnEdgeIds[i]] << endl;
#endif
    }
    for (size_t i = 0; i < clique.inversionIds.size(); ++i)
    {
        cut_lhs += inversionsVar[clique.inversionIds[i]];
    }
    if (!(clique.alnEdgeIds.empty() && clique.inversionIds.empty()))
    {
        cuts.add(cut_lhs <= 1);
        ++cutsAdded;
#ifdef DEBUGDLC
        cerr << "clique cut" << endl;
#endif
    }
    return cutsAdded;
}


//CLIQUE CUTS WITHOUT INVERSIONS(INDEPENDENT OF SYMMETRIC/ASYMMETRIC)
int _generateCliqueCuts(const IloNumArray &alignEdgesVal,
                        const IloNumVarArray &alignEdgesVar,
                        SparsePairGraph &pairGraph,
                        IloNum eps,
                        IloRangeArray &cuts)
{
    IloNumVarArray tmpVar(alignEdgesVal.getEnv());
    IloNumArray tmpVal(alignEdgesVal.getEnv());
    return _generateCliqueCuts(alignEdgesVal, tmpVal, alignEdgesVar, tmpVar, pairGraph, eps, cuts);
}


//DUPLICATION CYCLE CUTS (FOR SYMMETRIC MEDIAN)
int _generatedupCycleCuts(const ValHolderMedian2 & vals,
                          const VarHolderMedian2 & vars,
                          CallBackDataMedian2 &cbData,
                          size_t gId,
                          IloNum eps,
                          bool lifted,
                          bool addOnlyBestCycle,
                          IloRangeArray &cuts)

{
    //cerr << "IN UNPAIRED DUPCYCLE CUT SEARCH" << std::endl;
    const IloNumArray & duplicationsVal = vals.duplicationVals[gId];
    const IloNumVarArray & duplicationsVar = vars.duplicationVars[gId];
    DuplicationEventHolder & holder = cbData.dupEvents[gId];
    IloEnv env = cuts.getEnv();
    
    DuplicationEventHolder::TviolatedDupList cyclesIds;
    int cutsAdded = 0;
    
    if (lifted)
        holder.getLiftedDuplicationCycle(duplicationsVal, cyclesIds, 1. - eps, !addOnlyBestCycle);
    else
        holder.getDuplicationCycle(duplicationsVal, cyclesIds, 1. - eps, !addOnlyBestCycle);
    
    size_t num = cyclesIds.size();
    
    for (size_t i = 0; i < num; ++i)
    {
        IloExpr cycleCut(env);
        for (size_t j = 0; j < cyclesIds[i].dupIds.size(); ++j)
        {
            cycleCut += duplicationsVar[cyclesIds[i].dupIds[j]];
        }
#ifdef DEBUGDLC
        cerr << cycleCut << endl;
#endif
        
        if (!cyclesIds[i].dupIds.empty())
        {
            cuts.add(cycleCut <= (IloInt) cyclesIds[i].cycleLen - 1);
#ifdef DEBUGDLC
            cerr << cyclesIds[i].score << " <= " << cyclesIds[i].cycleLen - 1 << endl;
#endif
            ++cutsAdded;
        }
    }
    return cutsAdded;
}

/* DEPRECATED
 //DUPLICATION CYCLE CUTS (FOR ASYMMETRIC MEDIAN)
 int _generatedupCycleCutsPaired(const valHolderMedian2 & vals,
 const varHolderMedian2 & vars,
 CallBackDataMedian2 &cbData,
 size_t gId,
 IloNum eps,
 bool lifted,
 bool addOnlyBestCycle,
 IloRangeArray &cuts)
 
 {
 //cerr << "IN PAIRED DUPCYCLE CUT SEARCH" << std::endl;
 //const IloNumArray & duplicationsVal = vals.duplicationVals[gId];
 const IloNumVarArray & duplicationsVar = vars.duplicationVars[gId];
 const IloNumArray & pairedDupsVal = vals.pairedDuplicationVals;
 const IloNumVarArray & pairedDupsVar = vars.pairedDuplicationVars;
 DuplicationEventHolder & holder = cbData.dupEvents[gId];
 PairedDuplicationEventHolder2 & pholder = cbData.pairedDupEvents;
 
 IloNumArray duplicationsVal(vals.duplicationVals[gId].copy()); //local copy
 
 //PUT PAIRED VALUES ONTO DUP VALS
 //    for (size_t i = 0; i < pholder.duplications.size(); ++i)
 //    {
 //        if (pairedDupsVal[i] > eps)
 //        {
 //            cerr << "acitve pair: " << i << "  " << pairedDupsVal[i] << std::endl;
 //            if (gId == 1)
 //            {
 //                duplicationsVal[pholder.duplications[i].dupA] += pairedDupsVal[i];
 //                std::cerr << "New value for " << pholder.duplications[i].dupA << " : " << duplicationsVal[pholder.duplications[i].dupA] << std::endl;
 //            }
 //            else if (gId == 2)
 //            {
 //                duplicationsVal[pholder.duplications[i].dupB] += pairedDupsVal[i];
 //                std::cerr << "New value for " << pholder.duplications[i].dupB << " : " << duplicationsVal[pholder.duplications[i].dupB] << std::endl;
 //            }
 //        }
 //    }
 
 IloEnv env = cuts.getEnv();
 
 DuplicationEventHolder::TviolatedDupList cyclesIds;
 int cutsAdded = 0;
 
 if (lifted)
 holder.getLiftedDuplicationCycle(duplicationsVal, cyclesIds, 1. - eps, !addOnlyBestCycle);
 else
 holder.getDuplicationCycle(duplicationsVal, cyclesIds, 1. - eps, !addOnlyBestCycle);
 
 size_t num = cyclesIds.size();
 //std::cerr << "Detected cycles: " << num << std::endl;
 
 for (size_t i = 0; i < num; ++i)
 {
 IloExpr cycleCut(env);
 for (size_t j = 0; j < cyclesIds[i].dupIds.size(); ++j)
 {
 size_t dupId = cyclesIds[i].dupIds[j];
 cycleCut += duplicationsVar[dupId];
 
 //            //add also corresponding pairedDupVariables
 //            if (gId == 1)
 //            {
 //                for (size_t k = 0; k < pholder.invMapA[dupId].size(); ++k)
 //                    cycleCut += pairedDupsVar[pholder.invMapA[dupId][k]];
 //            }
 //            else if (gId == 2)
 //            {
 //                for (size_t k = 0; k < pholder.invMapB[dupId].size(); ++k)
 //                    cycleCut += pairedDupsVar[pholder.invMapB[dupId][k]];
 //            }
 }
 cerr << cycleCut << endl;
 
 if (!cyclesIds[i].dupIds.empty())
 {
 cuts.add(cycleCut <= (IloInt) cyclesIds[i].cycleLen - 1);
 cerr << cyclesIds[i].score << " <= " << cyclesIds[i].cycleLen - 1 << endl;
 ++cutsAdded;
 }
 }
 return cutsAdded;
 }
 */

void _computeMirrors (const DuplicationEventHolder & holder, TCutIndexList & cut_ids)
{
    DuplicationEventHolder::TDuplicationMap mapS(holder.dupMap.size()), mapT(holder.dupMap.size());

    for (size_t i = 0; i < holder.dupMap.size(); ++i)
    {
        for (size_t j = 0; j < holder.dupMap[i].size(); ++j)
        {
            if (holder.dupMap[i][j] % 2 == 1)
            {
                mapS[i].push_back(holder.dupMap[i][j] - 1);
            }
            else
            {
                mapT[i].push_back(holder.dupMap[i][j]);
            }
        }
        std::sort(mapS[i].begin(), mapS[i].end());
        std::sort(mapT[i].begin(), mapT[i].end());
    }

    for (size_t i = 0; i < holder.dupMap.size(); ++i)
    {
        for (size_t j = i+1; j < holder.dupMap.size(); ++j)
        {
            std::vector<size_t> intersec(2 * std::min(mapS[i].size(), mapT[j].size()));
            std::vector<size_t>::iterator it = std::set_intersection(mapS[i].begin(), mapS[i].end(), mapT[j].begin(), mapT[j].end(), intersec.begin());
            size_t offset = it - intersec.begin();

            //std::cerr << "offset: " << offset << "  " << mapS[i].size() << "  " << mapT[j].size() << std::endl;

            if (offset == 0)
                continue;

            for (size_t k = 0; k < offset; ++k)
            {
                intersec[k+offset] = intersec[k] + 1;
            }

            //            for (size_t k = 0; k < 2 * offset; ++k)
            //            {
            //                std::cerr << intersec[k] << std::endl;
            //            }
            intersec.resize(2 * offset);
            cut_ids.resize(cut_ids.size() + 1);
            std::sort(intersec.begin(), intersec.end());
            cut_ids.back().swap(intersec);
            //cut_ids.back() = intersec;
        }
    }

    //filter duplicates and subsets
    std::vector<size_t> intersec;
    TCutIndexList cut_ids_tmp;
    for (size_t i = 0; i < cut_ids.size(); ++i)
    {
        bool kickOut = false;
        for (size_t j = 0; j < cut_ids.size(); ++j)
        {
            if (i == j || cut_ids[i].size() > cut_ids[j].size())
                continue;

            intersec.resize(cut_ids[i].size());
            std::vector<size_t>::iterator it = std::set_intersection(cut_ids[i].begin(), cut_ids[i].end(), cut_ids[j].begin(), cut_ids[j].end(), intersec.begin());

            if(it == intersec.end())
            {
                kickOut = true;
                break;
            }
        }

        if (!kickOut)
            cut_ids_tmp.push_back(cut_ids[i]);
    }
    cut_ids.swap(cut_ids_tmp);
}



//MIXED CLIQUE CUTS (FOR SYMMETRIC MEDIAN)
int _generateMixedCliqueCuts(const ValHolderMedian2 & vals,
                             const VarHolderMedian2 & vars,
                             CallBackDataMedian2 &cbData,
                             size_t g1,
                             size_t g2,
                             IloNum eps,
                             IloRangeArray &cuts,
                             size_t id = 2)
{
    const IloNumArray & alignEdgesValXY = vals.alignEdgesVals[g1][g2];
    const IloNumArray & inversionsValXY = vals.inversionVals[g1][g2];
    const IloNumArray &duplicationsValX = vals.duplicationVals[g1];
    const IloNumArray &duplicationsValY = vals.duplicationVals[g2];
    const IloNumVarArray & alignEdgesVarXY = vars.alignEdgesVars[g1][g2];
    const IloNumVarArray & inversionsVarXY = vars.inversionVars[g1][g2];
    const IloNumVarArray &duplicationsVarX = vars.duplicationVars[g1];
    const IloNumVarArray &duplicationsVarY = vars.duplicationVars[g2];
    SparsePairGraph & pairGraphAug = cbData.pairGraphsAug[g1][g2];
    
    DuplicationEventHolder & holderX = cbData.dupEvents[g1];
    DuplicationEventHolder & holderY = cbData.dupEvents[g2];
    
    int cutsAdded = 0;
    IloEnv env = cuts.getEnv();
    
    SparsePairGraph::TMixedCliqueList cliques;
    pairGraphAug.assignWeights(alignEdgesValXY);
    
    if (id == 0 || id == 2)
        pairGraphAug.getMixedCliques(holderX, inversionsValXY, duplicationsValX, cliques, eps, 0);
    if (id == 1 || id == 2)
        pairGraphAug.getMixedCliques(holderY, inversionsValXY, duplicationsValY, cliques, eps, 1);
    
    
    for (size_t i = 0; i < cliques.size(); ++i)
    {
        IloExpr cutM_lhs(env);
        for (size_t j = 0; j < cliques[i].alnEdgeIds.size(); ++j)
        {
            cutM_lhs += alignEdgesVarXY[cliques[i].alnEdgeIds[j]];
#ifdef DEBUGDLC
            cerr << "align:  " <<  cliques[i].alnEdgeIds[j] << endl;
#endif
        }
        
        for (size_t j = 0; j < cliques[i].inversionIds.size(); ++j)
        {
            cutM_lhs += inversionsVarXY[cliques[i].inversionIds[j]];
        }
        
        for (size_t j = 0; j < cliques[i].dupIds.size(); ++j)
        {
            if (cliques[i].seq == 0)
                cutM_lhs += duplicationsVarX[cliques[i].dupIds[j]];
            else
                cutM_lhs += duplicationsVarY[cliques[i].dupIds[j]];
#ifdef DEBUGDLC
            cerr << (int)cliques[i].seq << "  " << cliques[i].dupIds[j] <<  endl;
#endif
        }
        cuts.add(cutM_lhs <= 1);
        ++cutsAdded;
    }
    return cutsAdded;
}


//MIXED CLIQUE CUTS (FOR ASYMMETRIC MEDIAN)
int _generateMixedCliqueCutsPaired(const ValHolderMedian2 & vals,
                                   const VarHolderMedian2 & vars,
                                   CallBackDataMedian2 &cbData,
                                   size_t g1,
                                   IloNum eps,
                                   IloRangeArray &cuts)
{
    const IloNumArray & alignEdgesVal = vals.alignEdgesVals[0][g1];
    const IloNumArray & inversionsVal = vals.inversionVals[0][g1];
    const IloNumVarArray & pairedDupsVar = vars.pairedDuplicationVars[g1 - 1];
    const IloNumArray & pairedDupsVal = vals.pairedDuplicationVals[g1 - 1];
    const IloNumArray & duplicationsVal = vals.duplicationVals[g1];
    PairedDuplicationEventHolder3 & pholder = cbData.pairedDupEvents[g1 - 1];
    
    const IloNumVarArray & alignEdgesVar = vars.alignEdgesVars[0][g1];
    const IloNumVarArray & inversionsVar = vars.inversionVars[0][g1];
    const IloNumVarArray & duplicationsVar = vars.duplicationVars[g1];
    
    SparsePairGraph & pairGraphAug = cbData.pairGraphsAug[0][g1];
    DuplicationEventHolder & holder = cbData.dupEvents[g1];
    
    int cutsAdded = 0;
    IloEnv env = cuts.getEnv();
    
    SparsePairGraph::TMixedCliqueList cliques;
    pairGraphAug.assignWeights(alignEdgesVal);
    pairGraphAug.getMixedCliquesPaired(pholder, holder, pairedDupsVal, duplicationsVal, cliques, eps);
    
    
    for (size_t i = 0; i < cliques.size(); ++i)
    {
        float score_debug = 0.;
        
        IloExpr cutM_lhs(env);
        float tmp3 = 0;
        for (size_t j = 0; j < cliques[i].alnEdgeIds.size(); ++j)
        {
            score_debug += alignEdgesVal[cliques[i].alnEdgeIds[j]];
            tmp3 += alignEdgesVal[cliques[i].alnEdgeIds[j]];
            cutM_lhs += alignEdgesVar[cliques[i].alnEdgeIds[j]];
#ifdef DEBUGDLC
            cerr << "align:  " <<  cliques[i].alnEdgeIds[j] << endl;
#endif
        }
        
        
        for (size_t j = 0; j < cliques[i].inversionIds.size(); ++j)
        {
            assert (1 == 0);
            score_debug += inversionsVal[cliques[i].inversionIds[j]];
            cutM_lhs += inversionsVar[cliques[i].inversionIds[j]];
        }
        
        float tmp1 = 0;
        for (size_t j = 0; j < cliques[i].dupIds.size(); ++j)
        {
            size_t dupId = cliques[i].dupIds[j];
            score_debug += duplicationsVal[dupId];
            tmp1 += duplicationsVal[dupId];
            cutM_lhs += duplicationsVar[dupId];
        }
        
        
        float tmp2 = 0;
        for (size_t j = 0; j < cliques[i].dupIdsP.size(); ++j)
        {
            size_t dupId = cliques[i].dupIdsP[j];
            tmp2 += pairedDupsVal[dupId];
            score_debug += pairedDupsVal[dupId];
            cutM_lhs += pairedDupsVar[dupId];
        }
#ifdef DEBUGDLC
        std::cerr << score_debug << " " << cliques[i].score << std::endl;
#endif
        assert(fabs(score_debug - cliques[i].score) < 0.01);
        
        cuts.add(cutM_lhs <= 1);
        ++cutsAdded;
    }
    return cutsAdded;
}


//helper to reduce code duplication in lazy and cut callback
int _commonUserCutStepsMedian(CallBackDataMedian2 & cbData,
                              const VarHolderMedian2 &vars,
                              const ValHolderMedian2 &vals,
                              IloNum eps,
                              IloRangeArray &cuts,
                              bool liftedDupCycles = true,
                              bool addOnlyBestCycle = true,
                              bool asymmetric = false,
                              bool stopAtFirstCut = true)
{
    int cutsAdded = 0;
    
#ifndef PAIRWISECONFLICT
    //SIMPLE CLIQUE CUTS
    for (size_t i = 0; i < cbData.pairGraphs.size(); ++i)
    {
        for (size_t j = i+1; j < cbData.pairGraphs[i].size(); ++j)
        {
            int cutsAddedC = _generateCliqueCuts(vals.alignEdgesVals[i][j],
                                                 vals.inversionVals[i][j],
                                                 vars.alignEdgesVars[i][j],
                                                 vars.inversionVars[i][j],
                                                 cbData.pairGraphs[i][j],
                                                 eps,
                                                 cuts);
            cutsAdded += cutsAddedC;
#ifdef DEBUGILP
            std::cerr << "Clique Cuts: " << cutsAddedC << std::endl;
#endif
        }
    }
#endif //PAIRWISECONFLICT
    
    if (cutsAdded && stopAtFirstCut)
        return cutsAdded;
    
    
    //TRANSITIVTY CUTS
    //if (asymmetric)
    //    cutsAdded += _generateTransitivityCuts(vals, vars, cbData, eps, cuts);
    
    //DUPLICATION CYCLE CUTS
    for (size_t i = 0; i < cbData.dupEvents.size(); ++i)
    {
        if (!asymmetric || i > 0)
        {
            int cutsAddedD = _generatedupCycleCuts(vals, vars, cbData, i, eps, liftedDupCycles, addOnlyBestCycle, cuts);
            cutsAdded += cutsAddedD;
#ifdef DEBUGILP
            std::cerr << "Dup Cycle Cuts: " << cutsAddedD << std::endl;
#endif
        }
    }
    
    if (cutsAdded && stopAtFirstCut)
        return cutsAdded;
    
    //TOO STRONG ASSUMPTION OF VISIBLE HISTORY
    //PENALTY VARIABLE CUTS
    
    //This penalty is required to force a visible history.
    //If some source of a shared duplication in Parent is lost in A and B
    //this must be explained by two losses in A and B separately
    if (asymmetric && cbData.opt.sourceDisruptConstraints)
    {
        int cutsAddedPen = 0;
        for (size_t i = 0; i < cbData.pairedDupEvents[0].dupMapS.size(); ++i)
        {
            for (size_t d = 0; d < cbData.pairedDupEvents[0].dupMapS[i].size(); ++d)
            {
                size_t dupId = cbData.pairedDupEvents[0].dupMapS[i][d];
                if (vals.pairedDuplicationVals[0][dupId] + vals.lossVals[0][1][i] + vals.lossVals[0][2][i] - 2 > vals.penaltyVals[0][i] + eps)
                {
                    //cuts.add(vars.pairedDuplicationVars[0][dupId] + vars.lossVars[0][1][i] - 1 - vars.penaltyVars[0][i] <= 0);@TODO undo after loss check
                    cuts.add(vars.pairedDuplicationVars[0][dupId] + vars.lossVars[0][1][i] + vars.lossVars[0][2][i] - 2 - vars.penaltyVars[0][i] <= 0);
                    ++cutsAddedPen;
                    ++cutsAdded;
                }
            }
        }
        
        
        for (size_t i = 0; i < cbData.pairedDupEvents[1].dupMapS.size(); ++i)
        {
            for (size_t d = 0; d < cbData.pairedDupEvents[1].dupMapS[i].size(); ++d)
            {
                size_t dupId = cbData.pairedDupEvents[1].dupMapS[i][d];
                if (vals.pairedDuplicationVals[1][dupId] + vals.lossVals[0][1][i] + vals.lossVals[0][2][i] - 2 > vals.penaltyVals[0][i] + eps)
                {
                    //cuts.add(vars.pairedDuplicationVars[1][dupId] + vars.lossVars[0][1][i] - 1 - vars.penaltyVars[0][i] <= 0);@TODO undo after loss check
                    cuts.add(vars.pairedDuplicationVars[1][dupId] + vars.lossVars[0][1][i] + vars.lossVars[0][2][i] - 2 - vars.penaltyVars[0][i] <= 0);
                    ++cutsAddedPen;
                    ++cutsAdded;
                }
            }
        }
#ifdef DEBUGILP
        std::cerr << "cuts added pen: " << cutsAddedPen << std::endl;
#endif
    }
    
    
    if (cutsAdded && stopAtFirstCut)
        return cutsAdded;
    
    
#ifndef PAIRWISECONFLICT
    //MIXED CLIQUES CUTS
    for (size_t i = 0; i < cbData.pairGraphs.size(); ++i)
    {
        for (size_t j = i+1; j < cbData.pairGraphs[i].size(); ++j)
        {
            int cutsAddedM = 0;
            if (asymmetric && i == 0)
            {
                cutsAddedM = _generateMixedCliqueCutsPaired(vals, vars, cbData, j, eps, cuts);
                IloEnv env = cuts.getEnv();
                IloRangeArray cutsDebug(env);
                int cutsAddedDebug = _generateMixedCliqueCuts(vals, vars, cbData, i, j, eps, cuts, 1);
                assert((bool) cutsAddedDebug <= (bool) cutsAddedM);
#ifdef DEBUGILP
                std::cerr << "Mixed Clique Cuts Paired: " << cutsAddedM << std::endl;
#endif
            }
            else
            {
                cutsAddedM += _generateMixedCliqueCuts(vals, vars, cbData, i, j, eps, cuts);
#ifdef DEBUGILP
                std::cerr << "Mixed Clique Cuts: " << cutsAddedM << std::endl;
#endif
            }
            cutsAdded += cutsAddedM;
        }
    }
#endif //PAIRWISECONFLICT
    
    if (cutsAdded && stopAtFirstCut)
        return cutsAdded;
    
    //MIXED CYCLE CUTS
    if (cbData.pairGraphs.size() > 2)
    {
        MixedCycleFinder::TMixedCycle cycle;
        size_t length = -1;
        
        cbData.mcFinder.findViolatedMixedCycle(vals.alignEdgesVals, cycle, length, eps);
        
        if (!cycle.empty())
        {
            IloEnv env = cuts.getEnv();
            IloExpr mCycleCut(env);
            double tmp = 0;
            for(size_t i = 0; i < cycle.size(); ++i)
            {
#ifdef DEBUGILP
                cerr << cycle[i].seqA << " " << cycle[i].seqB << " " << cycle[i].index << endl;
#endif
                mCycleCut += vars.alignEdgesVars[cycle[i].seqA][cycle[i].seqB][cycle[i].index];
                tmp += vals.alignEdgesVals[cycle[i].seqA][cycle[i].seqB][cycle[i].index];
            }
            //std::cerr << "add: " << tmp << " <= " << length - 1 << std::endl;
            cuts.add(mCycleCut <= (IloInt) length - 1);
            mCycleCut.end();
            ++cutsAdded;
#ifdef DEBUGILP
            std::cerr << "Mixed Cycle Cuts: " << 1 << std::endl;
#endif
        }
        
        if (cutsAdded && stopAtFirstCut)
            return cutsAdded;
        
        //PAIRED DUPLICATION CUTS
        if (asymmetric)
        {
            int cutsAddedPLift = _generatePairedDupCutsLifted(vals, vars, cbData, eps, cuts);
            cutsAdded += cutsAddedPLift;
#ifdef DEBUGILP
            std::cerr << "Paired Dup Cuts Lifted: " << cutsAddedPLift << std::endl;
#endif
            
#ifdef DEBUGCLP
            IloEnv env = cuts.getEnv();
            IloRangeArray cutsDebug(env);
            int cutsAddedP = _generatePairedDupCuts(vals, vars, cbData, eps, cutsDebug);
            std::cerr << "Paired Dup Cuts: " << cutsAddedP << std::endl;
            assert((bool)cutsAddedPLift >= (bool)cutsAddedP);
#endif
        }
    }
    return cutsAdded;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//---------------------------CALLBACK DEFINITIONS-------------------------------
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


ILOLAZYCONSTRAINTCALLBACK5(LazyCallback,
                           CallBackDataMedian2 & , cbData,
                           VarHolderMedian2 &, vars,
                           ValHolderMedian2 &, vals,
                           IloNum, eps,
                           bool, asymmetric)
{
#ifdef DEBUGILP
    std::cerr << "LAZY CallBack" << endl << endl;
#endif
    
    IloEnv env = getEnv();
    IloRangeArray cuts(env);
    
    //extract vals from cplex vars
    for (size_t i = 0; i < cbData.pairGraphs.size(); ++i)
    {
        getValues(vals.duplicationVals[i], vars.duplicationVars[i]);
        for (size_t j = 0; j < cbData.pairGraphs[i].size(); ++j)
        {
            if (i == j)
                continue;
            
            getValues(vals.lossVals[i][j], vars.lossVars[i][j]);
            if (i < j)
            {
                getValues(vals.alignEdgesVals[i][j], vars.alignEdgesVars[i][j]);
                getValues(vals.inversionVals[i][j], vars.inversionVars[i][j]);
            }
        }
    }
    
    if(asymmetric)
    {
        getValues(vals.pairedDuplicationVals[0], vars.pairedDuplicationVars[0]);
        getValues(vals.pairedDuplicationVals[1], vars.pairedDuplicationVars[1]);
        getValues(vals.wVals[0], vars.wVars[0]);
        getValues(vals.wVals[1], vars.wVars[1]);
        
        //for (size_t i = 0; i < vars.penaltyVars.size(); ++i)
        //{
        getValues(vals.penaltyVals[0], vars.penaltyVars[0]);
        //}
    }
    
    _commonUserCutStepsMedian(cbData, vars, vals, eps, cuts, cbData.opt.liftedDupCycles, cbData.opt.addOnlyBestDupCycle, asymmetric);
    
    try
    {
        for (size_t i = 0; i < (size_t)cuts.getSize(); ++i)
            add(cuts[i]);
        
        cuts.end();
    }
    catch (...)
    {
        cuts.end();
        throw;
    }
}

ILOUSERCUTCALLBACK5(UserCallback,
                    CallBackDataMedian2 &, cbData,
                    VarHolderMedian2 &, vars,
                    ValHolderMedian2 &, vals,
                    IloNum, eps,
                    bool, asymmetric)
{
#ifdef DEBUGILP
    std::cerr << "Cut" << std::endl;
    cerr << "BEST: " << getBestObjValue() << "  ACTUAL: " << getObjValue() << endl;
#endif 
    
    IloEnv env = getEnv();
    IloRangeArray cuts(env);
    
    
    if (getObjValue() > getBestObjValue() + cbData.opt.maxBestCurrentNodeDelta)
    {
        return;
    }
    
    //stop because only integer coefficients we can stop here
    if (getObjValue() + 1 > getIncumbentObjValue() + eps)
    {
        return;
    }
    
    bool stopAfterFirstCut = true;
    
    if (cbData.lastImprovedScore + eps < getObjValue())
    {
        cbData.lastImprovedScore = getObjValue();
        cbData.lastImproveRound = 0;
    }
    else
    {
        ++(cbData.lastImproveRound);
    }
    
    if (cbData.lastImproveRound >= 10)
    {
        stopAfterFirstCut = false;
        cbData.lastImproveRound = 0;
    }
    
    
    //extract vals from cplex vars
    for (size_t i = 0; i < cbData.pairGraphs.size(); ++i)
    {
        getValues(vals.duplicationVals[i], vars.duplicationVars[i]);
        for (size_t j = 0; j < cbData.pairGraphs[i].size(); ++j)
        {
            if (i == j)
                continue;
            
            getValues(vals.lossVals[i][j], vars.lossVars[i][j]);
            if (i < j)
            {
                getValues(vals.alignEdgesVals[i][j], vars.alignEdgesVars[i][j]);
                getValues(vals.inversionVals[i][j], vars.inversionVars[i][j]);
            }
        }
    }
    if(asymmetric)
    {
        getValues(vals.pairedDuplicationVals[0], vars.pairedDuplicationVars[0]);
        getValues(vals.pairedDuplicationVals[1], vars.pairedDuplicationVars[1]);
        getValues(vals.wVals[0], vars.wVars[0]);
        getValues(vals.wVals[1], vars.wVars[1]);
        getValues(vals.penaltyVals[0], vars.penaltyVars[0]);
        
    }
    
    
    int nof_rows_added = _commonUserCutStepsMedian(cbData, vars, vals, eps, cuts, cbData.opt.liftedDupCycles, cbData.opt.addOnlyBestDupCycle, asymmetric, stopAfterFirstCut);
    
    try
    {
        for (size_t i = 0; i < (size_t)cuts.getSize(); ++i)
            add(cuts[i]);
        
        cuts.end();
    }
    catch (...)
    {
        cuts.end();
        throw;
    }
    
    //cut subtours
    if (cbData.pairGraphs.size() == 2) //actually we only do this for two sequences
    {
        const IloNumArray & alignEdgesVal = vals.alignEdgesVals[0][1];
        const IloNumArray & duplicationsValA = vals.duplicationVals[0];
        const IloNumArray & duplicationsValB = vals.duplicationVals[1];
        const IloNumArray & lossValA = vals.lossVals[0][1];
        const IloNumArray & lossValB = vals.lossVals[1][0];
        const IloNumVarArray & alignEdgesVar = vars.alignEdgesVars[0][1];
        const IloNumVarArray & duplicationsVarA = vars.duplicationVars[0];
        const IloNumVarArray & duplicationsVarB = vars.duplicationVars[1];
        const IloNumVarArray & lossVarA = vars.lossVars[0][1];
        const IloNumVarArray & lossVarB = vars.lossVars[1][0];
        
        DuplicationEventHolder & dupEventsA = cbData.dupEvents[0];
        DuplicationEventHolder & dupEventsB = cbData.dupEvents[1];
        SparsePairGraph & pairGraph = cbData.pairGraphs[0][1];
        
        ListDigraph gA;
        ListDigraph::Node* nodesA = new ListDigraph::Node[dupEventsA.dupMap.size()];
        for(size_t p = 0; p < dupEventsA.dupMap.size(); ++p)
        {
            nodesA[p] = gA.addNode();
        }
        std::set<ViolatedCut> cutsA;
        get_violated_cuts(dupEventsA, duplicationsValA, lossValA, alignEdgesVal, pairGraph, dupEventsB.dupMap.size(), eps, gA, nodesA, cutsA, true);
        
        ListDigraph gB;
        ListDigraph::Node* nodesB = new ListDigraph::Node[dupEventsB.dupMap.size()];
        for(size_t p = 0; p < dupEventsB.dupMap.size(); ++p)
        {
            nodesB[p] = gB.addNode();
        }
        std::set<ViolatedCut> cutsB;
        get_violated_cuts(dupEventsB, duplicationsValB, lossValB, alignEdgesVal, pairGraph, dupEventsA.dupMap.size(), eps, gB, nodesB, cutsB, false);
        
        bool* cut_char_vec = new bool[dupEventsA.dupMap.size()];
        for(set<ViolatedCut>::iterator it = cutsA.begin(); it != cutsA.end(); it++)
        {
            for(size_t p = 0; p < dupEventsA.dupMap.size(); p++)
                cut_char_vec[p] = true;
            
            IloExpr subtour_cut(env);
            
            for(size_t z = 0; z < (*it).node_idc.size(); z++) {
                int p = (*it).node_idc[z];
                cut_char_vec[p] = false;
                subtour_cut += lossVarA[p];
                
                for(size_t q = 0; q < dupEventsB.dupMap.size(); ++q)
                {
                    if( pairGraph.hasAlignEdge(p,q))
                    {
                        subtour_cut += alignEdgesVar[pairGraph.vertexToIloId(p, q)];
                    }
                }
            }
            
            for(size_t i = 0; i < dupEventsA.duplications.size(); ++i)
            {
                for (size_t j = dupEventsA.duplications[i].targetBegin, k=0; j <= dupEventsA.duplications[i].targetEnd; ++j, ++k)
                {
                    if(!cut_char_vec[j] && cut_char_vec[dupEventsA.duplications[i].sourceBegin+k])
                    {
                        subtour_cut += duplicationsVarA[i];
                        break;
                    }
                }
            }
            add(subtour_cut >= 1);
            subtour_cut.end();
            nof_rows_added++;
        }
        delete[] nodesA;
        delete[] cut_char_vec;
        
        
        cut_char_vec = new bool[dupEventsB.dupMap.size()];
        for(set<ViolatedCut>::iterator it = cutsB.begin(); it != cutsB.end(); it++)
        {
            for(size_t p = 0; p < dupEventsB.dupMap.size(); p++)
                cut_char_vec[p] = true;
            
            IloExpr subtour_cut(env);
            
            for(size_t z = 0; z < (*it).node_idc.size(); z++) {
                int p = (*it).node_idc[z];
                cut_char_vec[p] = false;
                subtour_cut += lossVarB[p];
                
                for(size_t q = 0; q < dupEventsA.dupMap.size(); ++q)
                {
                    if( pairGraph.hasAlignEdge(q, p))
                    {
                        subtour_cut += alignEdgesVar[pairGraph.vertexToIloId(q, p)];
                    }
                }
            }
            
            for(size_t i = 0; i < dupEventsB.duplications.size(); ++i)
            {
                for (size_t j = dupEventsB.duplications[i].targetBegin, k=0; j <= dupEventsB.duplications[i].targetEnd; ++j, ++k)
                {
                    if(!cut_char_vec[j] && cut_char_vec[dupEventsB.duplications[i].sourceBegin+k])
                    {
                        subtour_cut += duplicationsVarB[i];
                        break;
                    }
                }
            }
            add(subtour_cut >= 1);
            subtour_cut.end();
            nof_rows_added++;
        }
        delete[] nodesB;
        delete[] cut_char_vec;
    }

#ifdef DEBUGILP
    if (nof_rows_added)
    {
        cerr << "added rows: " << nof_rows_added << endl;
        return;
    }
#endif
}


ILOBRANCHCALLBACK0(BranchCallbackMedian)
{
    if (getObjValue() + 1 > getIncumbentObjValue() + 0.0001)
    {
#ifdef DEBUGILP
        std::cerr << "Prune" << std::endl;
#endif
        prune();
    }
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//---------------------------THE MAIN ILP FUNCTION------------------------------
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


double MedianILPThreeLeaf(const std::vector<GeneString> & genomes, TAlignment & align, const IlpOptions & opts)
{
    const bool inversions = opts.inversions;
    bool asymmetric = opts.asymmetric;
    
    double score = -1;
    
    if (genomes.size() == 2)
        asymmetric = false;
    
    if (genomes.size() > 3)
    {
        std::cerr << "Actually at most 3 Genomes are supported - exit";
        exit(1);
    }
    
    
    
    CallBackDataMedian2 cbData;
    cbData.lastImproveRound = 0;
    cbData.lastImprovedScore = -1000;
    
    //setting options
    cbData.opt = opts;
    
    //create alignment edges with the SparsePairGraph
    cbData.pairGraphs.resize(genomes.size());
    cbData.pairGraphsAug.resize(genomes.size());
    
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        cbData.pairGraphs[i].resize(genomes.size());
        cbData.pairGraphsAug[i].resize(genomes.size());
        
        for (size_t j = i+1; j < genomes.size(); ++j)
        {
            cbData.pairGraphs[i][j].createPG(genomes[i], genomes[j], inversions);
            cbData.pairGraphsAug[i][j].createAugmentedPG(genomes[i], genomes[j], inversions);
        }
    }
    
    //create Duplications for each sequence
    cbData.dupEvents.resize(genomes.size());
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        if (asymmetric && i == 0) //no duplication inside parent genome
            continue;

        cbData.dupEvents[i].createDuplicationEventsLCP(genomes[i], inversions);
#ifdef DEBUGILP
        std::cerr << "len genome " << i << ": " << seqan::length(genomes[i]) << std::endl;
        std::cerr << "Num Duplications " << i <<": " << cbData.dupEvents[i].duplications.size() << std::endl;
#endif
    }
    
    //create pairedDuplications if asymmetric
    if (asymmetric)
    {
        cbData.pairedDupEvents.resize(2);
        cbData.pairedDupEvents[0].createDuplicationEventsLCP(genomes[0], genomes[1]);
        cbData.pairedDupEvents[1].createDuplicationEventsLCP(genomes[0], genomes[2]);
#ifdef DEBUGILP
        std::cerr << "num shared duplications A: " << cbData.pairedDupEvents[0].duplications.size() << std::endl;
        std::cerr << "num shared duplications B: " << cbData.pairedDupEvents[1].duplications.size() << std::endl;
#endif
    }
    
    
#ifdef DEBUGCLP
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        cerr << "duplications for genome " << i << endl;
        for (size_t j = 0; j < cbData.dupEvents[i].duplications.size(); ++j)
        {
            cerr << "[" << cbData.dupEvents[i].duplications[j].sourceBegin << "," << cbData.dupEvents[i].duplications[j].sourceEnd << "]["
            << cbData.dupEvents[i].duplications[j].targetBegin << "," << cbData.dupEvents[i].duplications[j].targetEnd << "] " << endl;
        }
    }
#endif
    
    
    IloEnv env;
    IloModel model(env);
    
    VarHolderMedian2 vars;
    ValHolderMedian2 vals;
    
    //IloVar Arrays are the variables
    typedef vector<IloNumVarArray> TIloNumVarMatrix;
    vars.alignEdgesVars.assign(genomes.size(), TIloNumVarMatrix(genomes.size()));
    vars.lossVars.assign(genomes.size(), TIloNumVarMatrix(genomes.size()));
    vars.inversionVars.assign(genomes.size(),TIloNumVarMatrix(genomes.size()));
    vars.duplicationVars.resize(genomes.size());
    vars.penaltyVars.resize(genomes.size());
    
    if (asymmetric)
    {
        vars.pairedDuplicationVars.resize(2);
        vars.pairedDuplicationVars[0] = IloNumVarArray(env, cbData.pairedDupEvents[0].duplications.size(), 0, 1, ILOINT);
        vars.pairedDuplicationVars[1] = IloNumVarArray(env, cbData.pairedDupEvents[1].duplications.size(), 0, 1, ILOINT);
        vars.pairedDuplicationVars[0].setNames("pairedA");
        vars.pairedDuplicationVars[1].setNames("pairedB");
        
        vars.wVars.resize(2);
        vars.wVars[0] = IloNumVarArray(env, cbData.pairGraphs[1][2].numAlignEdges(), 0, 1, ILOINT);
        vars.wVars[1] = IloNumVarArray(env, cbData.pairGraphs[1][2].numAlignEdges(), 0, 1, ILOINT);
    }
    else
    {
        vars.pairedDuplicationVars.assign(2, IloNumVarArray(env));
        vars.wVars.assign(2, IloNumVarArray(env));
    }
    
    
    //IloNum Arrays are used in the callback - defined here to prevent repeated contruction
    typedef vector<IloNumArray> TIloNumMatrix;
    vals.alignEdgesVals.assign(genomes.size(),TIloNumMatrix(genomes.size()));
    vals.lossVals.assign(genomes.size(),TIloNumMatrix(genomes.size()));
    vals.inversionVals.assign(genomes.size(), TIloNumMatrix(genomes.size()));
    vals.duplicationVals.resize(genomes.size());
    vals.penaltyVals.resize(genomes.size());
    
    if (asymmetric)
    {
        vals.pairedDuplicationVals.resize(2);
        vals.pairedDuplicationVals[0] = IloNumArray(env, cbData.pairedDupEvents[0].duplications.size());
        vals.pairedDuplicationVals[1] = IloNumArray(env, cbData.pairedDupEvents[1].duplications.size());
        
        vals.wVals.resize(2);
        vals.wVals[0] = IloNumArray(env, cbData.pairGraphs[1][2].numAlignEdges());
        vals.wVals[1] = IloNumArray(env, cbData.pairGraphs[1][2].numAlignEdges());
    }
    else
    {
        vals.pairedDuplicationVals.assign(2,IloNumArray(env));
        vals.wVals.assign(2,IloNumArray(env));
    }
    
    
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        //create duplication var arrays
        size_t numDup = cbData.dupEvents[i].duplications.size();
        vars.duplicationVars[i] = IloNumVarArray(env, numDup, 0, 1, ILOINT);
        vals.duplicationVals[i] = IloNumArray(env, numDup);
        
        std::stringstream strstr;
        strstr << "dup_" << i;
        vars.duplicationVars[i].setNames(strstr.str().c_str());
        
        strstr.str("");
        strstr << "pen_" << i;
        vars.penaltyVars[i] = IloNumVarArray(env, seqan::length(genomes[i]), 0, 1, ILOINT);
        vals.penaltyVals[i] = IloNumArray(env, seqan::length(genomes[i]));
        vars.penaltyVars[i].setNames(strstr.str().c_str());
        
        //create alignment edge vars and loss vars
        for (size_t j = 0; j < genomes.size(); ++j)
        {
            if(i == j)
                continue;
            
            if (i < j)
            {
                //ALIGN
                size_t numAlign = cbData.pairGraphs[i][j].numAlignEdges();
                vars.alignEdgesVars[i][j] = IloNumVarArray(env, numAlign, 0, 1, ILOINT);
                vals.alignEdgesVals[i][j] = IloNumArray(env, numAlign);
                std::stringstream strstr;
                strstr << "al_" << i << "-" << j;
                vars.alignEdgesVars[i][j].setNames(strstr.str().c_str());
                
                //INVERSION
                size_t numInversions = cbData.pairGraphs[i][j].numInversions();
                vars.inversionVars[i][j] = IloNumVarArray(env, numInversions, 0, 1, ILOINT);
                vals.inversionVals[i][j] = IloNumArray(env, numInversions);
                strstr.str("");
                strstr << "inv_" << i << "-" << j;
                vars.inversionVars[i][j].setNames(strstr.str().c_str());
            }
            
            //LOSS
            //no losses of parent and only a single array of losses for parent namely vars.lossVars[0][1]
            //if(asymmetric && ( (j == 0) || ((i == 0) && (j == 2) ) ))
            if(asymmetric && ( (j == 0) /*|| ((i == 0) && (j == 2) ) */))//@TODO undo after loss check
            {
                //create empty handles
                vars.lossVars[i][j] = IloNumVarArray(env);
                vals.lossVals[i][j] = IloNumArray(env);
            }
            else
            {
                vars.lossVars[i][j] = IloNumVarArray(env, seqan::length(genomes[i]), 0, 1, ILOINT);
                vals.lossVals[i][j] = IloNumArray(env, seqan::length(genomes[i]));
                std::stringstream strstr;
                strstr << "loss_" << i << "-" << j;
                vars.lossVars[i][j].setNames(strstr.str().c_str());
            }
        }
    }
    
    //formulate ILP
    //objective Function
    double lossWeight = 1;
    
    //if (genomes.size() == 3 && !asymmetric) //@TODO undo after loss check
    if (genomes.size() == 3) //@TODO undo after loss check
        lossWeight = 0.5;
    
    IloExpr obj(env);
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        obj += IloSum(vars.duplicationVars[i]);
        
        //        if (genomes.size() == 3 && (!asymmetric || i == 0) ) //@TODO undo after loss check
        //            obj += IloSum(vars.penaltyVars[i]);
        
        if (genomes.size() == 3)
        {
            if (i == 0)
                obj += IloSum(vars.penaltyVars[i]);
            else
                obj += lossWeight * IloSum(vars.penaltyVars[i]);
        }
        
        for (size_t j = i+1; j < genomes.size(); ++j)
        {
            obj += lossWeight * IloSum(vars.lossVars[i][j]);
            obj += lossWeight * IloSum(vars.lossVars[j][i]);
            
            if (inversions)
            {
#ifdef DEBUGILP
                std::cerr << "INVERSIONS" << std::endl;
#endif
                obj += IloSum(vars.inversionVars[i][j]);
            }
        }
    }
    
    if (asymmetric)
    {
        obj += IloSum(vars.pairedDuplicationVars[0]);
        obj += IloSum(vars.pairedDuplicationVars[1]);
    }
    
    
    
    //obj += IloSum(sionVarArray);
    model.add(IloMinimize(env, obj));
    obj.end();
    
    
    //PENALTY CONSTRAINTS -- THESE ARE ONLY VALID IN CASE OF 3 GENOMES
    
    //This penalty makes a loss in two sequences score 2 while a loss in two genomes has cost 1
    //This is necessary since in both cases 2 loss variables are active
    //In case of asymmetric ILP a loss in both child genomes only scores 1 and no losses in parent are allowed
    //Therefore no penalty is required in case of asymmetric formulation
    /*if (genomes.size() == 3 && !asymmetric)
     {
     for (size_t i = 0; i < genomes.size(); ++i)
     {
     if (asymmetric && i == 0)
     continue;
     
     for (size_t j = 0; j < seqan::length(genomes[i]); ++j)
     {
     IloExpr constr(env);
     for (size_t k = 0; k < genomes.size(); ++k)
     {
     if (i == k)
     continue;
     
     constr += vars.lossVars[i][k][j];
     }
     model.add(vars.penaltyVars[i][j] >= constr - 1);
     constr.end();
     }
     }
     }*/
    
    //Penalty for separate losses. If a position G1 is covered by paired dupl from G0 and is Lost in G2 the without penalty we charge only 0.5
    if (genomes.size() == 3 && asymmetric)
    {
        size_t count = 0;
        for (size_t i = 0; i < seqan::length(genomes[1]); ++i)
        {
            IloExpr constr(env);
            for (size_t d = 0; d < cbData.pairedDupEvents[0].dupMap[i].size(); ++d)
            {
                size_t dupId = cbData.pairedDupEvents[0].dupMap[i][d];
                constr += vars.pairedDuplicationVars[0][dupId];
            }
            constr += vars.lossVars[1][2][i];
            model.add(vars.penaltyVars[1][i] >= constr - 1);
            ++count;
        }
        for (size_t i = 0; i < seqan::length(genomes[2]); ++i)
        {
            IloExpr constr(env);
            for (size_t d = 0; d < cbData.pairedDupEvents[1].dupMap[i].size(); ++d)
            {
                size_t dupId = cbData.pairedDupEvents[1].dupMap[i][d];
                constr += vars.pairedDuplicationVars[1][dupId];
            }
            constr += vars.lossVars[2][1][i];
            model.add(vars.penaltyVars[2][i] >= constr - 1);
            ++count;
        }
#ifdef DEBUGILP
        std::cerr << "Penalty constraints for split loss vars: " << count << std::endl;
#endif
    }
    
    
    try{
        //Coupling the w variables to combinations of alignment and duplications
        if (asymmetric)
        {
            size_t count = 0 ;
            for (size_t i = 0; i < seqan::length(genomes[1]); ++i)
            {
                for (size_t j = 0; j < seqan::length(genomes[2]); ++j)
                {
                    if (!cbData.pairGraphs[1][2].hasAlignEdge(i, j))
                        continue;
                    
                    size_t alId = cbData.pairGraphs[1][2].vertexToIloId(i,j);
                    
                    IloExpr dupsB(env);
                    //sum up all duplications from Parent that cover j in B
                    for (size_t d = 0; d < cbData.pairedDupEvents[1].dupMap[j].size(); ++d)
                    {
                        assert(j < cbData.pairedDupEvents[1].dupMap.size());
                        size_t dupId = cbData.pairedDupEvents[1].dupMap[j][d];
                        //std::cerr << dupId << std::endl;
                        assert(dupId < (size_t)vars.pairedDuplicationVars[1].getSize());
                        dupsB += vars.pairedDuplicationVars[1][dupId];
                    }
                    
                    model.add(vars.alignEdgesVars[1][2][alId] + dupsB - 1 <= vars.wVars[0][alId]);
                    model.add(vars.wVars[0][alId] <= dupsB);
                    model.add(vars.wVars[0][alId] <= vars.alignEdgesVars[1][2][alId]);
                    dupsB.end();
                    count += 3;
                    
                    IloExpr dupsA(env);
                    //sum up all duplications from Parent that cover j in B
                    for (size_t d = 0; d < cbData.pairedDupEvents[0].dupMap[i].size(); ++d)
                    {
                        size_t dupId = cbData.pairedDupEvents[0].dupMap[i][d];
                        assert(dupId < (size_t)vars.pairedDuplicationVars[0].getSize());
                        dupsA += vars.pairedDuplicationVars[0][dupId];
                    }
                    
                    model.add(vars.alignEdgesVars[1][2][alId] + dupsA - 1 <= vars.wVars[1][alId]);
                    model.add(vars.wVars[1][alId] <= dupsA);
                    model.add(vars.wVars[1][alId] <= vars.alignEdgesVars[1][2][alId]);
                    dupsA.end();
                    count += 3;
                }
            }
#ifdef DEBUGILP
            std::cerr << "W coupling constraints: " << count << std::endl;
#endif
        }
        
        
#ifdef PAIRWISECONFLICT
        //for test reasons add pairwise exclusion constraints
        for (size_t i = 0; i < genomes.size(); ++i)
        {
            for (size_t j = i+1; j < genomes.size(); ++j)
            {
                const SparsePairGraph & pg = cbData.pairGraphs[i][j];
                for (size_t edge1 = 0; edge1 < vars.alignEdgesVars[i][j].getSize(); ++edge1)
                {
                    std::pair<size_t, size_t> pairI = pg.verticesInv[pg.IloIdToVertex[edge1]];
                    
                    for (size_t edge2 = edge1 + 1; edge2 < vars.alignEdgesVars[i][j].getSize(); ++edge2)
                    {
                        std::pair<size_t, size_t> pairJ = pg.verticesInv[pg.IloIdToVertex[edge2]];
                        
                        if(! (((pairI.first < pairJ.first) && (pairI.second < pairJ.second)) ||
                              ((pairI.first > pairJ.first) && (pairI.second > pairJ.second))))
                        {
                            IloExpr constr(env);
                            constr += vars.alignEdgesVars[i][j][edge1] + vars.alignEdgesVars[i][j][edge2];
                            model.add(constr <= 1);
                            constr.end();
                        }
                    }
                }
            }
        }
        
        
        /*
         
         for (size_t i = 0; i < dupAVarArray.getSize(); i+=2)
         {
         model.add(dupAVarArray[i] + dupAVarArray[i+1] <= 1);
         }
         for (size_t i = 0; i < dupBVarArray.getSize(); i+=2)
         {
         model.add(dupBVarArray[i] + dupBVarArray[i+1] <= 1);
         }
         */
#endif
        
        
        //adding cover constraint for parent in asymmetric median
        if (asymmetric)
        {
            for (size_t posI = 0; posI < seqan::length(genomes[0]); ++posI)
            {
                IloExpr constrAl1(env), constrAl2(env);
                
                for (size_t posJ = 0; posJ < seqan::length(genomes[1]); ++posJ)
                {
                    if (cbData.pairGraphs[0][1].hasAlignEdge(posI, posJ))
                    {
                        constrAl1 += vars.alignEdgesVars[0][1][cbData.pairGraphs[0][1].vertexToIloId(posI,posJ)];
                    }
                }
                for (size_t posJ = 0; posJ < seqan::length(genomes[2]); ++posJ)
                {
                    if (cbData.pairGraphs[0][2].hasAlignEdge(posI, posJ))
                    {
                        constrAl2 += vars.alignEdgesVars[0][2][cbData.pairGraphs[0][2].vertexToIloId(posI,posJ)];
                    }
                }
                
                model.add(constrAl1 + vars.lossVars[0][1][posI] == 1);
                model.add(constrAl2+ vars.lossVars[0][2][posI] == 1);
                //                model.add(constrAl1 + constrAl2 + vars.lossVars[0][1][posI] >= 1); //@TODO undo after loss check
                //                model.add(constrAl1 + vars.lossVars[0][1][posI] <= 1);
                //                model.add(constrAl2 + vars.lossVars[0][1][posI] <= 1);
            }
        }
        
        
        for (size_t i = 0; i < genomes.size(); ++i)
        {
            if (asymmetric && i == 0)
                continue;
            
            for (size_t posI = 0; posI < seqan::length(genomes[i]); ++posI)
            {
                IloExpr constrDup(env);
                
                //either a normal duplication
                for (size_t dupi = 0; dupi < cbData.dupEvents[i].dupMap[posI].size(); ++dupi)
                {
                    constrDup += vars.duplicationVars[i][cbData.dupEvents[i].dupMap[posI][dupi]];
                }
                
                for (size_t j = 0; j < genomes.size(); ++j)
                {
                    if (i == j)
                        continue;
                    
                    IloExpr constr(env);
                    constr += constrDup;
                    
                    //or a duplication from A to X
                    if (asymmetric && j == 0)
                    {
                        if (i == 1)
                        {   //Duplication from Parent to X covering A
                            TSizeArray::const_iterator dupIt = cbData.pairedDupEvents[0].dupMap[posI].begin();
                            for (; dupIt != cbData.pairedDupEvents[0].dupMap[posI].end(); ++dupIt)
                            {
                                constr += vars.pairedDuplicationVars[0][*dupIt];
                            }
                            
                            for (size_t pos2 = 0; pos2 < seqan::length(genomes[2]); ++pos2)
                            {
                                if (cbData.pairGraphs[1][2].hasAlignEdge(posI, pos2))
                                    constr += vars.wVars[0][cbData.pairGraphs[1][2].vertexToIloId(posI, pos2)];
                            }
                        }
                        if (i == 2)
                        {   //Duplication from Parent to X covering B
                            TSizeArray::const_iterator dupIt = cbData.pairedDupEvents[1].dupMap[posI].begin();
                            for (; dupIt != cbData.pairedDupEvents[1].dupMap[posI].end(); ++dupIt)
                            {
                                constr += vars.pairedDuplicationVars[1][*dupIt];
                            }
                            
                            for (size_t pos1 = 0; pos1 < seqan::length(genomes[1]); ++pos1)
                            {
                                if (cbData.pairGraphs[1][2].hasAlignEdge(pos1, posI))
                                    constr += vars.wVars[1][cbData.pairGraphs[1][2].vertexToIloId(pos1, posI)];
                            }
                        }
                    }
                    
                    //or inversion
                    if (inversions)
                    {
                        if (i < j)
                        {
                            for (size_t inv = 0; inv < cbData.pairGraphs[i][j].getInversionHolder().invMapA[posI].size(); ++inv)
                            {
                                constr += vars.inversionVars[i][j][cbData.pairGraphs[i][j].getInversionHolder().invMapA[posI][inv]];
                            }
                        }
                        else
                        {
                            for (size_t inv = 0; inv < cbData.pairGraphs[j][i].getInversionHolder().invMapB[posI].size(); ++inv)
                            {
                                constr += vars.inversionVars[j][i][cbData.pairGraphs[j][i].getInversionHolder().invMapB[posI][inv]];
                            }
                        }
                    }
                    
                    //or a loss
                    if (!asymmetric || j != 0)
                        constr += vars.lossVars[i][j][posI];
                    
                    //or aligned
                    for (size_t posJ = 0; posJ < seqan::length(genomes[j]); ++posJ)
                    {
                        if (i < j)
                        {
                            if (cbData.pairGraphs[i][j].hasAlignEdge(posI, posJ))
                            {
                                constr += vars.alignEdgesVars[i][j][cbData.pairGraphs[i][j].vertexToIloId(posI,posJ)];
                            }
                        }
                        else
                        {
                            if (cbData.pairGraphs[j][i].hasAlignEdge(posJ, posI))
                            {
                                constr += vars.alignEdgesVars[j][i][cbData.pairGraphs[j][i].vertexToIloId(posJ, posI)];
                            }
                        }
                    }
                    
                    model.add(constr == 1);
                    constr.end();
                }
            }
        }
    }
    catch (IloException& e) {
        e.print(cerr);
        throw;
    }
    
    
    //#ifndef PAIRWISECONFLICT
    
    //adding mirror constraints
    for (size_t i = 0; i < genomes.size(); ++i)
    {
        TCutIndexList ids;
        _computeMirrors(cbData.dupEvents[i], ids);
#ifdef DEBUGILP
        cerr << "adding "<< ids.size() <<"  mirror constraints " << i <<  endl;
#endif
        
        for (size_t id = 0; id < ids.size(); ++id)
        {
            IloExpr constr(env);
            for (size_t j = 0; j < ids[id].size(); ++j)
            {
                constr += vars.duplicationVars[i][ids[id][j]];
            }
            model.add(constr <= 1);
            constr.end();
        }
    }
 
    //adding "gene id constraints"
    //for every gene id at least one copy must be a loss or aligned
    if (!asymmetric && !inversions)
    {
        for (size_t i = 0; i < genomes.size(); ++i)
        {
            if (asymmetric && i == 0)
                continue;
            
            std::map<unsigned, vector<size_t> > genePos;
            for (size_t pos = 0; pos < seqan::length(genomes[i]); ++pos)
            {
                genePos[std::abs(genomes[i][pos])].push_back(pos);
            }
            for (size_t j = 0; j < genomes.size(); ++j)
            {
                if (i == j)
                    continue;
                
                if (asymmetric && j == 0)
                    continue;
#ifdef DEBUGILP
                cerr << "adding gene id constraints " << i << "  -> " << j << endl;
#endif
                std::map<unsigned, vector<size_t> >::iterator geneIt = genePos.begin();
                
                while(geneIt != genePos.end())
                {
                    IloExpr constr(env);
                    vector<size_t>::iterator posIt;
                    for (posIt = geneIt->second.begin(); posIt != geneIt->second.end(); ++posIt)
                    {
                        constr += vars.lossVars[i][j][*posIt];
                        for (size_t posJ = 0; posJ < seqan::length(genomes[j]); ++posJ)
                        {
                            if (i < j && cbData.pairGraphs[i][j].hasAlignEdge(*posIt, posJ))
                            {
                                constr += vars.alignEdgesVars[i][j][cbData.pairGraphs[i][j].vertexToIloId(*posIt, posJ)];
                            }
                            if (i > j && cbData.pairGraphs[j][i].hasAlignEdge(posJ, *posIt))
                            {
                                constr += vars.alignEdgesVars[j][i][cbData.pairGraphs[j][i].vertexToIloId(posJ, *posIt)];
                            }
                        }
                    }
                    model.add(constr >= 1);
                    constr.end();
                    
                    ++geneIt;
                }
            }
        }
    }
  
    //#endif
    
#ifdef PAIRWISECONFLICT
    cerr << "Simulate Competitor Model" << endl;
    IloCplex cplex(model);
    cplex.exportModel("lpex1.lp");
    cplex.setOut(cerr);
    
    try{
        
        MixedCycleFinder mcFinder(cbData.pairGraphs);
        int cuts = 1;
        
        while (cuts)
        {
            cuts = 0;
            double eps = cplex.getParam(cplex.EpRHS);
            cplex.solve();
            std::cerr << "solved iteration" << endl;
            
            
            //Check all genomes for duplication cycles.
            for (size_t i = 0; i < genomes.size(); ++i)
            {
                IloNumArray duplicationVals(env, vars.duplicationVars[i].getSize());
                cplex.getValues(duplicationVals, vars.duplicationVars[i]);
                
                
                //adding cut for each cyclic duplication
                DuplicationEventHolder::TviolatedDupList cyclesIds;
                cbData.dupEvents[i].getDuplicationCycle(duplicationVals, cyclesIds, 1. - eps, true);
                
                size_t num = cyclesIds.size();
                for (size_t i = 0; i < num; ++i)
                {
                    //double score = 0;
                    IloExpr cycleCut(env);
                    for (size_t j = 0; j < cyclesIds[i].dupIds.size(); ++j)
                    {
                        cycleCut += vars.duplicationVars[i][cyclesIds[i].dupIds[j]];
                    }
                    cerr << cycleCut << endl;
                    if (!cyclesIds[i].dupIds.empty())
                    {
                        model.add(cycleCut <= (IloInt) cyclesIds[i].cycleLen - 1);
                        ++cuts;
                        //cerr << cycleCutA << "  " << cyclesIdsA[i].score << " " << score << "  " << cyclesIdsA[i].cycleLen <<  endl;
                    }
                    cycleCut.end();
                }
            }
            
            //search for violated mixed cycle
            for (size_t i = 0; i < genomes.size(); ++i)
            {
                for (size_t j = i+1; j < genomes.size(); ++j)
                {
                    cplex.getValues(vals.alignEdgesVals[i][j], vars.alignEdgesVars[i][j]);
                }
            }
            MixedCycleFinder::TMixedCycle cycle;
            size_t length = -1;
            mcFinder.findViolatedMixedCycle(vals.alignEdgesVals, cycle, length, eps);
            
            if (!cycle.empty())
            {
                IloExpr mCycleCut(env);
                for(size_t i = 0; i < cycle.size(); ++i)
                {
                    cerr << cycle[i].seqA << " " << cycle[i].seqB << " " << cycle[i].index << endl;
                    mCycleCut += vars.alignEdgesVars[cycle[i].seqA][cycle[i].seqB][cycle[i].index];
                }
                model.add(mCycleCut <= (IloInt) length - 1);
                ++cuts;
                mCycleCut.end();
            }
        }
    }
    catch (IloException& e)
    {
        e.getMessage();
        throw;
    }
    
#else
   
    //solve model
    try
    {
#ifdef DEBUGILP
        cerr << "generate Model" << endl;
#endif
        IloCplex cplex(model);
        cplex.setParam(IloCplex::PreInd, IloFalse);
        cplex.setParam(IloCplex::MIPSearch, IloCplex::Traditional);
        cplex.setParam(IloCplex::Threads, (IloNum)1);
        cplex.setParam(IloCplex::NodeSel, IloCplex::BestBound);
        cplex.setParam(IloCplex::BtTol, 0.0);
        cplex.setParam(IloCplex::CutUp, 1000.0);

#ifdef DEBUGILP
        cplex.setOut(cerr);
#else
        cplex.setOut(env.getNullStream());
#endif
        cbData.mcFinder = MixedCycleFinder(cbData.pairGraphs);
        
        cplex.use(LazyCallback(env, cbData, vars, vals, cplex.getParam(cplex.EpRHS), asymmetric));
        cplex.use(UserCallback(env, cbData, vars, vals, cplex.getParam(cplex.EpRHS), asymmetric));
        cplex.use(BranchCallbackMedian(env));

#ifdef DEBUGILP
        cerr << "Rows: " << cplex.getNrows() << endl;
        cerr << "Cols: " << cplex.getNcols() << endl;
#endif
        
        try
        {
            cplex.solve();
#ifdef DEBUGILP
            cerr << cplex.getStatus() << endl;
#endif
        }
        catch(IloException &e)
        {
            e.print(cerr);
            throw;
        }
#endif
        
#ifdef DEBUGILP
        cerr << "done solving!" << endl;
        cerr << "Distance: " << cplex.getObjValue() << endl;
#endif
        score = cplex.getObjValue();
        
        //now get result and output nonzero variables
        double eps = cplex.getParam(cplex.EpInt);
        //extract vals from cplex vars
        for (size_t i = 0; i < cbData.pairGraphs.size(); ++i)
        {
            cplex.getValues(vals.duplicationVals[i], vars.duplicationVars[i]);
            for (size_t j = 0; j < cbData.pairGraphs[i].size(); ++j)
            {
                if (i < j)
                {
                    cplex.getValues(vals.lossVals[i][j], vars.lossVars[i][j]);
                    cplex.getValues(vals.lossVals[j][i], vars.lossVars[j][i]);
                    cplex.getValues(vals.alignEdgesVals[i][j], vars.alignEdgesVars[i][j]);
                    cplex.getValues(vals.inversionVals[i][j], vars.inversionVars[i][j]);
                }
            }
        }
        
        IloNumArray penaltyVals(env, vars.penaltyVars[0].getSize());
        if (asymmetric)
        {
            cplex.getValues(vals.pairedDuplicationVals[0], vars.pairedDuplicationVars[0]);
            cplex.getValues(vals.pairedDuplicationVals[1], vars.pairedDuplicationVars[1]);
            cplex.getValues(vals.wVals[0], vars.wVars[0]);
            cplex.getValues(vals.wVals[1], vars.wVars[1]);
            cplex.getValues(penaltyVals, vars.penaltyVars[0]);
        }
        
        
        align.clear();
        
        //_getAlignment(vals, cbData,genomes, align, eps, asymmetric, penaltyVals);
        _getAlignment2(vals, cbData,genomes, align, eps, asymmetric, penaltyVals);
#ifdef DEBUGDLC
        _printAlignment(align, genomes);
#endif
        
        
#ifdef DEBUGILP
        size_t numLosses = 0;
        size_t numDuplications = 0;
        for (size_t i = 0; i < genomes.size(); ++i)
        {
            for (size_t j = 0; j < genomes.size(); ++j)
            {
                if (i == j)
                    continue;
                
                for (size_t posI = 0; posI < seqan::length(genomes[i]); ++posI)
                {
                    for (size_t posJ = 0; posJ < seqan::length(genomes[j]); ++posJ)
                    {
                        if (i < j && cbData.pairGraphs[i][j].hasAlignEdge(posI, posJ))
                        {
                            if (cplex.getValue(vars.alignEdgesVars[i][j][cbData.pairGraphs[i][j].vertexToIloId(posI, posJ)]) > 0.1)
                                cerr << "ALIGN: " << i << " " << j << " : " << posI << " -> " << posJ << std::endl;
                            
                            if (i == 1 && j == 2 && cplex.getValue(vars.wVars[0][cbData.pairGraphs[1][2].vertexToIloId(posI, posJ)]) > 0.1)
                                cerr << "W1Var: " << posI << " " << posJ << std::endl;
                            
                            if (i == 1 && j == 2 && cplex.getValue(vars.wVars[1][cbData.pairGraphs[1][2].vertexToIloId(posI, posJ)]) > 0.1)
                                cerr << "W2Var: " << posJ << " " << posI << std::endl;
                        }
                    }
                }
                
                
                for (size_t k = 0; k < (size_t)vars.lossVars[i][j].getSize(); ++k)
                {
                    if (cplex.getValue(vars.lossVars[i][j][k]) > 0.1)
                    {
                        cerr << " Loss: " << i << "  " << j << "  " << k << endl;
                        //cerr << " Penalty: " << i << "  " << k << "  : " <<  cplex.getValue(vars.penaltyVars[i][k]) << endl;
                        ++numLosses;
                    }
                }
            }
            
            for (size_t j = 0; j < (size_t)vars.duplicationVars[i].getSize(); ++j)
            {
                if (cplex.getValue(vars.duplicationVars[i][j]) > 0.1)
                    ++numDuplications;
            }
        }
        cerr << "#LOSS: " << numLosses << "  #Dup: " << numDuplications << std::endl;
#endif
        
    }
    catch(IloException &e)
    {
        e.print(cerr);
        throw;
    }
    
    env.end();
    return score;
}