// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include<MixedCycleFinder.h>
#include<CycleDetect.h>


MixedCycleFinder::MixedCycleFinder (const TPairGraphMatrix & pairGraphs)
{
    _pGs.resize(pairGraphs.size());
    for (size_t i = 0; i < pairGraphs.size(); ++i)
        _pGs[i].resize(pairGraphs.size());

    for (size_t i = 0; i < pairGraphs.size(); ++i)
    {
        for (size_t j = i + 1; j < pairGraphs.size(); ++j)
        {
            //we need two versions of each pg to compute the weights for the liftedmixedclique graph
            //the first needs to be transposed so we can compute the lifted scores for edge in direction i-->j
            //source and sink must also be swapped
            _pGs[i][j] = pairGraphs[i][j];
            SparsePairGraph::TVertexDescriptor tmp = _pGs[i][j]._source;
            _pGs[i][j]._source = _pGs[i][j]._sink;
            _pGs[i][j]._sink = tmp;
            seqan::transpose(_pGs[i][j].graph);

            //otherwise nothing needs to be done to use this graph to compute lifted scores in direction j-->i
            _pGs[j][i] = pairGraphs[i][j];
        }
    }

    _buildAlignGraph();
}

void MixedCycleFinder::_buildAlignGraph()
{
    _vMat.clear();
    _vMat.resize(_pGs.size());
    _vMat[0].resize(_pGs[0][1].strLenA());

    for (size_t i = 1; i < _pGs.size(); ++i)
    {
        _vMat[i].resize(_pGs[0][i].strLenB());
    }

    //create the vertices
    for (size_t i = 0; i < _vMat.size(); ++i)
    {
        for (size_t j = 0; j < _vMat[i].size(); ++j)
        {
            _vMat[i][j] = seqan::addVertex(_alnGraph);

            if (j > 0)
            {
                seqan::addEdge(_alnGraph, _vMat[i][j-1], _vMat[i][j]);
            }
        }
    }

    //create the inverted map to get info from vertexdecriptor
    seqan::resizeVertexMap(_alnGraph, _alnGraphToPG);
    for (size_t i = 0; i < _vMat.size(); ++i)
    {
        for (size_t j = 0; j < _vMat[i].size(); ++j)
        {
            seqan::assignProperty(_alnGraphToPG, _vMat[i][j], TVertexInfo(i,j));
        }
    }

    //create the alignment edges
    for (size_t i = 0; i < _vMat.size(); ++i)
    {
        for (size_t j = i+1; j < _vMat.size(); ++j)
        {
            for (size_t posI = 0; posI < _vMat[i].size(); ++posI)
            {
                for (size_t posJ = 0; posJ < _vMat[j].size(); ++posJ)
                {
                    if (_pGs[i][j].hasAlignEdge(posI, posJ))
                    {
                        seqan::addEdge(_alnGraph, _vMat[i][posI], _vMat[j][posJ]);
                        seqan::addEdge(_alnGraph, _vMat[j][posJ], _vMat[i][posI]); //the reversal
                    }
                }
            }
        }
    }
    seqan::resizeEdgeMap(_alnGraph, _edgeWeights, 0.);
}


void MixedCycleFinder::findViolatedMixedCycle(const TWeightMatrix & weights,
                                              TMixedCycle & CycleIloIds,
                                              size_t & length,
                                              double eps)
{
    typedef seqan::VertexDescriptor<TDirGraph>::Type TVertexDescriptor;
    typedef seqan::String<TVertexDescriptor> TPredMap;
    
    std::vector<TEdgeInfo> initialCycle;
    _prepareAlnGraph(weights);

    double min_dist = 1 - eps;

    //search for violated mixed cycle in this graph
    for (size_t i = 0; i < _vMat.size(); ++i)
    {
        for (size_t j = 1; j < _vMat[i].size(); ++j)
        {
            TVertexDescriptor startV = _vMat[i][j];
            TVertexDescriptor stopV = _vMat[i][j-1];
            TDistanceMap dist;
            TPredMap predMap;
            dijkstra(_alnGraph, startV, _edgeWeights, predMap, dist, min_dist);

            if (dist[stopV] < min_dist)
            {                
                initialCycle.clear();
                min_dist = dist[stopV];

                TVertexDescriptor currentV = stopV;
                do
                {
                    TVertexDescriptor pred = seqan::getProperty(predMap, currentV);
                    const TVertexInfo & predInfo = seqan::property(_alnGraphToPG, pred);
                    const TVertexInfo & currentInfo = seqan::property(_alnGraphToPG, currentV);
                    if (predInfo.seq != currentInfo.seq) //if aln edge and no arc
                    {
                        TEdgeInfo tmp = {predInfo.seq, currentInfo.seq, predInfo.pos, currentInfo.pos};             
                        initialCycle.push_back(tmp);
                    }
                    currentV = pred;
                }
                while(currentV != startV);
            }
        }
    }

    //find all associated alignment edges
    std::vector<TEdgeInfo>::const_iterator it = initialCycle.begin();
    for (; it != initialCycle.end(); ++it)
    {
        size_t seqs = it->seqS;
        size_t seqt = it->seqT;
        const SparsePairGraph & tmpPG = _pGs[seqs][seqt];
        const TPredMap & tmpPred = _predMaps[seqs][seqt];

        //std::cerr << seqs << "  " << seqt << std::endl;
        //std::cerr << it->posS << " " << it->posT << " size 1: " << tmpPG.vertices.size() << std::endl;
        TVertexDescriptor v = seqs < seqt ? tmpPG.getVertexDesc(it->posS,it->posT) : tmpPG.getVertexDesc(it->posT, it->posS);
        TVertexDescriptor source = tmpPG._source;

        while(v != source)
        {
            size_t iloId = tmpPG.vertexToIloId(v);
            std::pair<size_t, size_t>tmppos;
            tmpPG.iloIdToVertex(iloId, tmppos.first, tmppos.second);
            
            if (seqs < seqt)
                CycleIloIds.push_back(TMixedCycleIndex(seqs, seqt, iloId, tmppos.first, tmppos.second));
            else
                CycleIloIds.push_back(TMixedCycleIndex(seqt, seqs, iloId, tmppos.second, tmppos.first));

            v = seqan::getProperty(tmpPred, v);
        }
    }

    length = initialCycle.size();
}




void MixedCycleFinder::_prepareAlnGraph(const TWeightMatrix & weights)
{
    //do for all pairgraphs:
    //assign weights to pairGraph vertices
    //compute shortest path tree in each pairgraph
    //assign the weights to the edges in the alignment graph
    typedef seqan::String< seqan::String<TDistanceMap> > TDistMapMatrix;
    typedef seqan::VertexDescriptor<TDirGraph>::Type TVertexDescriptor;

    //prepare maps
    TDistMapMatrix distMaps;
    seqan::resize(distMaps, _pGs.size());
    seqan::resize(_predMaps, _pGs.size());
    for (size_t i = 0; i < _pGs.size(); ++i)
    {
        seqan::resize(_predMaps[i], _pGs.size());
        seqan::resize(distMaps[i], _pGs.size());
    }

    for (size_t i = 0; i < _pGs.size(); ++i)
    {
        for (size_t j = 0; j < _pGs.size(); ++j)
        {
            if (i == j)
                continue;
            _pGs[i][j].assignWeights(weights[std::min(i,j)][std::max(i,j)]);
            _pGs[i][j].createShortestPathTree(distMaps[i][j], _predMaps[i][j], _pGs[i][j]._source);
        }
    }

    //use the computed dists to create weights in our alignment graph
    typedef seqan::Iterator<TGraph, seqan::EdgeIterator>::Type TEdgeIterator;
    for (TEdgeIterator edge(_alnGraph); !seqan::atEnd(edge);  seqan::goNext(edge))
    {
        const TVertexInfo & sv = seqan::property(_alnGraphToPG, seqan::sourceVertex(_alnGraph, *edge));
        const TVertexInfo & tv = seqan::property(_alnGraphToPG, seqan::targetVertex(_alnGraph, *edge));

        if (sv.seq == tv.seq)
        {
            seqan::assignProperty(_edgeWeights, *edge, 0);
            continue;
        }

        const SparsePairGraph & pgtmp = _pGs[sv.seq][tv.seq];
        TVertexDescriptor vId;

        if (sv.seq > tv.seq)
            vId = pgtmp.getVertexDesc(tv.pos, sv.pos);
        else
            vId = pgtmp.getVertexDesc(sv.pos, tv.pos);

        size_t index = pgtmp.vertexToIloId(vId);
        double weight = weights[std::min(sv.seq,tv.seq)][std::max(sv.seq,tv.seq)][index];
        //std::cerr << "before: " << weight << " ";
        weight -= seqan::property(distMaps[sv.seq][tv.seq], vId);
        //std::cerr << "  after: " << weight << std::endl;
        seqan::assignProperty(_edgeWeights, *edge, 1. - weight);
    }
}
