// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <SparsePairGraph.h>
#include <DuplicationEventHolder.h>

const size_t MINUSONE = -1;

void SparsePairGraph::createPG(const GeneString & strA,
                               const GeneString & strB,
                               bool inversions)
{
    _inversions = inversions;
    _strLenA = seqan::length(strA);
    _strLenB = seqan::length(strB);

    if (_inversions)
        _inversionHolder.createInversionEvents(strA, strB);

    std::vector<std::list<std::pair<int, TVertexDescriptor> > >rows(_strLenA);

    _vertices.assign(_strLenA, TVertexList(_strLenB, seqan::getNil<TVertexDescriptor>()));

    TVertexDescriptor v;
    for (size_t col = 0; col < _strLenB; ++col)
    {
        for (int row = _strLenA - 1; row >= 0; --row)
        {
            if (strA[row] == strB[col] || _inversionHolder.hasInversion(row, col)) // add node to pairgraph and edges
            {
                v = seqan::addVertex(graph);
#ifdef DEBUGDLC
                std::cerr << v << " == " <<"[" << row <<"][" << col << "]" << std::endl;
#endif
                _vertices[row][col] = v;

                if (!rows[row].empty())
                {
                    TVertexDescriptor pred = rows[row].back().second;
                    seqan::addEdge(graph, v, pred);
                }
                rows[row].push_back(std::make_pair(col, v));

                //check upwards to add the required edges
                int maxCol = -1;
                for (size_t upRow = row + 1; upRow < _strLenA; ++upRow)
                {
                    if (!rows[upRow].empty() && rows[upRow].back().first > maxCol)
                    {
                        maxCol = rows[upRow].back().first;
                        seqan::addEdge(graph, v, rows[upRow].back().second);
                    }
                }
            }
        }
    }
    //adding source and sink
    _source = seqan::addVertex(graph);
    _sink = seqan::addVertex(graph);

    for (size_t row = 0; row < _strLenA; ++ row)
    {
        if(!rows[row].empty())
        {
            if (seqan::inDegree(graph, rows[row].back().second) == 0)
            {
                seqan::addEdge(graph, _source, rows[row].back().second);
            }
            if (seqan::outDegree(graph, rows[row].front().second) == 0)
            {
                seqan::addEdge(graph, rows[row].front().second, _sink);
            }
        }
    }
    //fill inverse matrix
    seqan::resizeVertexMap(graph, _verticesInv, std::make_pair(-1,-1));
    seqan::resizeVertexMap(graph, _vertexToIloId, -1);
    _iloIdToVertex.resize(seqan::numVertices(graph) - 2); // -2 because of artificial vertices

    if (_inversions)
        seqan::resizeVertexMap(graph,_inversionMap);


    size_t id = 0;
    for (size_t col = 0; col < _strLenB; ++col)
    {
        for (int row = _strLenA - 1; row >= 0; --row)
        {
            if(_vertices[row][col] != seqan::getNil<TVertexDescriptor>())
            {
                seqan::assignProperty(_verticesInv, _vertices[row][col], std::make_pair(row, col));
                if (strA[row] == strB[col])
                {
                    seqan::assignProperty(_vertexToIloId, _vertices[row][col], id);
                    _iloIdToVertex[id] = _vertices[row][col];
                    ++id;
                }
                if (_inversions && _inversionHolder.hasInversion(row,col))
                {
                    seqan::assignProperty(_inversionMap, _vertices[row][col], _inversionHolder.EdgeMap[row][col]);
                }
            }
        }
    }

    _iloIdToVertex.resize(id);
    _numEdges = id;
    //prepare the edge weight map
    seqan::resizeEdgeMap(graph, _edgeWeights, 0);
}


void SparsePairGraph::createAugmentedPG(const GeneString & strA,
                                        const GeneString & strB,
                                        bool inversions)
{
    _inversions = inversions;
    
    _strLenA = seqan::length(strA);
    _strLenB = seqan::length(strB);

    if (_inversions)
        _inversionHolder.createInversionEvents(strA, strB);

    std::vector<std::list<std::pair<int, TVertexDescriptor> > >rows(_strLenA);

    _vertices.assign(_strLenA, TVertexList(_strLenB, seqan::getNil<TVertexDescriptor>()));

    TVertexDescriptor v;
    for (size_t col = 0; col < _strLenB; ++col)
    {
        for (size_t row = _strLenA - 1; (int)row >= 0; --row)
        {
            if (strA[row] == strB[col] || _inversionHolder.hasInversion(row, col) || row == 0 || row == _strLenA - 1 || col == 0 || col == _strLenB - 1) // add node to pairgraph and edges
            {
                v = seqan::addVertex(graph);
#ifdef DEBUGDLC
                std::cerr << v << " == " <<"[" << row <<"][" << col << "]" << std::endl;
#endif
                _vertices[row][col] = v;

                if (!rows[row].empty())
                {
                    TVertexDescriptor pred = rows[row].back().second;
                    seqan::addEdge(graph, v, pred);
                }
                rows[row].push_back(std::make_pair(col, v));

                //check upwards to add the required edges
                int maxCol = -1;
                for (size_t upRow = row + 1; upRow < _strLenA; ++upRow)
                {
                    if (!rows[upRow].empty() && rows[upRow].back().first > maxCol)
                    {
                        maxCol = rows[upRow].back().first;
                        seqan::addEdge(graph, v, rows[upRow].back().second);
                    }
                }
            }
        }
    }
    //adding source and sink
    _source = seqan::addVertex(graph);
    _sink = seqan::addVertex(graph);

    for (size_t row = 0; row < _strLenA; ++ row)
    {
        if(!rows[row].empty())
        {
            if (seqan::inDegree(graph, rows[row].back().second) == 0)
            {
                seqan::addEdge(graph, _source, rows[row].back().second);
            }
            if (seqan::outDegree(graph, rows[row].front().second) == 0)
            {
                seqan::addEdge(graph, rows[row].front().second, _sink);
            }
        }
    }
    //fill inverse matrix
    seqan::resizeVertexMap(graph, _verticesInv, std::make_pair(-1,-1));
    seqan::resizeVertexMap(graph, _vertexToIloId, -1);
    _iloIdToVertex.resize(seqan::numVertices(graph) - 2); // -2 because of artificial vertices
    
    if (_inversions)
        seqan::resizeVertexMap(graph,_inversionMap);

    size_t id = 0;
    for (size_t col = 0; col < _strLenB; ++col)
    {
        for (int row = _strLenA - 1; row >= 0; --row)
        {
            if(_vertices[row][col] != seqan::getNil<TVertexDescriptor>())
            {
                seqan::assignProperty(_verticesInv, _vertices[row][col], std::make_pair(row, col));

                if (strA[row] == strB[col]) //only for real vertices add the corresponding entries
                {
                    seqan::assignProperty(_vertexToIloId, _vertices[row][col], id);
                    _iloIdToVertex[id] = _vertices[row][col];
                    ++id;
                }
                if (_inversions && _inversionHolder.hasInversion(row,col))
                {
                    seqan::assignProperty(_inversionMap, _vertices[row][col], _inversionHolder.EdgeMap[row][col]);
                }
            }
        }
    }

    _iloIdToVertex.resize(id);
    _numEdges = id;
    //prepare the edge weight map
    seqan::resizeEdgeMap(graph, _edgeWeights, 0);
}


void SparsePairGraph::getMixedCliques(DuplicationEventHolder &holderA,
                                      DuplicationEventHolder &holderB,
                                      const IloNumArray &inversionScores,
                                      const IloNumArray &dupAScores,
                                      const IloNumArray &dupBScores,
                                      TMixedCliqueList &cliques,
                                      IloNum eps)
{
    if (_inversions)
        _inversionsToEdges(inversionScores, eps);
//    else
//        std::cerr << "No Inversions Defined - Ignoring scores" << std::endl;
    
    return getMixedCliques(holderA, holderB, dupAScores, dupBScores, cliques, eps);
}

void SparsePairGraph::getMixedCliques(DuplicationEventHolder &holderA,
                                      const IloNumArray &inversionScores,
                                      const IloNumArray &dupAScores,
                                      TMixedCliqueList &cliques,
                                      IloNum eps,
                                      size_t id)
{
    if (_inversions)
        _inversionsToEdges(inversionScores, eps);
//    else
//        std::cerr << "No Inversions Defined - Ignoring scores" << std::endl;
    
    return getMixedCliques(holderA, dupAScores, cliques, eps, id);
}



//TODO: get rid of massive duplication - only very few differences for A and B
void SparsePairGraph::getMixedCliques(DuplicationEventHolder &holderA,
                                      DuplicationEventHolder &holderB,
                                      const IloNumArray &dupAScores,
                                      const IloNumArray &dupBScores,
                                      TMixedCliqueList &cliques,
                                      IloNum eps)
{
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;
    
    holderA.DPCompute(dupAScores);
    holderB.DPCompute(dupBScores);

    size_t lenA = seqan::length(_vertices);
    size_t lenB = seqan::length(_vertices[0]);

    //search mixed cliques with duplication in strA
    if (!holderA.duplications.empty())
    {
        seqan::String<seqan::String<TVertexDescriptor> >predMapsA;
        seqan::String<seqan::String<double> > distMapsA;

        seqan::resize(predMapsA, lenA);
        seqan::resize(distMapsA, lenA);

        //compute shortest path trees in sequence A
        for (size_t i = 0; i < lenA; ++i)
        {
            if (holderA.piV[i][i] > eps && holderA.piV[i][i] < 1)
                seqan::dagShortestPath(graph, _vertices[i][lenB - 1], _edgeWeights, predMapsA[i], distMapsA[i]);
        }

        for (size_t i = 0; i < lenA; ++i)
        {
            if (holderA.piV[i][i] < eps || holderA.piV[i][i] > 1 - eps)
                continue; //no violated mixed clique possible if this holds

            TVertexDescriptor source = _vertices[i][lenB - 1];
            for (size_t j = i; j < lenA && holderA.sigV[i][j] > eps; ++j)
            {   //stop as soon as holderA.sigV[i][j] becomes 0 -> will never increase again for greater j

                if (holderA.hasDup(i,j) && holderA.pirV[j][j] > eps && holderA.pirV[j][j] < 1)
                {
                    TVertexDescriptor sink = _vertices[j][0];

                    double score = (-1) * seqan::getProperty(distMapsA[i], sink);
                    
                    if (_vertexToIloId[sink] != MINUSONE)
                    {
                        TOutEdgeIterator oit(graph,sink);
                        score -= seqan::getProperty(_edgeWeights, *oit); //correct for missing score on sink vertex in PG
                    }
                    
                    score += holderA.sigV[i][j];

                    if(score > 1. + eps)
                    {
                        TVertexList clique;
                        TSizeArray alnEdgeIds;
                        //traceback to get the violated clique
                        TVertexDescriptor v = sink;
                        while (true)
                        {
                            if (seqan::getProperty(_vertexToIloId, v) != MINUSONE)
                            {
                                clique.push_back(v); //if not a fake
                                alnEdgeIds.push_back(seqan::getProperty(_vertexToIloId, v));
                            }

                            if (v == source)
                                break;

                            v = seqan::getProperty(predMapsA[i], v);
                        }
                        MixedClique tmp = {0, i, j, score, clique, alnEdgeIds, holderA.sigI[i][j]};
                        cliques.push_back(tmp);
                    }
                }
            }
        }
    }

    //search mixed cliques with duplication in strA
    if (!holderB.duplications.empty())
    {       
        seqan::String<seqan::String<TVertexDescriptor> >predMapsB;
        seqan::String<seqan::String<double> > distMapsB;
        seqan::resize(predMapsB, lenB);
        seqan::resize(distMapsB, lenB);

        //compute shortest path trees in sequence B
        for (size_t i = 0; i < lenB; ++i)
        {
            if (holderB.pirV[i][i] > eps && holderB.pirV[i][i] < 1)
                seqan::dagShortestPath(graph, _vertices[0][i], _edgeWeights, predMapsB[i], distMapsB[i]);
        }

        for (size_t i = 0; i < lenB; ++i)
        {
            if (holderB.piV[i][i] < eps || holderB.piV[i][i] > 1 - eps)
                continue;

            TVertexDescriptor sink = _vertices[lenA - 1][i];
            for (size_t j = i; j < lenB && holderB.sigV[i][j] > eps; ++j)
            {
                if (holderB.hasDup(i,j) && holderB.pirV[j][j] > eps && holderB.pirV[j][j] < 1)
                {
                    TVertexDescriptor source = _vertices[0][j];

                    double score = (-1) * seqan::getProperty(distMapsB[j], sink);
                    
                    if (_vertexToIloId[sink] != MINUSONE)
                    {
                        TOutEdgeIterator oit(graph,sink);
                        score -= seqan::getProperty(_edgeWeights, *oit); //correct for missing score on sink vertex in PG
                    }
                    
                    score += holderB.sigV[i][j];
                    
                    if(score > 1. + eps)
                    {
                        TVertexList clique;
                        TSizeArray alnEdgeIds;
                        //traceback to get the violated clique
                        TVertexDescriptor v = sink;
                        while (true)
                        {
                            if (seqan::getProperty(_vertexToIloId, v) != MINUSONE)
                            {
                                clique.push_back(v); //if not a fake
                                alnEdgeIds.push_back(seqan::getProperty(_vertexToIloId, v));
                            }

                            if (v == source)
                                break;

                            v = seqan::getProperty(predMapsB[j], v);
                        }

                        MixedClique tmp = {1, i, j, score, clique, alnEdgeIds, holderB.sigI[i][j]};
                        cliques.push_back(tmp);
                    }
                }
            }
        }
    }
}

//TODO: get rid of massive duplication - only very few differences for A and B
void SparsePairGraph::getMixedCliques(DuplicationEventHolder &holder,
                                      const IloNumArray &dupScores,
                                      TMixedCliqueList &cliques,
                                      IloNum eps,
                                      size_t id)
{    
    holder.DPCompute(dupScores);
    
    size_t lenA = seqan::length(_vertices);
    size_t lenB = seqan::length(_vertices[0]);
    
    //switch len if cuts for second sequence
    if (id == 1)
    {
        size_t tmp = lenA;
        lenA = lenB;
        lenB = tmp;
    }
    
    //search mixed cliques with duplication in str id
    if (!holder.duplications.empty())
    {
        seqan::String<seqan::String<TVertexDescriptor> >predMaps;
        seqan::String<seqan::String<double> > distMaps;
        
        seqan::resize(predMaps, lenA);
        seqan::resize(distMaps, lenA);
        
        //compute shortest path trees in sequence A
        for (size_t i = 0; i < lenA; ++i)
        {
            if (id == 0 && holder.piV[i][i] > eps && holder.piV[i][i] < 1)
                seqan::dagShortestPath(graph, _vertices[i][lenB - 1], _edgeWeights, predMaps[i], distMaps[i]);
            
            else if (id == 1 && holder.pirV[i][i] > eps && holder.pirV[i][i] < 1)
                seqan::dagShortestPath(graph, _vertices[0][i], _edgeWeights, predMaps[i], distMaps[i]);
        }
        
        TVertexDescriptor source, sink;
        for (size_t i = 0; i < lenA; ++i)
        {
            if (holder.piV[i][i] < eps || holder.piV[i][i] > 1 - eps)
                continue; //no violated mixed clique possible if this holds
            
            if (id == 0)
                source = _vertices[i][lenB - 1]; //lower right
            else if (id == 1)
                sink = _vertices[lenB - 1][i]; //upper left
            
            for (size_t j = i; j < lenA && holder.sigV[i][j] > eps; ++j)
            {   //stop as soon as holderA.sigV[i][j] becomes 0 -> will never increase again for greater j
                
                if (holder.hasDup(i,j) && holder.pirV[j][j] > eps && holder.pirV[j][j] < 1)
                {
                    if (id == 0)
                        sink = _vertices[j][0]; //upper left
                    else if (id == 1)
                        source = _vertices[0][j]; //lower right
                    
                    double score = 0;
                    if (id == 0)
                        score = (-1) * seqan::getProperty(distMaps[i], sink);
                    else if (id == 1)
                        score = (-1) * seqan::getProperty(distMaps[j], sink);
                    
                    if (_vertexToIloId[sink] != MINUSONE)
                    {
                        TOutEdgeIterator oit(graph,sink);
                        score -= seqan::getProperty(_edgeWeights, *oit); //correct for missing score on sink vertex in PG
                    }
                    
                    score += holder.sigV[i][j];
                    
                    if(score > 1. + eps)
                    {
                        TVertexList clique;
                        TSizeArray alnEdgeIds;
                        //traceback to get the violated clique
                        TVertexDescriptor v = sink;
                        while (true)
                        {
                            if (seqan::getProperty(_vertexToIloId, v) != MINUSONE)
                            {
                                clique.push_back(v); //if not a fake
                                alnEdgeIds.push_back(seqan::getProperty(_vertexToIloId, v));
                            }
                            
                            if (v == source)
                                break;
                            
                            if (id == 0)
                                v = seqan::getProperty(predMaps[i], v);
                            else if (id == 1)
                                v = seqan::getProperty(predMaps[j], v);
                        }                       

                        MixedClique tmp = {id, i, j, score, clique, alnEdgeIds, holder.sigI[i][j]};
                        cliques.push_back(tmp);
                    }
                }
            }
        }
    }
}

void SparsePairGraph::getMixedCliquesPaired(PairedDuplicationEventHolder3 &holderP,
                                            DuplicationEventHolder & holderC,
                                            const IloNumArray &dupPScores,
                                            const IloNumArray &dupCScores,
                                            TMixedCliqueList &cliques,
                                            IloNum eps)
{
    holderC.DPCompute(dupCScores);
    holderP.DPCompute(dupPScores);    
    holderP.combineScores(holderC);
    
    size_t lenA = seqan::length(_vertices);
    size_t lenB = seqan::length(_vertices[0]);

    //search mixed cliques with duplication in strC
    if (!holderP.duplications.empty() || !holderC.duplications.empty())
    {       
        seqan::String<seqan::String<TVertexDescriptor> > predMapsB;
        seqan::String<seqan::String<double> > distMapsB;
        seqan::resize(predMapsB, lenB);
        seqan::resize(distMapsB, lenB);
        
        //compute shortest path trees in sequence B
        for (size_t i = 0; i < lenB; ++i)
        {
            if (holderP.pirV[i][i] > eps && holderP.pirV[i][i] < 1)
                seqan::dagShortestPath(graph, _vertices[0][i], _edgeWeights, predMapsB[i], distMapsB[i]);
        }
        
        for (size_t i = 0; i < lenB; ++i)
        {
            if (holderP.piV[i][i] < eps || holderP.piV[i][i] > 1 - eps)
                continue;
            
            TVertexDescriptor sink = _vertices[lenA - 1][i];
            for (size_t j = i; j < lenB && holderP.sigV[i][j] > eps; ++j)
            {
                if ( (holderP.hasDup(i,j) || holderC.hasDup(i,j)) && holderP.pirV[j][j] > eps && holderP.pirV[j][j] < 1)
                {
                    TVertexDescriptor source = _vertices[0][j];

                    double score = (-1) * seqan::getProperty(distMapsB[j], sink);
                    
                    if (_vertexToIloId[sink] != MINUSONE)
                    {
                        TOutEdgeIterator oit(graph,sink);
                        score -= seqan::getProperty(_edgeWeights, *oit); //correct for missing score on sink vertex in PG
                    }
                    
                    double scoreTmp = score;
                    
                    score += holderP.sigV[i][j];
                    
                    if(score > 1. + eps)
                    {
                        //traceback to get the violated clique
                        TVertexList clique;
                        TSizeArray alnEdgeIds;

                        TVertexDescriptor v = sink;
                        double tmpS = 0;
                        while (true)
                        {
                            if (seqan::getProperty(_vertexToIloId, v) != MINUSONE)
                            {
                                clique.push_back(v); //if not a fake
                                alnEdgeIds.push_back(seqan::getProperty(_vertexToIloId, v));
                                
                                TOutEdgeIterator oit(graph,v);
                                tmpS -= seqan::getProperty(_edgeWeights, *oit);
                            }
                            
                            if (v == source)
                                break;
                            
                            v = seqan::getProperty(predMapsB[j], v);
                        }
                        assert (fabs(tmpS - scoreTmp) < 0.01);
                        MixedClique tmp = {1, i, j, score, clique, alnEdgeIds, holderC.sigI[i][j], TSizeArray(), holderP.sigI[i][j]};
                        cliques.push_back(tmp);
                    }
                }
            }
        }
    }
}


void SparsePairGraph::prepareIntervalQueries()
{
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;
    
    size_t lenA = seqan::length(_vertices);
    size_t lenB = seqan::length(_vertices[0]);
   
    seqan::resize(_predMapsB, lenB);
    seqan::resize(_distMapsB, lenB);
    
    seqan::resize(_predMapsA, lenA);
    seqan::resize(_distMapsA, lenA);
   
    
    //compute shortest path trees in sequence B
    for (size_t i = 0; i < lenA; ++i)
    {
        seqan::dagShortestPath(graph, _vertices[i][lenB - 1], _edgeWeights, _predMapsA[i], _distMapsA[i]);
    }
    
    //compute shortest path trees in sequence B
    for (size_t i = 0; i < lenB; ++i)
    {
        seqan::dagShortestPath(graph, _vertices[0][i], _edgeWeights, _predMapsB[i], _distMapsB[i]);
    }
}

double SparsePairGraph::getMaxIntervalClique(size_t startPos,
                                            size_t endPos,
                                            size_t id,
                                            std::vector<size_t> & clique)
{
    typedef seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type TOutEdgeIterator;

    TVertexDescriptor source, sink;
    if (id == 0)
    {
        source = _vertices[startPos][strLenB() - 1];
        sink = _vertices[endPos][0];
    }
    else if(id == 1)
    {
        source = _vertices[0][endPos];
        sink = _vertices[strLenA() - 1][startPos];
    }
    else
    {
        std::cerr << "INVALID ID" << std::endl;
        exit(0);
    }
    

    double score = 0;
    if (_vertexToIloId[sink] != MINUSONE)
    {
        //correct for sink weight
        TOutEdgeIterator outIter(graph, sink);
        score += seqan::getProperty(_edgeWeights, *outIter);
    }

    if (id == 0)
        score += seqan::getProperty(_distMapsA[startPos], sink);
    else if (id == 1)
    {
        score += seqan::getProperty(_distMapsB[endPos], sink);
    }
    
    //create the clique
    clique.clear();
    TVertexDescriptor v = sink;
    //std::cerr << source << " <---> " << sink << std::endl;
    while (true)
    {
        //std::cerr << "V: " << v << std::endl;
        if (seqan::getProperty(_vertexToIloId, v) != MINUSONE)
        {
            clique.push_back(seqan::getProperty(_vertexToIloId, v));
        }
        
        if (v == source)
            break;
        
        if (id == 0)
            v = seqan::getProperty(_predMapsA[startPos], v);
        else if (id == 1)
            v = seqan::getProperty(_predMapsB[endPos], v);
    }
    return (-1) * score;
}
