// Copyright (c) 2013, Sandro Andreotti and Stefan Canzar,
// FU Berlin and Johns Hopkins University School of Medicine, Baltimore
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the authors or the contributing institutions nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <Subtour.h>

#include <lemon/preflow.h>

using namespace lemon;

void _get_violated_cuts(ListDigraph& g, 
                        ArcLookUp<ListDigraph> & ae,
                        ListDigraph::Node* nodes,
                        ListDigraph::ArcMap<double>& cap,
                        int s_idx,
                        int length_string,
                        IloNum eps,
                        std::set<ViolatedCut>& cuts)
{
    //concepts::WriteMap<ListDigraph::Node, bool> cut;

    ListDigraph::Node s = nodes[s_idx];
    ListDigraph::NodeMap<bool> cut(g);
    for(size_t p = 0; p < length_string; ++p)
    {
        if(p == s_idx)
            continue;

        ListDigraph::Arc a  = ae(s, nodes[p]);
        if( a != INVALID && cap[a] >= 1. - eps)
            continue;

        Preflow<ListDigraph, ListDigraph::ArcMap<double> > maxflow(g, cap, s, nodes[p]);
        //maxflow.run();
        maxflow.init();
        maxflow.startFirstPhase();
        if( maxflow.flowValue() < 1. - eps ) {
            ViolatedCut tmpcut;
            tmpcut.cutvalue = maxflow.flowValue();
            maxflow.minCutMap(cut);
            for(int j=0; j<length_string; j++) {
                if(!cut[nodes[j]]) {
                    tmpcut.node_idc.push_back(j);
                }
            }
            cuts.insert(tmpcut);
        }
    }
}

void _get_reverse_violated_cuts(ListDigraph& g, 
                                ListDigraph::Node* nodes,
                                int s_idx,
                                ArcLookUp<ListDigraph>& ae,
                                ListDigraph::ArcMap<double>& cap,
                                ListDigraph::ArcMap<double>& dupcap,
                                ListDigraph::ArcMap<bool>& is_duparc,
                                ListDigraph::NodeMap<double>& sum_lossalign,
                                int length_string,
                                std::set<ViolatedCut>& cuts,
                                IloNum eps)
{
    ListDigraph::Node s = nodes[s_idx];
    ListDigraph::NodeMap<bool> cut(g);
    for(size_t p = 0; p < length_string; ++p)
    {
        if(p == s_idx)
            continue;
        //update graph for max flow from nodes[p] to s
        for(ListDigraph::ArcIt a(g); a != INVALID; ++a)
        {
            if(!is_duparc[a]) {
                ListDigraph::Node source = g.source(a);
                g.erase(a);
                ae.refresh(source);
            }
            else {
                cap[a] = dupcap[a];
            }
        }

        ListDigraph::Arc a = ae(nodes[p], s);
        if( a != INVALID && cap[a] + sum_lossalign[s] >= 1. - eps )
            continue;

        ListDigraph::Node t = nodes[p];
        for(size_t pp = 0; pp < length_string; ++pp)
        {
            if(pp == p)
                continue;

            if(sum_lossalign[nodes[pp]] >= eps)
            {
                ListDigraph::Arc a = ae(t, nodes[pp]);
                if( a != INVALID )
                    cap[a] += sum_lossalign[nodes[pp]];
                else {
                    ListDigraph::Arc a = g.addArc(t, nodes[pp]);
                    cap[a] = sum_lossalign[nodes[pp]];
                    is_duparc[a] = false;
                    ae.refresh(t);
                }
            }


        }

        Preflow<ListDigraph, ListDigraph::ArcMap<double> > maxflow(g, cap, t, s);
        //maxflow.run();
        maxflow.init();
        maxflow.startFirstPhase();
        if( maxflow.flowValue() < 1. - eps ) {
            ViolatedCut tmpcut;
            tmpcut.cutvalue = maxflow.flowValue();
            maxflow.minCutMap(cut);
            for(int j=0; j<length_string; j++) {
                if(!cut[nodes[j]])
                    tmpcut.node_idc.push_back(j);
            }
            cuts.insert(tmpcut);
        }
    }
}


void _create_flowgraph_s(ListDigraph & g,
                         ListDigraph::Node* nodes,
                         ArcLookUp<ListDigraph> & ae,
                         ListDigraph::ArcMap<double> & cap,
                         ListDigraph::ArcMap<double> & dupcap,
                         ListDigraph::ArcMap<bool> & is_duparc,
                         int s_idx,
                         DuplicationEventHolder& dupEvents,
                         const IloNumArray &duplicationsVal,
                         const IloNumArray &lossVal,
                         const IloNumArray &alignEdgesVal,
                         SparsePairGraph & pairGraph,
                         ListDigraph::NodeMap<double> & sum_lossalign,
                         int length_otherstring,
                         IloNum eps,
                         bool stringA)
{
    for(size_t i = 0; i < dupEvents.duplications.size(); ++i)
    {
        if( duplicationsVal[i] >= eps )
        {
            for (size_t j = dupEvents.duplications[i].sourceBegin, k = 0; j <= dupEvents.duplications[i].sourceEnd; ++j, ++k)
            {
                ListDigraph::Arc a = ae(nodes[j], nodes[dupEvents.duplications[i].targetBegin + k]);
                if( a != INVALID ) {
                    cap[a] += duplicationsVal[i];
                    dupcap[a] += duplicationsVal[i];
                }
                else
                {
                    ListDigraph::Arc a = g.addArc(nodes[j], nodes[dupEvents.duplications[i].targetBegin + k]);
                    cap[a] = duplicationsVal[i];
                    dupcap[a] = duplicationsVal[i];
                    is_duparc[a] = true;
                    ae.refresh(nodes[j]);
                }
            }
        }
    }
    
    
    ListDigraph::Node s = nodes[s_idx];
    
    for(size_t p = 0; p < dupEvents.dupMap.size(); ++p)
    {
        sum_lossalign[nodes[p]] = 0.;
        //sum losses of vertex p
        if(lossVal[p]>=eps) {
            sum_lossalign[nodes[p]] += lossVal[p];
            if(p != s_idx)
            {
                ListDigraph::Arc a = ae(s, nodes[p]);
                if( a != INVALID )
                    cap[a] += lossVal[p];
                else
                {
                    ListDigraph::Arc a = g.addArc(s,nodes[p]);
                    cap[a] = lossVal[p];
                    is_duparc[a] = false;
                    ae.refresh(s);
                }
            }
        }
        
        //sum alignments of vertex p
        for(size_t q = 0; q < length_otherstring; ++q)
        {
            if( (stringA && pairGraph.hasAlignEdge(p,q)) ||
                    (!stringA && pairGraph.hasAlignEdge(q,p)) )
            {

                //std::cerr << pairGraph.vertices[p][q] << " !!" << std::endl;
                //std::cerr << pairGraph.VertexToIloId[pairGraph.vertices[p][q]] << std::endl;
                IloNum pq_edge_score;
                if(stringA)
                    pq_edge_score = alignEdgesVal[pairGraph.vertexToIloId(p, q)];
                else
                    pq_edge_score = alignEdgesVal[pairGraph.vertexToIloId(q, p)];
                if(pq_edge_score >= eps)
                {
                    sum_lossalign[nodes[p]] += pq_edge_score;
                    if(p != s_idx)
                    {
                        ListDigraph::Arc a = ae(s, nodes[p]);
                        if( a != INVALID )
                            cap[a] += pq_edge_score;
                        else
                        {
                            ListDigraph::Arc a = g.addArc(s,nodes[p]);
                            cap[a] = pq_edge_score;
                            is_duparc[a] = false;
                            ae.refresh(s);
                        }
                    }
                }
            }
        }
    }

}


void get_violated_cuts(DuplicationEventHolder& dupEvents,
                       const IloNumArray& duplicationsVal,
                       const IloNumArray& lossVal,
                       const IloNumArray& alignEdgesVal,
                       SparsePairGraph& pairGraph,
                       int length_otherstring,
                       IloNum eps,
                       ListDigraph& g,
                       ListDigraph::Node* nodes,
                       std::set<ViolatedCut>& cuts,
                       bool stringA)
{
    ArcLookUp<ListDigraph> ae(g);
    ListDigraph::ArcMap<double> cap(g);
    ListDigraph::ArcMap<double> dupcap(g);
    ListDigraph::ArcMap<bool> is_duparc(g);
    ListDigraph::NodeMap<double> sum_lossalign(g);
    
    int s_idx = 0; //better choice: vertex for which loss + alignment = 1
    ListDigraph::Node s = nodes[s_idx];
    _create_flowgraph_s(g, nodes, ae, cap, dupcap, is_duparc, s_idx,
                        dupEvents, duplicationsVal, lossVal,
                        alignEdgesVal, pairGraph,  sum_lossalign,
                        length_otherstring, eps, stringA);

    _get_violated_cuts(g, ae, nodes, cap, s_idx, dupEvents.dupMap.size(), eps, cuts);

    if( sum_lossalign[s] < 1. - eps )
    {
        _get_reverse_violated_cuts(g, nodes, s_idx, ae, cap, dupcap, is_duparc, sum_lossalign, dupEvents.dupMap.size(), cuts, eps);
    }
} 
